/*
 * mmu.c - memory management routines
 *
 * This file is a part of the multithreading version of the
 * SimpleScalar tool suite written by Todd M. Austin as a part
 * of the Multiscalar Research Project.
 * 
 * Copyright (C) 1997 by Dominik Madon
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dominik.madon@epfl.ch
 * Snail: LSL-DI/EPFL, In-Ecublens, 1015 Lausanne
 *
 * $Id: mmu.c,v 1.2 2002/01/08 14:00:26 yeung Exp $
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "misc.h"
#include "ss.h"
#include "mmu.h"
#include "memory.h"




/*                                                                      */
/* mmu_hash_iter applies a function to all elements of the hash list.   */
/*                                                                      */
/* The function given can stop the iteration by returning 0.            */
/*                                                                      */

void mmu_hash_iter(
  MMU_ENTRY_TYPE **mmu_hash_table,
  unsigned int (*iterator)(MMU_ENTRY_TYPE *))
{
  int hash_position;
  int exit = 0;
  MMU_ENTRY_TYPE *mmu_entry;

  for(hash_position = 0; hash_position < MMU_HASH_SIZE; hash_position++) {
    mmu_entry = mmu_hash_table[hash_position];
    while ((mmu_entry != NULL) && !exit) {
      exit = (*iterator)(mmu_entry);
      mmu_entry = mmu_entry->next;  
    } /* while */                   
  } /* for */
} /* mmu_hash_iter */




/*                                                                      */
/* mmu_hash_init creates the hash table and initialize it.              */
/* Each empty entry is filled with NULL.                                */
/*                                                                      */

MMU_ENTRY_TYPE ** mmu_hash_init()
{
  MMU_ENTRY_TYPE **mmu_hash_table;
  unsigned int   hash_entry;

  mmu_hash_table = (MMU_ENTRY_TYPE **)
    malloc(sizeof(MMU_ENTRY_TYPE *) * MMU_HASH_SIZE);
  if (mmu_hash_table == NULL) {
    fatal("cannot allocate memory for MMU's hash table\n");
  } /* if */

  for(hash_entry = 0; hash_entry < MMU_HASH_SIZE; hash_entry++)
    mmu_hash_table[hash_entry] = NULL;

  return mmu_hash_table;

} /* mmu_hash_init */



/*                                                                      */
/* mmu_hash_iter_free is used to deallocate memory reserved for the     */
/* hash table. It is applied to all entries of the table.               */
/* For all of them, it frees all the entries included in the chain.     */
/*                                                                      */
/* The first element is not freed because the whole table will be       */
/* managed one level upper.                                             */ 

unsigned int mmu_hash_iter_free(MMU_ENTRY_TYPE *mmu_hash_entry)
{
  MMU_ENTRY_TYPE *next, *node;

  node = mmu_hash_entry;
  while (node != NULL) {
    next = node->next;
    free(node);
    node = next;
  } /* while */

  return 1;

} /* mmu_hash_iter_free */



/*                                                                      */
/*  mmu_hash_free frees the mmu structure.                              */
/*                                                                      */

void mmu_hash_free(MMU_ENTRY_TYPE **mmu_hash_table)
{
  mmu_hash_iter(mmu_hash_table, mmu_hash_iter_free);
  free(mmu_hash_table);

} /* mmu_hash_free */



/*                                                                      */
/* mmu_hash_add adds an entry (logical addr, physical addr)             */
/* in the hash table.                                                   */
/*                                                                      */
/* Each hash table entry has a next field where another entry can be    */
/* chained. Entries are sorted by logical address in ascending order.   */

void mmu_hash_add(
  SS_CONTEXT_TYPE context,
  SS_ADDR_TYPE logical,
  SS_ADDR_TYPE physical,
  MMU_ENTRY_TYPE **mmu_hash_table)
{
  unsigned int   hash_index;
  MMU_ENTRY_TYPE *node, *ref;

        /* Determines hash key */
  hash_index = MMU_HASH_KEY(logical);

        /* Is the first entry of the table empty ? */
  if (mmu_hash_table[hash_index] == NULL) {
    mmu_hash_table[hash_index] =
      (MMU_ENTRY_TYPE *) malloc(sizeof(MMU_ENTRY_TYPE));
    node = mmu_hash_table[hash_index];
    node->next = NULL;
  } else {

        /* if not, starts a search to place the entry sorted */
    node = mmu_hash_table[hash_index];
    while ((node->logical < logical) && (node->next != NULL)) {
      node = node->next;
    } /* while */

        /* if it is not a replacement insert a new pair */
    if (node->logical > logical) {
      ref = (MMU_ENTRY_TYPE *) malloc(sizeof(MMU_ENTRY_TYPE));
      ref->next = node->next;
      ref->logical = node->logical;
      ref->physical = node->physical;
      node->next = ref;
    } else if (node->logical < logical) {
      node->next = (MMU_ENTRY_TYPE *) malloc(sizeof(MMU_ENTRY_TYPE));
      node = node->next;
      node->next = NULL;
    } /* if */
  } /* if */

  node->logical = logical;
  node->physical = physical;

} /* mmu_hash_add */




/*                                                                      */
/* mmu_hash_get_elem gives a pointer to the element containing the      */
/* logical address passed in parameter. If found, another pointer to    */
/* its predecessor in the chain list is given, or if it's the first     */
/* one, the predecessor points to the same element. If not found, the   */
/* pointer to the entry is NULL.                                        */
/*                                                                      */

void mmu_hash_get_elem(
  SS_ADDR_TYPE logical_addr,
  MMU_ENTRY_TYPE **node,
  MMU_ENTRY_TYPE **prev,
  MMU_ENTRY_TYPE **mmu_hash_table)
{
  unsigned int hash_index;
  MMU_ENTRY_TYPE *l_node, *l_previous;

  hash_index = MMU_HASH_KEY(logical_addr);
  l_node = mmu_hash_table[hash_index];
  l_previous = l_node;
  while ((l_node != NULL) && (l_node->logical < logical_addr)) {
    l_previous = l_node;
    l_node = l_node->next;
  } /* while */

  /* Didn't we found the searched node ? */
  if (!((l_node != NULL) && (l_node->logical == logical_addr)))
    l_node = NULL;

  *node = l_node;
  *prev = l_previous;

} /* mmu_hash_get_elem */



/*                                                                      */
/* mmu_hash_remove suppress from the hash list the specified entry      */
/* if it exists.                                                        */
/*                                                                      */
/* WARNING: Just invalidation with logical_address index is supported.  */
/*                                                                      */

void mmu_hash_remove(
  unsigned int command,                       
  SS_ADDR_TYPE logical_addr,
  SS_ADDR_TYPE physical_addr,
  MMU_ENTRY_TYPE **mmu_hash_table)
{
  MMU_ENTRY_TYPE *node, *previous;
  unsigned int hash_index;

  switch (command) {
  case MMU_INV_F_LOGICAL:
    mmu_hash_get_elem(logical_addr, &node, &previous, mmu_hash_table);
    if (node != NULL) {
      if (previous == node) {
          hash_index = MMU_HASH_KEY(logical_addr);
          mmu_hash_table[hash_index] = mmu_hash_table[hash_index]->next;
      } /* if */
      free(node);
    } /* if */
    break;
  case MMU_INV_F_PHYSICAL:
    /* Should I really implement that ? */
    break;
  case MMU_INV_F_CONTEXT:
    for(hash_index = 0; hash_index < MMU_HASH_SIZE; hash_index++) {
      node = mmu_hash_table[hash_index];
      while (node != NULL) {
        previous = node;
        node = node->next;
        free(previous);
      } /* while */
      mmu_hash_table[hash_index] = NULL;
    } /* while */
    break;
  default:
  } /* switch */
} /* mmu_hash_remove */




/* m m u _ r e g _ s t a t s                                            */
/*                                                                      */
/* Not implemented yet.                                                 */
/*                                                                      */
/*                                                                      */

/* void mmu_reg_stats(struct stat_sdb_t *sdb) */





/* m m u _ d u m p                                                      */
/*                                                                      */
/* For test purposes only.                                              */
/*                                                                      */
/*                                                                      */

void mmu_dump(MMU_DB_TYPE *mmu_db, FILE *stream)
{
  unsigned int   hash_index, context, found;
  MMU_ENTRY_TYPE *mmu_node;

  for(context = 0; context < SS_NUM_CONTEXT; context++){
    for(hash_index = 0; hash_index < MMU_HASH_SIZE; hash_index++) {
      found = FALSE;
      if (mmu_db->thread_context[context] != NULL) {
        mmu_node = mmu_db->thread_context[context][hash_index];
        if (mmu_node != NULL) {
          fprintf(stream, "[%d,%02d] ", context, hash_index);
          found = TRUE;
          while (mmu_node != NULL) {
            fprintf(stream, "0x%08x/0x%08x ",
                    mmu_node->logical, 
                    mmu_node->physical);
            mmu_node = mmu_node->next;
          } /* while */
        } /* if */
      } else {
        fatal("mmu_db structure not initialized\n");
      } /* if */
      if (found)
        fprintf(stream, "\n");
    } /* for */
  } /* for */
} /* mmu_dump */





/* m m u _ i n i t                                                      */
/*                                                                      */
/* Initialize the Memory Managment Unit.                                */
/*                                                                      */
/* Returns a pointer to the mmu.                                        */
/*                                                                      */

MMU_DB_TYPE * mmu_init(void)
{
  MMU_DB_TYPE      *mmu_db;
  SS_CONTEXT_TYPE  context;

  mmu_db = (MMU_DB_TYPE *) malloc(sizeof(MMU_DB_TYPE));

  if (mmu_db == NULL) {
    fatal("cannot allocate memory for MMU\n");
  } else {
    for (context = 0; context < SS_NUM_CONTEXT; context++) {
      mmu_db->thread_context[context] = mmu_hash_init();
      mmu_db->config[context] = context;
    } /* for */

  } /* if */

  return mmu_db;

} /* mmu_init */




/* m m u _ f r e e                                                      */
/*                                                                      */
/* Destroy the mmu and frees the allocated memory.                      */
/*                                                                      */
/* Params:                                                              */
/*                                                                      */
/*   in: pointer to a *mmu.                                             */
/*                                                                      */

void mmu_free(MMU_DB_TYPE *mmu_db)
{
  SS_CONTEXT_TYPE context;

  for (context = 0; context < SS_NUM_CONTEXT; context++) {
    mmu_hash_free(mmu_db->thread_context[context]);
  } /* for */

  free(mmu_db);
} /* mmu_free */





/*                                                                      */
/* m m u _ s e t _ c o n t e x t _ c o n f i g                          */
/*                                                                      */
/* Configure the mmu internal context registers                         */
/*                                                                      */
/* Params:                                                              */
/*                                                                      */
/*   in: thread context.                                                */
/*   in: the new config.                                                */
/*   inout: the mmu.                                                    */
/*                                                                      */





void mmu_set_context_config(
  SS_CONTEXT_TYPE context,
  unsigned int config,
  MMU_DB_TYPE *mmu_db)
{
  if ((config < SS_NUM_CONTEXT) && (config >= 0))
    mmu_db->config[context] = config;
  else
    fatal("wrong mmu context configuration");

} /* mmu_set_context_config */






/*                                                                      */
/* m m u _ g e t _ c o n t e x t _ c o n f i g                          */
/*                                                                      */
/* Get the value of a context registers configuration                   */
/*                                                                      */
/* Params:                                                              */
/*                                                                      */
/*   in: thread context.                                                */
/*   inout: the mmu.                                                    */
/*                                                                      */
/*   returned: the config.                                              */
/*                                                                      */

unsigned int mmu_get_context_config(
   SS_CONTEXT_TYPE context,
   MMU_DB_TYPE *mmu_db)
{
  return mmu_db->config[context];
} /* mmu_get_context_config */





/* m m u _ m a p                                                        */
/*                                                                      */
/* Returns the physical address corresponding to the logical given.     */
/* If the page is not in the MMU, this function returns 0 else 1.       */
/*                                                                      */
/* Params:                                                              */
/*                                                                      */
/*   in: thread context.                                                */
/*   in: a logical address.                                             */
/*   inout: the mmu.                                                    */
/*                                                                      */

SS_ADDR_TYPE mmu_map(
  SS_CONTEXT_TYPE context,
  SS_ADDR_TYPE logical_addr,
  MMU_DB_TYPE *mmu_db)
{
  MMU_ENTRY_TYPE *previous, *current;
  SS_ADDR_TYPE physical_addr;

  if ((context >= SS_NUM_CONTEXT) || (context <0))
    fatal("context number out of range in mmu_map\n");

  /* DY:  alias all ASIDs between SS_SHARED_BASE and SS_TEXT_BASE */
  if ((logical_addr >= SS_SHARED_BASE) && (logical_addr < SS_TEXT_BASE))
    context = 0;

  mmu_hash_get_elem(logical_addr & ~(MEM_BLOCK_SIZE - 1),
                    &current, &previous,
                    mmu_db->thread_context[mmu_db->config[context]]);

  if (current != NULL) {
    physical_addr = current->physical + MEM_OFFSET(logical_addr);
/*     fprintf(stderr, "[%d] 0x%08x translated to physical 0x%08x\n", */
/*             context, logical_addr, physical_addr); */
    return physical_addr;
  } else
    return 0;
} /* mmu_map */




/* m m u _ e x i s t s _ p a g e                                        */
/*                                                                      */
/* Returns TRUE if the specified exists in the MMU else FLASE.          */
/*                                                                      */
/* Params:                                                              */
/*                                                                      */
/*   in: thread context.                                                */
/*   in: a logical address.                                             */
/*   inout: the mmu.                                                    */
/*                                                                      */

unsigned int mmu_exists_page(
  SS_CONTEXT_TYPE context,
  SS_ADDR_TYPE logical_addr,
  MMU_DB_TYPE *mmu_db)
{
  MMU_ENTRY_TYPE *previous, *current;

  if ((context >= SS_NUM_CONTEXT) || (context <0))
    fatal("context number out of range in mmu_exists_page\n");

  /* DY:  alias all ASIDs between SS_SHARED_BASE and SS_TEXT_BASE */
  if ((logical_addr >= SS_SHARED_BASE) && (logical_addr < SS_TEXT_BASE))
    context = 0;

  mmu_hash_get_elem(logical_addr & ~(MEM_BLOCK_SIZE - 1),
                    &current, &previous,
                    mmu_db->thread_context[mmu_db->config[context]]);

  if (current != NULL) {
    return TRUE;
  } else
    return FALSE;
} /* mmu_exists_entry */




/* m m u _ i s _ m a p p e d _ b y                                      */
/*                                                                      */
/* Returns the logical address corresponding to the physical given.     */
/* If the logical page is found, 1 is returned else 0.                  */
/*                                                                      */
/* Params:                                                              */
/*                                                                      */
/*   in: thread context.                                                */
/*   in: a physical address.                                            */
/*   inout: the mmu.                                                    */
/*                                                                      */

SS_ADDR_TYPE mmu_is_mapped_by(
  SS_CONTEXT_TYPE context,
  SS_ADDR_TYPE physical_addr,
  MMU_DB_TYPE *mmu_db)
{
  /* Should I implement that ? */
  fatal("mmu_is_mapped_by is not implemented\n");

  return 0;  
} /* mmu_is_mapped_by */



/* m m u _ a d d _ e n t r y                                            */
/*                                                                      */
/* Add the couple (logical address, physical address) to the            */
/* memory manager unit.                                                 */
/*                                                                      */
/* Params:                                                              */
/*                                                                      */
/*   in: logical address.                                               */
/*   in: physical address.                                              */
/*   inout: a pointer to the mmu.                                       */
/*                                                                      */

int mmu_add_entry(
  SS_CONTEXT_TYPE context,
  SS_ADDR_TYPE logical_addr,
  SS_ADDR_TYPE physical_addr,
  MMU_DB_TYPE *mmu_db)
{
  if ((context >= SS_NUM_CONTEXT) || (context <0))
    fatal("context number out of range in mmu_add_entry\n");

  /* DY:  alias all ASIDs between SS_SHARED_BASE and SS_TEXT_BASE */
  if ((logical_addr >= SS_SHARED_BASE) && (logical_addr < SS_TEXT_BASE))
    context = 0;

  mmu_hash_add(context,
               logical_addr & ~(MEM_BLOCK_SIZE - 1),
               physical_addr,
               mmu_db->thread_context[mmu_db->config[context]]);

  return 1;
} /* mmu_add_entry */




/*                                                                      */
/* m m u _ i n v a l i d a t e  _ e n t r y                             */
/*                                                                      */
/* Suppress the entry given. Depending on the first parameter, it is    */
/* possible to invalidate a full context, a specific entry or a         */
/* combination of them. The target is specified with MMU_INV_*          */
/* constants. The unused parameter can be just NULL.                    */
/*                                                                      */
/* Params:                                                              */
/*                                                                      */
/*   in: invalidation type                                              */
/*   in: thread context.                                                */
/*   in: logical address.                                               */
/*   in: physical address.                                              */
/*   inout: the mmu.                                                    */
/*                                                                      */

int mmu_invalidate_entry(
  unsigned int command,
  SS_CONTEXT_TYPE context,
  SS_ADDR_TYPE logical_addr,
  SS_ADDR_TYPE physical_addr,
  MMU_DB_TYPE *mmu_db)
{
  if ((command != MMU_INV_F_LOGICAL) && (command != MMU_INV_F_CONTEXT))
    fatal("MMU_INV_F_PHYSICAL command not implemented in mmu_invalidate_entry\n");
  if ((context >= SS_NUM_CONTEXT) || (context < 0))
    fatal("Invalide context number in mmu_invalidate_entry\n");

  /* DY:  alias all ASIDs between SS_SHARED_BASE and SS_TEXT_BASE */
  if ((logical_addr >= SS_SHARED_BASE) && (logical_addr < SS_TEXT_BASE))
    context = 0;

  mmu_hash_remove(command,
                  logical_addr & ~(MEM_BLOCK_SIZE - 1),
                  physical_addr,
                  mmu_db->thread_context[mmu_db->config[context]]);

  return 1;
} /* mmu_invalidate_entry */
