/*
 * sim-ssmt.c - simultaneous multithreading simulator implementation
 *
 * This file is part of my Ph. D. thesis work
 * 
 * Copyright (C) 1998, 1999, 2000 by Dominik Madon
 * Copyright (C) 1994, 1995, 1996, 1997 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 *
 * $Id: sim-ssmt.c,v 1.20 2002/08/19 15:31:16 dongkeun Exp dongkeun $
 *
 * Revision 1.1  1996/12/05  18:52:32  taustin
 * Initial revision
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <signal.h>

#include "misc.h"
#include "ss.h"
#include "regs.h"
#include "mmu.h"
#include "memory.h"
#include "cache.h"
#include "loader.h"
#include "syscall.h"
#include "bpred.h"
#include "resource.h"
#include "bitmap.h"
#include "options.h"
#include "stats.h"
#include "ptrace.h"
#include "dlite.h"
#include "sim.h"
#include "context.h"
#include "interrupt.h"

/* added by dongkeun */
#include "vortex.h"
void sim_vsim(void);
static int ruu_table[SS_NUM_CONTEXT][256], lsq_table[SS_NUM_CONTEXT][128];
static int sim_num_insn_commit[SS_NUM_CONTEXT];
int vsim_mode = FALSE;

#ifdef CACHE_PROF
int px_mode = FALSE;
#endif

static int exception_array[10];
int exception_signal = 0;

/* 
 * 1 = bad size or alignment - mem_valid in memory.c
 * 2 = segmentation violation - mem_valid in memory.c 
 * 3 = bad size or alignment - mem_access in memory.c
 * 4 = segmentation violation - mem_access in memory.c
 * 5 = bad jump alignment - TALIGN in ss.def
 */   

/* DK: This is to keep track of how many kills were actually performed. */
#define MAX_KILL_ENTRY	100

typedef struct kill_info
{
  SS_ADDR_TYPE PC;
  int access;
  int kill;
} kill_info_t;

kill_info_t kill_table[MAX_KILL_ENTRY];
int kill_table_num = 0;

/* DK: When a pre-execution thread is killed due to an exception or a kill
   instruction, set the proper element of this array. After all the 
   instructions that belong to the killed context are squashed in the 
   ruu_recover(), reset it and resume the context. */
int kill_mode[SS_NUM_CONTEXT];

/*
 * This file implements a very detailed out-of-order issue superscalar
 * processor with a two-level memory system and speculative execution support.
 * This simulator is a performance simulator, tracking the latency of all
 * pipeline operations.
 */

/*
 * DM debug variables
 */

/*  static int DM_debug_tag = 0; */

/*
 * simulator options
 */

/* DK: maximum number of inst's to execute */
static unsigned int max_insts;

/* DK: number of insts skipped before timing starts */
static int fastfwd_count;

/* DK: offset value to be added sub-threads' icount */
static int main_offset;

/* DK: offset value to be added sub-threads' icount */
static int debug_cycle;

/* DK: TRUE: perfect cache, FALSE: real cache */
static int perfi, perfd;

/* DK: number of threads to fetch from at a cycle */ 
static int num_thread;

/* pipeline trace range and output filename */
static int ptrace_nelt = 0;
static char *ptrace_opts[2];

/* instruction fetch queue size (in insts) */
static int ruu_ifq_size;

/* extra branch mis-prediction latency */
static int ruu_branch_penalty;

/* speed of front-end of machine relative to execution core */
static int fetch_speed;

/* instructions' block size to fetch at one time for one context */
static int fetch_block_size;

/* number of waiting cycles before exiting when there is no active context */
static int fetch_waiting;

/* number of cacheprobe executed per cycle by the prefetch stage */
static int prefetch_cache_probe;

/* branch predictor type {nottaken|taken|perfect|bimod|2lev} */
static char *pred_type;

/* bimodal predictor config (<table_size>) */
static int bimod_nelt = 1;
static int bimod_config[1] =
  { /* bimod tbl size */2048 };

/* 2-level predictor config (<l1size> <l2size> <hist_size> <xor>) */
static int twolev_nelt = 4;
static int twolev_config[4] =
  { /* l1size */1, /* l2size */2048, /* hist */11, /* xor */TRUE};

/* combining predictor config (<meta_table_size> */
static int comb_nelt = 1;
static int comb_config[1] =
  { /* meta_table_size */1024 };

/* return address stack (RAS) size */
static int ras_size = 8;

/* BTB predictor config (<num_sets> <associativity>) */
static int btb_nelt = 2;
static int btb_config[2] =
  { /* nsets */512, /* assoc */4 };

/* instruction decode B/W (insts/cycle) */
static int ruu_decode_width;

/* instruction issue B/W (insts/cycle) */
static int ruu_issue_width;

/* run pipeline with in-order issue */
static int ruu_inorder_issue;

/* issue instructions down wrong execution paths */
static int ruu_include_spec = TRUE;

/* instruction commit B/W (insts/cycle) */
static int ruu_commit_width;

/* register update unit (RUU) size */
static int RUU_size = 8;

/* load/store queue (LSQ) size */
static int LSQ_size = 4;

/* l1 data cache config, i.e., {<config>|none} */
static char *cache_dl1_opt;

/* l1 data cache hit latency (in cycles) */
static int cache_dl1_lat;

/* l2 data cache config, i.e., {<config>|none} */
static char *cache_dl2_opt;

/* l2 data cache hit latency (in cycles) */
static int cache_dl2_lat;

/* l1 instruction cache config, i.e., {<config>|dl1|dl2|none} */
static char *cache_il1_opt;

/* l1 instruction cache hit latency (in cycles) */
static int cache_il1_lat;

/* l2 instruction cache config, i.e., {<config>|dl1|dl2|none} */
static char *cache_il2_opt;

/* l2 instruction cache hit latency (in cycles) */
static int cache_il2_lat;

/* flush caches on system calls */
static int flush_on_syscalls;

/* convert 64-bit inst addresses to 32-bit inst equivalents */
static int compress_icache_addrs;

/* shared il1 cache between context */
static int shared_il1 = TRUE;

/* shared il2 cache between context */
static int shared_il2 = TRUE;

/* shared dl1 cache between context */
static int shared_dl1 = TRUE;

/* shared dl2 cache between context */
static int shared_dl2 = TRUE;

/* shared itlb cache between context */
static int shared_itlb = TRUE;

/* shared il1 cache between context */
static int shared_dtlb = TRUE;

/* memory access latency (<first_chunk> <inter_chunk>) */
static int mem_nelt = 2;
static int mem_lat[2] =
  { /* lat to first chunk */80, /* lat between remaining chunks */6 };

/* memory access bus width (in bytes) */
static int mem_bus_width;

/* instruction TLB config, i.e., {<config>|none} */
static char *itlb_opt;

/* data TLB config, i.e., {<config>|none} */
static char *dtlb_opt;

/* inst/data TLB miss latency (in cycles) */
static int tlb_miss_lat;

/* total number of integer ALU's available */
static int res_ialu;

/* total number of integer multiplier/dividers available */
static int res_imult;

/* total number of memory system ports available (to CPU) */
static int res_memport;

/* total number of floating point ALU's available */
static int res_fpalu;

/* total number of floating point multiplier/dividers available */
static int res_fpmult;

/* total number of context managment unit available */
static int res_cxtmsg;

/* text-based stat profiles */
#define MAX_PCSTAT_VARS 8
static int pcstat_nelt = 0;
static char *pcstat_vars[MAX_PCSTAT_VARS];

/* external interruption configuration  */
static int interrupt_nelt = 3;
static unsigned int interrupt_config[3] =
{ /* rise at cycle */0, /* port */256, /* then every  */0 /* cycles */};

/* interruption structure db (see interrupt.h) )*/
static INTERRUPT_TYPE *interrupt_db;

/* next cycle where an interruption should occured. Off if 0 */
static unsigned long int next_interruption = 0;

/* value of the PC of the next instrcution to execute when the interrupt
   occurs */
static SS_ADDR_TYPE interrupted_PC = 0;

/* interrupt vector (address of the interrupt handler) */
static SS_ADDR_TYPE interrupt_vector = 0;

/* an interrupt has been raised */
static unsigned int interrupt_raised = FALSE;

/* operate in backward-compatible bugs mode (for testing only) */
static int bugcompat_mode;

/* DY: flag for ignoring memory access violations in sub-threads */
int ignore_sub_thread_seg;

/* DY: flag for printing on cache miss events during profiling */
int print_cache_miss_events;

/*
 * functional unit resource configuration
 */

/* resource pool indices, NOTE: update these if you change FU_CONFIG */
#define FU_IALU_INDEX			0
#define FU_IMULT_INDEX			1
#define FU_MEMPORT_INDEX		2
#define FU_FPALU_INDEX			3
#define FU_FPMULT_INDEX			4
#define FU_CXTMSG_INDEX                 5

/* resource pool definition, NOTE: update FU_*_INDEX defs if you change this */
struct res_desc fu_config[] = {
  {
    "integer-ALU",
    8,
    0,
    {
      { IntALU, 1, 1 }
    }
  },
  {
    "integer-MULT/DIV",
    2,
    0,
    {
      { IntMULT, 3, 1 },
      { IntDIV, 20, 19 }
    }
  },
  {
    "memory-port",
    8,
    0,
    {
      { RdPort, 1, 1 },
      { WrPort, 1, 1 }
    }
  },
  {
    "FP-adder",
    8,
    0,
    {
      { FloatADD, 2, 1 },
      { FloatCMP, 2, 1 },
      { FloatCVT, 2, 1 }
    }
  },
  {
    "FP-MULT/DIV",
    2,
    0,
    {
      { FloatMULT, 4, 1 },
      { FloatDIV, 12, 12 },
      { FloatSQRT, 24, 24 }
    }
  },
  {
    "Context/Msg",
    1,
    0,
    {
      { CxtMsg, 1, 1 },
      { CxtReg, 1, 1 },
      { CxtLock, 2, 2 }
    }
  },
};


/*
 * simulator stats
 */

/* total number of instructions committed */
static SS_COUNTER_TYPE sim_num_insn = 0;

/* total number of instructions committed for contexts */
static SS_COUNTER_TYPE sim_num_insn_ctxt[SS_NUM_CONTEXT];

/* total number of instructions committed for contexts */
static SS_COUNTER_TYPE sim_num_start[SS_NUM_CONTEXT];

/* total number of instructions executed */
static SS_COUNTER_TYPE sim_total_insn = 0;

/* total number of memory references committed */
static SS_COUNTER_TYPE sim_num_refs = 0;

/* total number of memory references executed */
static SS_COUNTER_TYPE sim_total_refs = 0;

/* total number of loads committed */
static SS_COUNTER_TYPE sim_num_loads = 0;

/* total number of loads executed */
static SS_COUNTER_TYPE sim_total_loads = 0;

/* total number of branches committed */
static SS_COUNTER_TYPE sim_num_branches = 0;

/* total number of branches executed */
static SS_COUNTER_TYPE sim_total_branches = 0;

/* cycle counter */
SS_TIME_TYPE sim_cycle = 0;

/* DY:  hardware semaphore stats */
static int sem_p_ops[SS_NUM_CONTEXT][16];
static int sem_v_ops[SS_NUM_CONTEXT][16];
static int sem_suspends[SS_NUM_CONTEXT][16];
static unsigned prev_sem_time[SS_NUM_CONTEXT][16];
static unsigned total_sleep_time[SS_NUM_CONTEXT][16]; 

/* GT: flag to check whether predictor is shared */
int bpred_shared; 

/*
 * simulator state variables
 */


/* instruction sequence counter, used to assign unique id's to insts */
static unsigned int inst_seq = 0;

/* pipetrace instruction sequence counter */
static unsigned int ptrace_seq[SS_NUM_CONTEXT];

/* speculation mode, non-zero when mis-speculating, i.e., executing
   instructions down the wrong path, thus state recovery will eventually have
   to occur that resets processor register and memory state back to the last
   precise state */
int spec_mode[SS_NUM_CONTEXT];

/* cycles until fetch issue resumes */
static unsigned ruu_fetch_issue_delay[SS_NUM_CONTEXT];
/* DY:  prevent icache thrashing */
static unsigned icache_thrash_detect[SS_NUM_CONTEXT];

/* cycles until fetch unit will be able to start a new memory fetch */
static unsigned ruu_fetch_issue_ready = 0;

/* fetch priority for next fetch */
static unsigned int fetch_dynamic_priority[SS_NUM_CONTEXT];
#define FETCH_CXT_CACHE_MISS    0x00000100
#define FETCH_CXT_IDLE          0x00000200
#define FETCH_CXT_USED          0x00000400

/* perfect prediction enabled */
static int pred_perfect = FALSE;

/* speculative bpred-update enabled */
static char *bpred_spec_opt;
static enum { spec_ID, spec_WB, spec_CT } bpred_spec_update;

/* level 1 instruction cache, entry level instruction cache */
static struct cache *cache_il1[SS_NUM_CONTEXT];

/* level 1 instruction cache */
static struct cache *cache_il2[SS_NUM_CONTEXT];

/* level 1 data cache, entry level data cache */
static struct cache *cache_dl1[SS_NUM_CONTEXT];

/* level 2 data cache */
static struct cache *cache_dl2[SS_NUM_CONTEXT];

/* instruction TLB */
static struct cache *itlb[SS_NUM_CONTEXT];

/* data TLB */
static struct cache *dtlb[SS_NUM_CONTEXT];

/* branch predictor */
static struct bpred *pred[SS_NUM_CONTEXT];

/* GT: return of stack */
static struct bpred *retstack[SS_NUM_CONTEXT];

/* functional unit resource pool */
static struct res_pool *fu_pool = NULL;

/* prefetch cache accessor. If NULL prefetch is disabled  */
static struct cache *ruu_prefetch_icache[SS_NUM_CONTEXT];

/* text-based stat profiles */
static struct stat_stat_t *pcstat_stats[MAX_PCSTAT_VARS];
static SS_COUNTER_TYPE pcstat_lastvals[MAX_PCSTAT_VARS];
static struct stat_stat_t *pcstat_sdists[MAX_PCSTAT_VARS];

/* current context */
static SS_CONTEXT_TYPE current_context = 0;

/* simulator's active context. Because instructions are executed in
   the dispatch stage, the simulator must know if instructions placed
   in the reservation stations have been executed or not */
static BITMAP_TYPE(SS_NUM_CONTEXT, sim_active_context);
#define SIM_ACTIVE_SZ (BITMAP_SIZE(SS_NUM_CONTEXT))

/* Context's current phase. If a stop occured all instructions
   in the fetch queue will not be executed in the decode/dispatch
   stage but marked with the idle flag. Nevertheless they will be
   dispatched to the functional unit (consuming time) */
static BITMAP_TYPE(SS_NUM_CONTEXT, sim_shutdown_context);
#define SIM_SHUTDOWN_SZ (BITMAP_SIZE(SS_NUM_CONTEXT))

/* Context's restart is set when a context is being shutdown and
   another context wants to restart it. The variable is clear
   when the context has flushed the fetch queue (and thus
   sim_context_active is set) or has been stopped once again. */
static BITMAP_TYPE(SS_NUM_CONTEXT, sim_restart_context);
#define SIM_RESTART_SZ (BITMAP_SIZE(SS_NUM_CONTEXT))

/* wedge all stat values into a SS_COUNTER_TYPE */
#define STATVAL(STAT)							\
  ((STAT)->sc == sc_int							\
   ? (SS_COUNTER_TYPE)*((STAT)->variant.for_int.var)			\
   : ((STAT)->sc == sc_uint						\
      ? (SS_COUNTER_TYPE)*((STAT)->variant.for_uint.var)		\
      : ((STAT)->sc == sc_counter					\
	 ? *((STAT)->variant.for_counter.var)				\
	 : (panic("bad stat class"), 0))))


/* memory access latency, assumed to not cross a page boundary */
static unsigned int			/* total latency of access */
mem_access_latency(int blk_sz)		/* block size accessed */
{
  int chunks = (blk_sz + (mem_bus_width - 1)) / mem_bus_width;

  assert(chunks > 0);

  return (/* first chunk latency */mem_lat[0] +
	  (/* remainder chunk latency */mem_lat[1] * (chunks - 1)));
}


/*
 * cache miss handlers
 */

/* l1 data cache l1 block miss handler function */
static unsigned int			/* latency of block access */
dl1_access_fn(enum mem_cmd cmd,		/* access cmd, Read or Write */
	      SS_ADDR_TYPE baddr,	/* block address to access */
	      unsigned int ASID,        /* DY:  address space to access */
	      int bsize,		/* size of block to access */
	      struct cache_blk *blk,	/* ptr to block in upper level */
	      SS_TIME_TYPE now)		/* time of access */
{
  unsigned int lat;

  if (cache_dl2[current_context])
    {
      /* access next level of data cache hierarchy */
      lat = cache_access(cache_dl2[current_context], cmd, baddr, ASID, 
			 NULL, bsize,
			 /* now */now, /* pudata */NULL, /* repl addr */NULL);

      if (perfd) /* DK: perfect D-cache */
	lat = 1;	

      if (cmd == Read)
	return lat;
      else
	{
	  /* FIXME: unlimited write buffers */
	  return 0;
	}
    }
  else
    {
      /* access main memory */
      if (cmd == Read) {
	if (perfd)
	  return 1;
	else
	  return mem_access_latency(bsize);
      }
      else
	{
	  /* FIXME: unlimited write buffers */
	  return 0;
	}
    }
}

/* l2 data cache block miss handler function */
static unsigned int			/* latency of block access */
dl2_access_fn(enum mem_cmd cmd,		/* access cmd, Read or Write */
	      SS_ADDR_TYPE baddr,	/* block address to access */
	      unsigned int ASID,        /* DY:  address space to access */
	      int bsize,		/* size of block to access */
	      struct cache_blk *blk,	/* ptr to block in upper level */
	      SS_TIME_TYPE now)		/* time of access */
{
  /* this is a miss to the lowest level, so access main memory */
  if (cmd == Read) {
    if (perfd)
      return 1;
    else 
      return mem_access_latency(bsize);
  }
  else
    {
      /* FIXME: unlimited write buffers */
      return 0;
    }
}

/* l1 inst cache l1 block miss handler function */
static unsigned int			/* latency of block access */
il1_access_fn(enum mem_cmd cmd,		/* access cmd, Read or Write */
	      SS_ADDR_TYPE baddr,	/* block address to access */
	      unsigned int ASID,        /* DY:  address space to access */
	      int bsize,		/* size of block to access */
	      struct cache_blk *blk,	/* ptr to block in upper level */
	      SS_TIME_TYPE now)		/* time of access */
{
  unsigned int lat;

if (cache_il2[current_context])
    {
      /* access next level of inst cache hierarchy */
      lat = cache_access(cache_il2[current_context], cmd, baddr, ASID,
			 NULL, bsize,
			 /* now */now, /* pudata */NULL, /* repl addr */NULL);
      if (perfi) /* DK: perfect I-cache */
	lat = 1;	

      if (cmd == Read)
	return lat;
      else
	panic("writes to instruction memory not supported");
    }
  else
    {
      /* access main memory */
      if (cmd == Read) {
	if (perfi)
	  return 1;
	else 
	  return mem_access_latency(bsize);
      }	
      else
	panic("writes to instruction memory not supported");
    }
}

/* l2 inst cache block miss handler function */
static unsigned int			/* latency of block access */
il2_access_fn(enum mem_cmd cmd,		/* access cmd, Read or Write */
	      SS_ADDR_TYPE baddr,	/* block address to access */
	      unsigned int ASID,        /* DY:  address space to access */
	      int bsize,		/* size of block to access */
	      struct cache_blk *blk,	/* ptr to block in upper level */
	      SS_TIME_TYPE now)		/* time of access */
{
  /* this is a miss to the lowest level, so access main memory */
  if (cmd == Read) {
    if (perfi)
      return 1;
    else 
      return mem_access_latency(bsize);
  }
  else
    panic("writes to instruction memory not supported");
}


/*
 * TLB miss handlers
 */

/* inst cache block miss handler function */
static unsigned int			/* latency of block access */
itlb_access_fn(enum mem_cmd cmd,	/* access cmd, Read or Write */
	       SS_ADDR_TYPE baddr,	/* block address to access */
	       unsigned int ASID,       /* DY:  address space to access */
	       int bsize,		/* size of block to access */
	       struct cache_blk *blk,	/* ptr to block in upper level */
	       SS_TIME_TYPE now)	/* time of access */
{
  SS_ADDR_TYPE *phy_page_ptr = (SS_ADDR_TYPE *)blk->user_data;

  /* no real memory access, however, should have user data space attached */
  assert(phy_page_ptr);

  /* fake translation, for now... */
  *phy_page_ptr = 0;

  /* return tlb miss latency */
  return tlb_miss_lat;
}

/* data cache block miss handler function */
static unsigned int			/* latency of block access */
dtlb_access_fn(enum mem_cmd cmd,	/* access cmd, Read or Write */
	       SS_ADDR_TYPE baddr,	/* block address to access */
	       unsigned int ASID,       /* DY:  address space to access */
	       int bsize,		/* size of block to access */
	       struct cache_blk *blk,	/* ptr to block in upper level */
	       SS_TIME_TYPE now)	/* time of access */
{
  SS_ADDR_TYPE *phy_page_ptr = (SS_ADDR_TYPE *)blk->user_data;

  /* no real memory access, however, should have user data space attached */
  assert(phy_page_ptr);

  /* fake translation, for now... */
  *phy_page_ptr = 0;

  /* return tlb miss latency */
  return tlb_miss_lat;
}


/* register simulator-specific options */
void
sim_reg_options(struct opt_odb_t *odb)
{
  opt_reg_header(odb, 
"sim-ssmt: This simulator implements a very detailed out-of-order issue\n"
"superscalar processor with a two-level memory system and speculative\n"
"execution support. This simulator is able to execute multiple contexts\n"
"simultaneously i.e. dispatch instructions provided by multiple contexts\n"
"to functional units. This simulator is a performance simulator, tracking\n"
"the latency of all pipeline operations.\n"
		 );

  /* DK: instruction limit */

  opt_reg_uint(odb, "-max:inst", "maximum number of instructions to execute",
	       &max_insts, /* default */0,
	       /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-fastfwd", "number of insts skipped before timing starts",
	      &fastfwd_count, /* default */0,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-offset", "offset value to be added to sub-threads' icount",
	      &main_offset, /* default */0,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-debug", "turn on debugging feature after this many cycles",
	      &debug_cycle, /* default */0,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-num:thread", "number of threads to fetch from at a cycle",
	      &num_thread, /* default */4,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-igsubseg", "ignore memory access violations from sub-threads",
	      &ignore_sub_thread_seg, /* default */0,
	      /* print */TRUE, /* format */NULL);

  opt_reg_flag(odb, "-printmiss", "print cache miss events",
	       &print_cache_miss_events, /* default */FALSE,
	       /* print */TRUE, /* format */NULL);

  /* trace options */

  opt_reg_string_list(odb, "-ptrace",
	      "generate pipetrace, i.e., <fname|stdout|stderr> <range>",
	      ptrace_opts, /* arr_sz */2, &ptrace_nelt, /* default */NULL,
	      /* !print */FALSE, /* format */NULL, /* !accrue */FALSE);

  opt_reg_note(odb,
"  Pipetrace range arguments are formatted as follows:\n"
"\n"
"    {{@|#}<start>}:{{@|#|+}<end>}\n"
"\n"
"  Both ends of the range are optional, if neither are specified, the entire\n"
"  execution is traced.  Ranges that start with a `@' designate an address\n"
"  range to be traced, those that start with an `#' designate a cycle count\n"
"  range.  All other range values represent an instruction count range.  The\n"
"  second argument, if specified with a `+', indicates a value relative\n"
"  to the first argument, e.g., 1000:+100 == 1000:1100.  Program symbols may\n"
"  be used in all contexts.\n"
"\n"
"    Examples:   -ptrace FOO.trc #0:#1000\n"
"                -ptrace BAR.trc @2000:\n"
"                -ptrace BLAH.trc :1500\n"
"                -ptrace UXXE.trc :\n"
"                -ptrace FOOBAR.trc @main:+278\n"
	       );

  /* ifetch options */

  opt_reg_int(odb, "-fetch:ifqsize", "instruction fetch queue size (in insts)",
	      &ruu_ifq_size, /* default */32,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-fetch:mplat", "extra branch mis-prediction latency",
	      &ruu_branch_penalty, /* default */3,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-fetch:speed",
	      "speed of front-end of machine relative to execution core",
	      &fetch_speed, /* default */1,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-fetch:blocksize",
	      "instructions block size to fetch for a context",
	      &fetch_block_size, /* default */4,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-prefetch:cacheprobe",
	      "number of cache probe executed by the prefetch stage each cycle",
	      &prefetch_cache_probe, /* default */2,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-fetch:waiting",
	      "Number of waiting cycles to exit when there is no active context",
	      &fetch_waiting, /* default */1000,
	      /* print */TRUE, /* format */NULL);

  /* branch predictor options */

  opt_reg_note(odb,
"  Branch predictor configuration examples for 2-level predictor:\n"
"    Configurations:   N, M, W, X\n"
"      N   # entries in first level (# of shift register(s))\n"
"      W   width of shift register(s)\n"
"      M   # entries in 2nd level (# of counters, or other FSM)\n"
"      X   (yes-1/no-0) xor history and address for 2nd level index\n"
"    Sample predictors:\n"
"      GAg     : 1, W, 2^W, 0\n"
"      GAp     : 1, W, M (M > 2^W), 0\n"
"      PAg     : N, W, 2^W, 0\n"
"      PAp     : N, W, M (M == 2^(N+W)), 0\n"
"      gshare  : 1, W, 2^W, 1\n"
"  Predictor `comb' combines a bimodal and a 2-level predictor.\n"
               );

  opt_reg_int(odb, "-bpred:shared", "shared branch predictor and BTB",
	      &bpred_shared, /* default */FALSE,
	      /* print */TRUE, /* format */NULL);

  opt_reg_string(odb, "-bpred",
		 "branch predictor type {nottaken|taken|perfect|bimod|2lev|comb}",
                 &pred_type, /* default */"2lev",
                 /* print */TRUE, /* format */NULL);

  opt_reg_int_list(odb, "-bpred:bimod",
		   "bimodal predictor config (<table size>)",
		   bimod_config, bimod_nelt, &bimod_nelt,
		   /* default */bimod_config,
		   /* print */TRUE, /* format */NULL, /* !accrue */FALSE);

  opt_reg_int_list(odb, "-bpred:2lev",
                   "2-level predictor config "
		   "(<l1size> <l2size> <hist_size> <xor>)",
                   twolev_config, twolev_nelt, &twolev_nelt,
		   /* default */twolev_config,
                   /* print */TRUE, /* format */NULL, /* !accrue */FALSE);

  opt_reg_int_list(odb, "-bpred:comb",
		   "combining predictor config (<meta_table_size>)",
		   comb_config, comb_nelt, &comb_nelt,
		   /* default */comb_config,
		   /* print */TRUE, /* format */NULL, /* !accrue */FALSE);

  opt_reg_int(odb, "-bpred:ras",
              "return address stack size (0 for no return stack)",
              &ras_size, /* default */ras_size,
              /* print */TRUE, /* format */NULL);

  opt_reg_int_list(odb, "-bpred:btb",
		   "BTB config (<num_sets> <associativity>)",
		   btb_config, btb_nelt, &btb_nelt,
		   /* default */btb_config,
		   /* print */TRUE, /* format */NULL, /* !accrue */FALSE);

  opt_reg_string(odb, "-bpred:spec_update",
		 "speculative predictors update in {ID|WB} (default non-spec)",
		 &bpred_spec_opt, /* default */NULL,
		 /* print */TRUE, /* format */NULL);

  /* decode options */

  opt_reg_int(odb, "-decode:width",
	      "instruction decode B/W (insts/cycle)",
	      &ruu_decode_width, /* default */8,
	      /* print */TRUE, /* format */NULL);

  /* issue options */

  opt_reg_int(odb, "-issue:width",
	      "instruction issue B/W (insts/cycle)",
	      &ruu_issue_width, /* default */8,
	      /* print */TRUE, /* format */NULL);

  opt_reg_flag(odb, "-issue:inorder", "run pipeline with in-order issue",
	       &ruu_inorder_issue, /* default */FALSE,
	       /* print */TRUE, /* format */NULL);

  opt_reg_flag(odb, "-issue:wrongpath",
	       "issue instructions down wrong execution paths",
	       &ruu_include_spec, /* default */TRUE,
	       /* print */TRUE, /* format */NULL);

  /* commit options */

  opt_reg_int(odb, "-commit:width",
	      "instruction commit B/W (insts/cycle)",
	      &ruu_commit_width, /* default */8,
	      /* print */TRUE, /* format */NULL);

  /* register scheduler options */

  opt_reg_int(odb, "-ruu:size",
	      "register update unit (RUU) size",
	      &RUU_size, /* default */128,
	      /* print */TRUE, /* format */NULL);

  /* memory scheduler options  */

  opt_reg_int(odb, "-lsq:size",
	      "load/store queue (LSQ) size",
	      &LSQ_size, /* default */64,
	      /* print */TRUE, /* format */NULL);

  /* cache options */

  opt_reg_string(odb, "-cache:dl1",
		 "l1 data cache config, i.e., {<config>|none}",
		 &cache_dl1_opt, "dl1:1024:32:2:l",
		 /* print */TRUE, NULL);

  opt_reg_note(odb,
"  The cache config parameter <config> has the following format:\n"
"\n"
"    <name>:<nsets>:<bsize>:<assoc>:<repl>\n"
"\n"
"    <name>   - name of the cache being defined\n"
"    <nsets>  - number of sets in the cache\n"
"    <bsize>  - block size of the cache\n"
"    <assoc>  - associativity of the cache\n"
"    <repl>   - block replacement strategy, 'l'-LRU, 'f'-FIFO, 'r'-random\n"
"\n"
"    Examples:   -cache:dl1 dl1:4096:32:1:l\n"
"                -dtlb dtlb:128:4096:32:r\n"
	       );

  opt_reg_int(odb, "-cache:dl1lat",
	      "l1 data cache hit latency (in cycles)",
	      &cache_dl1_lat, /* default */1,
	      /* print */TRUE, /* format */NULL);

  opt_reg_string(odb, "-cache:dl2",
		 "l2 data cache config, i.e., {<config>|none}",
		 &cache_dl2_opt, "ul2:4096:64:4:l",
		 /* print */TRUE, NULL);

  opt_reg_int(odb, "-cache:dl2lat",
	      "l2 data cache hit latency (in cycles)",
	      &cache_dl2_lat, /* default */10,
	      /* print */TRUE, /* format */NULL);

  opt_reg_string(odb, "-cache:il1",
		 "l1 inst cache config, i.e., {<config>|dl1|dl2|none}",
		 &cache_il1_opt, "il1:1024:64:2:l",
		 /* print */TRUE, NULL);

  opt_reg_note(odb,
"  Cache levels can be unified by pointing a level of the instruction cache\n"
"  hierarchy at the data cache hiearchy using the \"dl1\" and \"dl2\" cache\n"
"  configuration arguments.  Most sensible combinations are supported, e.g.,\n"
"\n"
"    A unified l2 cache (il2 is pointed at dl2):\n"
"      -cache:il1 il1:128:64:1:l -cache:il2 dl2\n"
"      -cache:dl1 dl1:256:32:1:l -cache:dl2 ul2:1024:64:2:l\n"
"\n"
"    Or, a fully unified cache hierarchy (il1 pointed at dl1):\n"
"      -cache:il1 dl1\n"
"      -cache:dl1 ul1:256:32:1:l -cache:dl2 ul2:1024:64:2:l\n"
	       );

  opt_reg_int(odb, "-cache:il1lat",
	      "l1 instruction cache hit latency (in cycles)",
	      &cache_il1_lat, /* default */1,
	      /* print */TRUE, /* format */NULL);

  opt_reg_string(odb, "-cache:il2",
		 "l2 instruction cache config, i.e., {<config>|dl2|none}",
		 &cache_il2_opt, "dl2",
		 /* print */TRUE, NULL);

  opt_reg_int(odb, "-cache:il2lat",
	      "l2 instruction cache hit latency (in cycles)",
	      &cache_il2_lat, /* default */10,
	      /* print */TRUE, /* format */NULL);

  opt_reg_flag(odb, "-cache:flush", "flush caches on system calls",
	       &flush_on_syscalls, /* default */FALSE, /* print */TRUE, NULL);

  opt_reg_flag(odb, "-cache:icompress",
	       "convert 64-bit inst addresses to 32-bit inst equivalents",
	       &compress_icache_addrs, /* default */FALSE,
	       /* print */TRUE, NULL);

  opt_reg_flag(odb, "-cache:il1shared", "share il1 cache between contexts",
	       &shared_il1, /* default */TRUE,
	       /* print */TRUE, /* format */NULL);

  opt_reg_flag(odb, "-cache:il2shared", "share il2 cache between contexts",
	       &shared_il2, /* default */TRUE,
	       /* print */TRUE, /* format */NULL);

  opt_reg_flag(odb, "-cache:dl1shared", "share dl1 cache between contexts",
	       &shared_dl1, /* default */TRUE,
	       /* print */TRUE, /* format */NULL);

  opt_reg_flag(odb, "-cache:dl2shared", "share dl2 cache between contexts",
	       &shared_dl2, /* default */TRUE,
	       /* print */TRUE, /* format */NULL);


  /* mem options */
  
  opt_reg_int(odb, "-perf:icache", "perfect I-cache",
	      &perfi, /* default */FALSE,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-perf:dcache", "perfect D-cache",
	      &perfd, /* default */FALSE,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int_list(odb, "-mem:lat",
		   "memory access latency (<first_chunk> <inter_chunk>)",
		   mem_lat, mem_nelt, &mem_nelt, mem_lat,
		   /* print */TRUE, /* format */NULL, /* !accrue */FALSE);

  opt_reg_int(odb, "-mem:width", "memory access bus width (in bytes)",
	      &mem_bus_width, /* default */8,
	      /* print */TRUE, /* format */NULL);

  /* TLB options */

  opt_reg_string(odb, "-tlb:itlb",
		 "instruction TLB config, i.e., {<config>|none}",
		 &itlb_opt, "itlb:1:4096:64:l", /* print */TRUE, NULL);

  opt_reg_string(odb, "-tlb:dtlb",
		 "data TLB config, i.e., {<config>|none}",
		 &dtlb_opt, "dtlb:1:4096:128:l", /* print */TRUE, NULL);

  opt_reg_int(odb, "-tlb:lat",
	      "inst/data TLB miss latency (in cycles)",
	      &tlb_miss_lat, /* default */30,
	      /* print */TRUE, /* format */NULL);

  opt_reg_flag(odb, "-tlb:itlbshared", "share itlb cache between contexts",
	       &shared_itlb, /* default */TRUE,
	       /* print */TRUE, /* format */NULL);

  opt_reg_flag(odb, "-tlb:dtlbshared", "share dtlb cache between contexts",
	       &shared_dtlb, /* default */TRUE,
	       /* print */TRUE, /* format */NULL);


  /* resource configuration */

  opt_reg_int(odb, "-res:ialu",
	      "total number of integer ALU's available",
	      &res_ialu, /* default */fu_config[FU_IALU_INDEX].quantity,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-res:imult",
	      "total number of integer multiplier/dividers available",
	      &res_imult, /* default */fu_config[FU_IMULT_INDEX].quantity,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-res:memport",
	      "total number of memory system ports available (to CPU)",
	      &res_memport, /* default */fu_config[FU_MEMPORT_INDEX].quantity,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-res:fpalu",
	      "total number of floating point ALU's available",
	      &res_fpalu, /* default */fu_config[FU_FPALU_INDEX].quantity,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-res:fpmult",
	      "total number of floating point multiplier/dividers available",
	      &res_fpmult, /* default */fu_config[FU_FPMULT_INDEX].quantity,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-res:cxtmsg",
	      "total number of msg unit available",
	      &res_cxtmsg, /* default */fu_config[FU_CXTMSG_INDEX].quantity,
	      /* print */TRUE, /* format */NULL);

  opt_reg_string_list(odb, "-pcstat",
		      "profile stat(s) against text addr's (mult uses ok)",
		      pcstat_vars, MAX_PCSTAT_VARS, &pcstat_nelt, NULL,
		      /* !print */FALSE, /* format */NULL, /* accrue */TRUE);

  opt_reg_uint_list(odb, "-interrupt",
                    "external interruption config "
		    "(<first_cycle> <port> <delta_cycle>)",
                    interrupt_config, interrupt_nelt, &interrupt_nelt,
		   /* default */interrupt_config,
                   /* print */TRUE, /* format */NULL, /* !accrue */FALSE);

  opt_reg_flag(odb, "-bugcompat",
	       "operate in backward-compatible bugs mode (for testing only)",
	       &bugcompat_mode, /* default */FALSE, /* print */TRUE, NULL);

}

/* check simulator-specific option values */
void
sim_check_options(struct opt_odb_t *odb,        /* options database */
		  int argc, char **argv)        /* command line arguments */
{
  char name[128], c;
  int nsets, bsize, assoc;
  SS_CONTEXT_TYPE context;

  if (fastfwd_count < 0 || fastfwd_count >= 2147483647)
    fatal("bad fast forward count: %d", fastfwd_count);

  if (main_offset < 0)
    fatal("offset value should be non-negative: %d", main_offset);

  if (!(num_thread > 0) || (num_thread & (num_thread-1)) != 0)
    fatal("num:thread should be positive and power of two: %d", num_thread);

  if (ruu_ifq_size < 1 || (ruu_ifq_size & (ruu_ifq_size - 1)) != 0)
    fatal("inst fetch queue size must be positive > 0 and a power of two");

  if (ruu_branch_penalty < 1)
    fatal("mis-prediction penalty must be at least 1 cycle");

  if (fetch_speed < 1)
    fatal("front-end speed must be positive and non-zero");

  if (fetch_block_size < 1)
    fatal("fetch block size must be positive and non-zero");

  if (prefetch_cache_probe < 0)
    fatal("cacheprobe must be positive or zero");

  if (fetch_waiting < 1)
    fatal("waiting idle cycles must be positive and non-zero");

  if (!mystricmp(pred_type, "perfect"))
    {
      /* perfect predictor */

      for(context = 0; context < context_num; context++)
        pred[context] = NULL;
      pred_perfect = TRUE;
    }
  else if (!mystricmp(pred_type, "taken"))
    {
      /* static predictor, not taken */
      for(context = 0; context < context_num; context++) {
	if (!bpred_shared)
	  pred[context] = bpred_create(BPredTaken, 0, 0, 0, 0, 0, 0, 0, 0);
	else {
	  if (context == 0)
	    pred[context] = bpred_create(BPredTaken, 0, 0, 0, 0, 0, 0, 0, 0);
	  else
	    pred[context] = pred[0];
	}
	retstack[context] = retstack_create(0); /* GT: create private stack */
      }
    }
  else if (!mystricmp(pred_type, "nottaken"))
    {
      /* static predictor, taken */
      for(context = 0; context < context_num; context++) {
	if (!bpred_shared)
	  pred[context] =  bpred_create(BPredNotTaken, 0, 0, 0, 0, 0, 0, 0, 0);
	else {
	  if (context == 0)
	    pred[context] = bpred_create(BPredNotTaken, 0, 0, 0, 0, 0, 0, 0, 0);
	  else
	    pred[context] = pred[0];
	}
	retstack[context] = retstack_create(0); /* GT: create private stack */
      }
    }
  else if (!mystricmp(pred_type, "bimod"))
    {
      /* bimodal predictor, bpred_create() checks BTB_SIZE */
      if (bimod_nelt != 1)
	fatal("bad bimod predictor config (<table_size>)");
      if (btb_nelt != 2)
	fatal("bad btb config (<num_sets> <associativity>)");

      /* bimodal predictor, bpred_create() checks BTB_SIZE */
      for(context = 0; context < context_num; context++) {
	if (!bpred_shared)
	  pred[context] = bpred_create(BPred2bit,
				       /* bimod table size */bimod_config[0],
				       /* 2lev l1 size */0,
				       /* 2lev l2 size */0,
				       /* meta table size */0,
				       /* history reg size */0,
				       /* history xor address */0,
				       /* btb sets */btb_config[0],
				       /* btb assoc */btb_config[1]);
	else {
	  if (context == 0)
	    pred[context] = bpred_create(BPred2bit,
					 /* bimod table size */bimod_config[0],
					 /* 2lev l1 size */0,
					 /* 2lev l2 size */0,
					 /* meta table size */0,
					 /* history reg size */0,
					 /* history xor address */0,
					 /* btb sets */btb_config[0],
					 /* btb assoc */btb_config[1]);
	  else
	    pred[context] = pred[0];
	}
	retstack[context] = retstack_create(ras_size); /* GT: create private stack */
      }
    }
  else if (!mystricmp(pred_type, "2lev"))
    {
      /* 2-level adaptive predictor, bpred_create() checks args */
      if (twolev_nelt != 4)
	fatal("bad 2-level pred config (<l1size> <l2size> <hist_size> <xor>)");
      if (btb_nelt != 2)
	fatal("bad btb config (<num_sets> <associativity>)");

      for(context = 0; context < context_num; context++) {
	if (!bpred_shared)
	  pred[context] = bpred_create(BPred2Level,
				       /* bimod table size */0,
				       /* 2lev l1 size */twolev_config[0],
				       /* 2lev l2 size */twolev_config[1],
				       /* meta table size */0,
				       /* history reg size */twolev_config[2],
				       /* history xor address */twolev_config[3],
				       /* btb sets */btb_config[0],
				       /* btb assoc */btb_config[1]);
	else {
	  if (context == 0)
	    pred[context] = bpred_create(BPred2Level,
					 /* bimod table size */0,
					 /* 2lev l1 size */twolev_config[0],
					 /* 2lev l2 size */twolev_config[1],
					 /* meta table size */0,
					 /* history reg size */twolev_config[2],
					 /* history xor address */twolev_config[3],
					 /* btb sets */btb_config[0],
					 /* btb assoc */btb_config[1]);
	  else
	    pred[context] = pred[0];
	}
        retstack[context] = retstack_create(ras_size); /* GT: create private stack */
      }
    }
  else if (!mystricmp(pred_type, "comb"))
    {
      /* combining predictor, bpred_create() checks args */
      if (twolev_nelt != 4)
	fatal("bad 2-level pred config (<l1size> <l2size> <hist_size> <xor>)");
      if (bimod_nelt != 1)
	fatal("bad bimod predictor config (<table_size>)");
      if (comb_nelt != 1)
	fatal("bad combining predictor config (<meta_table_size>)");
      if (btb_nelt != 2)
	fatal("bad btb config (<num_sets> <associativity>)");

      for(context = 0; context < context_num; context++) {
	if (!bpred_shared)
	  pred[context] = bpred_create(BPredComb,
				       /* bimod table size */bimod_config[0],
				       /* l1 size */twolev_config[0],
				       /* l2 size */twolev_config[1],
				       /* meta table size */comb_config[0],
				       /* history reg size */twolev_config[2],
				       /* history xor address */twolev_config[3],
				       /* btb sets */btb_config[0],
				       /* btb assoc */btb_config[1]);
	else {
	  if (context == 0)
	    pred[context] = bpred_create(BPredComb,
					 /* bimod table size */bimod_config[0],
					 /* l1 size */twolev_config[0],
					 /* l2 size */twolev_config[1],
					 /* meta table size */comb_config[0],
					 /* history reg size */twolev_config[2],
					 /* history xor address */twolev_config[3],
					 /* btb sets */btb_config[0],
					 /* btb assoc */btb_config[1]);
	  else
	    pred[context] = pred[0];
	}
	retstack[context] = retstack_create(ras_size); /* GT: create private stack */
      }
    }
  else
    fatal("cannot parse predictor type `%s'", pred_type);

  if (!bpred_spec_opt)
    bpred_spec_update = spec_CT;
  else if (!mystricmp(bpred_spec_opt, "ID"))
    bpred_spec_update = spec_ID;
  else if (!mystricmp(bpred_spec_opt, "WB"))
    bpred_spec_update = spec_WB;
  else
    fatal("bad speculative update stage specifier, use {ID|WB}");

  if (ruu_decode_width < 1 || (ruu_decode_width & (ruu_decode_width-1)) != 0)
    fatal("issue width must be positive non-zero and a power of two");

  if (ruu_issue_width < 1 || (ruu_issue_width & (ruu_issue_width-1)) != 0)
    fatal("issue width must be positive non-zero and a power of two");

  if (ruu_commit_width < 1)
    fatal("commit width must be positive non-zero");

  if (RUU_size < 2 || (RUU_size & (RUU_size-1)) != 0)
    fatal("RUU size must be a positive number > 1 and a power of two");

  if (LSQ_size < 2 || (LSQ_size & (LSQ_size-1)) != 0)
    fatal("LSQ size must be a positive number > 1 and a power of two");


  /* use a level 1 D-cache? */
  if (!mystricmp(cache_dl1_opt, "none"))
    {
      for(context = 0; context < (shared_dl1 ? 1 : context_num); context++)
        cache_dl1[context] = NULL;

      /* the level 2 D-cache cannot be defined */
      if (strcmp(cache_dl2_opt, "none"))
        fatal("the l1 data cache must defined if the l2 cache is defined");
      for(context = 0; context < (shared_dl2 ? 1 : context_num); context++)
        cache_dl2[context] = NULL;
    }
  else /* dl1 is defined */
    {
      if (sscanf(cache_dl1_opt, "%[^:]:%d:%d:%d:%c",
		 name, &nsets, &bsize, &assoc, &c) != 5)
	fatal("bad l1 D-cache parms: <name>:<nsets>:<bsize>:<assoc>:<repl>");
      for(context = 0; context < (shared_dl1 ? 1 : context_num); context++)
        cache_dl1[context] = cache_create(name, nsets, bsize,
                                          /* balloc */FALSE,
                                          /* usize */0, assoc,
                                          cache_char2policy(c),
                                          dl1_access_fn,
                                          /* hit lat */cache_dl1_lat);

      /* is the level 2 D-cache defined? */
      if (!mystricmp(cache_dl2_opt, "none")) {
        for(context = 0; context < (shared_dl2 ? 1 : context_num); context++)
          cache_dl2[context] = NULL;
        for(context = 1; context < (shared_dl2 ? 1 : context_num); context++)
          cache_connect_bus(cache_dl1[0], cache_dl1[context]);
      } else
	{
	  if (sscanf(cache_dl2_opt, "%[^:]:%d:%d:%d:%c",
		     name, &nsets, &bsize, &assoc, &c) != 5)
	    fatal("bad l2 D-cache parms: "
		  "<name>:<nsets>:<bsize>:<assoc>:<repl>");
          for(context = 0; context < (shared_dl2 ? 1 : context_num); context++)
            cache_dl2[context] = cache_create(name, nsets, bsize,
                                              /* balloc */FALSE,
                                              /* usize */0, assoc,
                                              cache_char2policy(c),
                                              dl2_access_fn,
                                              /* hit lat */cache_dl2_lat);
          for(context = 1; context < (shared_dl2 ? 1 : context_num); context++)
            cache_connect_bus(cache_dl2[0], cache_dl2[context]);

        }
    }

  /* use a level 1 I-cache? */
  if (!mystricmp(cache_il1_opt, "none"))
    {
      for(context = 0; context < (shared_il1 ? 1 : context_num); context++) {
        cache_il1[context] = NULL;
        ruu_prefetch_icache[context] = NULL;
      }
      /* the level 2 I-cache cannot be defined */
      if (strcmp(cache_il2_opt, "none"))
	fatal("the l1 inst cache must defined if the l2 cache is defined");
      for(context = 0; context < (shared_il2 ? 1 : context_num); context++)
        cache_il2[context] = NULL;
    }
  else if (!mystricmp(cache_il1_opt, "dl1"))
    {
      for(context = 0; context < (shared_dl1 ? 1 : context_num); context++)
        if (!cache_dl1[context])
          fatal("I-cache l1 cannot access D-cache l1 as it's undefined");
      for(context = 0; context < (shared_il1 ? 1 : context_num); context++) {
        cache_il1[context] = cache_dl1[context];
        ruu_prefetch_icache[context] = cache_dl1[context];
      }

      /* the level 2 I-cache cannot be defined */
      if (strcmp(cache_il2_opt, "none"))
	fatal("the l1 inst cache must defined if the l2 cache is defined");
      for(context = 0; context < (shared_il2 ? 1 : context_num); context++)
        cache_il2[context] = NULL;
    }
  else if (!mystricmp(cache_il1_opt, "dl2"))
    {
      for(context = 0; context < (shared_dl2 ? 1 : context_num); context++)
        if (!cache_dl2[context])
          fatal("I-cache l1 cannot access D-cache l2 as it's undefined");
      for(context = 0; context < (shared_il2 ? 1 : context_num); context++) {
        cache_il1[context] = cache_dl2[context];
        ruu_prefetch_icache[context] = cache_dl2[context];
      }

      /* the level 2 I-cache cannot be defined */
      if (strcmp(cache_il2_opt, "none"))
	fatal("the l1 inst cache must defined if the l2 cache is defined");
      for(context = 0; context < (shared_il2 ? 1 : context_num); context++)
        cache_il2[context] = NULL;
    }
  else /* il1 is defined */
    {
      if (sscanf(cache_il1_opt, "%[^:]:%d:%d:%d:%c",
		 name, &nsets, &bsize, &assoc, &c) != 5)
	fatal("bad l1 I-cache parms: <name>:<nsets>:<bsize>:<assoc>:<repl>");
      for(context = 0; context < (shared_il1 ? 1 : context_num); context++) {
        cache_il1[context] = cache_create(name, nsets, bsize,
                                          /* balloc */FALSE,
                                          /* usize */0, assoc,
                                          cache_char2policy(c),
                                          il1_access_fn,
                                          /* hit lat */cache_il1_lat);
        ruu_prefetch_icache[context] = cache_il1[context];
      }

      /* is the level 2 D-cache defined? */
      if (!mystricmp(cache_il2_opt, "none")) {
        for(context = 0; context < (shared_il2 ? 1 : context_num); context++)
          cache_il2[context] = NULL;
        for(context = 0; context < (shared_il1 ? 1 : context_num); context++)
          cache_connect_bus(cache_il1[0], cache_il1[context]);
      } else if (!mystricmp(cache_il2_opt, "dl2"))
	{
          for(context = 0; context < (shared_dl2 ? 1 : context_num); context++)
            if (!cache_dl2[context])
              fatal("I-cache l2 cannot access D-cache l2 as it's undefined");
          for(context = 0; context < (shared_il2 ? 1 : context_num); context++){
            cache_il2[context] = cache_dl2[context];
          }
	}
      else
	{
	  if (sscanf(cache_il2_opt, "%[^:]:%d:%d:%d:%c",
		     name, &nsets, &bsize, &assoc, &c) != 5)
	    fatal("bad l2 I-cache parms: "
		  "<name>:<nsets>:<bsize>:<assoc>:<repl>");
          for(context = 0; context < (shared_il2 ? 1 :context_num); context++)
            cache_il2[context] = cache_create(name, nsets, bsize,
                                              /* balloc */FALSE,
                                              /* usize */0, assoc,
                                              cache_char2policy(c),
                                              il2_access_fn,
                                              /* hit lat */cache_il2_lat);
          for(context = 0; context < (shared_il1 ? 1 : context_num); context++)
            cache_connect_bus(cache_il2[0], cache_il2[context]);
	}
    }

  /* use an I-TLB? */
  if (!mystricmp(itlb_opt, "none"))
    for(context = 0; context < context_num; context++)
      itlb[context] = NULL;
  else
    {
      if (sscanf(itlb_opt, "%[^:]:%d:%d:%d:%c",
		 name, &nsets, &bsize, &assoc, &c) != 5)
	fatal("bad TLB parms: <name>:<nsets>:<page_size>:<assoc>:<repl>");
      for(context = 0; context < (shared_itlb ? 1 : context_num); context++)
        itlb[context] = cache_create(name, nsets, bsize, /* balloc */FALSE,
                                     /* usize */sizeof(SS_ADDR_TYPE), assoc,
                                     cache_char2policy(c), itlb_access_fn,
                                     /* hit latency */1);
    }

  /* use a D-TLB? */
  if (!mystricmp(dtlb_opt, "none"))
    for(context = 0; context < context_num; context++)
      dtlb[context] = NULL;
  else
    {
      if (sscanf(dtlb_opt, "%[^:]:%d:%d:%d:%c",
		 name, &nsets, &bsize, &assoc, &c) != 5)
	fatal("bad TLB parms: <name>:<nsets>:<page_size>:<assoc>:<repl>");
      for(context = 0; context < (shared_dtlb ? 1 : context_num); context++)
        dtlb[context] = cache_create(name, nsets, bsize, /* balloc */FALSE,
                                     /* usize */sizeof(SS_ADDR_TYPE), assoc,
                                     cache_char2policy(c), dtlb_access_fn,
                                     /* hit latency */1);
    }
  
    for(context = 1; context < (shared_il1 ? context_num : 0); context++)
      cache_il1[context] = cache_il1[0];
    for(context = 1; context < (shared_il2 ? context_num : 0); context++)
      cache_il2[context] = cache_il2[0];
    for(context = 1; context < (shared_dl1 ? context_num : 0); context++)
      cache_dl1[context] = cache_dl1[0];
    for(context = 1; context < (shared_dl2 ? context_num : 0); context++)
      cache_dl2[context] = cache_dl2[0];
    for(context = 1; context < (shared_itlb ? context_num : 0); context++)
      itlb[context] = itlb[0];
    for(context = 1; context < (shared_dtlb ? context_num : 0); context++)
      dtlb[context] = dtlb[0];


  if (cache_dl1_lat < 1)
    fatal("l1 data cache latency must be greater than zero");

  if (cache_dl2_lat < 1)
    fatal("l2 data cache latency must be greater than zero");

  if (cache_il1_lat < 1)
    fatal("l1 instruction cache latency must be greater than zero");

  if (cache_il2_lat < 1)
    fatal("l2 instruction cache latency must be greater than zero");

  if (mem_nelt != 2)
    fatal("bad memory access latency (<first_chunk> <inter_chunk>)");

  if (mem_lat[0] < 1 || mem_lat[1] < 1)
    fatal("all memory access latencies must be greater than zero");

  if (mem_bus_width < 1 || (mem_bus_width & (mem_bus_width-1)) != 0)
    fatal("memory bus width must be positive non-zero and a power of two");

  if (tlb_miss_lat < 1)
    fatal("TLB miss latency must be greater than zero");

  if (res_ialu < 1)
    fatal("number of integer ALU's must be greater than zero");
  if (res_ialu > MAX_INSTS_PER_CLASS)
    fatal("number of integer ALU's must be <= MAX_INSTS_PER_CLASS");
  fu_config[FU_IALU_INDEX].quantity = res_ialu;
  
  if (res_imult < 1)
    fatal("number of integer multiplier/dividers must be greater than zero");
  if (res_imult > MAX_INSTS_PER_CLASS)
    fatal("number of integer mult/div's must be <= MAX_INSTS_PER_CLASS");
  fu_config[FU_IMULT_INDEX].quantity = res_imult;
  
  if (res_memport < 1)
    fatal("number of memory system ports must be greater than zero");
  if (res_memport > MAX_INSTS_PER_CLASS)
    fatal("number of memory system ports must be <= MAX_INSTS_PER_CLASS");
  fu_config[FU_MEMPORT_INDEX].quantity = res_memport;
  
  if (res_fpalu < 1)
    fatal("number of floating point ALU's must be greater than zero");
  if (res_fpalu > MAX_INSTS_PER_CLASS)
    fatal("number of floating point ALU's must be <= MAX_INSTS_PER_CLASS");
  fu_config[FU_FPALU_INDEX].quantity = res_fpalu;
  
  if (res_fpmult < 1)
    fatal("number of floating point multiplier/dividers must be > zero");
  if (res_fpmult > MAX_INSTS_PER_CLASS)
    fatal("number of FP mult/div's must be <= MAX_INSTS_PER_CLASS");
  fu_config[FU_FPMULT_INDEX].quantity = res_fpmult;

  if (res_cxtmsg < 1)
    fatal("number of context management unit must be > zero");
  if (res_fpmult > MAX_INSTS_PER_CLASS)
    fatal("number of context management unit must be <= MAX_INSTS_PER_CLASS");
  fu_config[FU_FPMULT_INDEX].quantity = res_cxtmsg;

  if ((interrupt_config[1] < 256) && (interrupt_config[1] != 0))
    fatal ("port used to catch interrupt should be >= 256 or 0");

}

/* print simulator-specific configuration information */
void
sim_aux_config(FILE *stream)            /* output stream */
{
  /* nada */
}

/* register simulator-specific statistics */
void
sim_reg_stats(struct stat_sdb_t *sdb)   /* stats database */
{
  int i;
  char *buffer;

  SS_CONTEXT_TYPE context;

  /* register baseline stats */
  stat_reg_counter(sdb, "sim_num_insn",
		   "total number of instructions committed",
		   &sim_num_insn, 0, NULL);

  for(context = 0; context < context_num; context++) {
    char s_description[80], s_variable_name[80];

    buffer = calloc(20, sizeof(char));
    if (!buffer)
      fatal("unable to allocate memory\n");
    
    /* FIXME DM: Change strcpy to sprintf */

    if (strcpy(s_description,
               "number of instr committed for context ") == NULL)
      fatal("strcpy failed to copy");
    if (strcpy(s_variable_name, "sim_num_insn.") == NULL)
      fatal("strcpy failed to copy");

    sprintf(buffer, "%d", context);

    strcat(s_variable_name, buffer);
    strcat(s_description, buffer);
    stat_reg_counter(sdb, s_variable_name, s_description,
                     &sim_num_insn_ctxt[context], 0, NULL);
    if (strcpy(s_description,
               "number of resumings for context ") == NULL)
      fatal("strcpy failed to copy");
    if (strcpy(s_variable_name, "sim_num_start.") == NULL)
      fatal("strcpy failed to copy");

    sprintf(buffer, "%d", context);

    strcat(s_variable_name, buffer);
    strcat(s_description, buffer);
    stat_reg_counter(sdb, s_variable_name, s_description,
                     &sim_num_start[context], 0, NULL);
    free(buffer);

  } /* for */

  stat_reg_counter(sdb, "sim_num_refs",
		   "total number of loads and stores committed",
		   &sim_num_refs, 0, NULL);
  stat_reg_counter(sdb, "sim_num_loads",
		   "total number of loads committed",
		   &sim_num_loads, 0, NULL);
  stat_reg_formula(sdb, "sim_num_stores",
		   "total number of stores committed",
		   "sim_num_refs - sim_num_loads", NULL);
  stat_reg_counter(sdb, "sim_num_branches",
		   "total number of branches committed",
		   &sim_num_branches, /* initial value */0, /* format */NULL);
  stat_reg_int(sdb, "sim_elapsed_time",
	       "total simulation time in seconds",
	       (int *)&sim_elapsed_time, 0, NULL);
  stat_reg_formula(sdb, "sim_inst_rate",
		   "simulation speed (in insts/sec)",
		   "sim_num_insn / sim_elapsed_time", NULL);

  stat_reg_counter(sdb, "sim_total_insn",
		   "total number of instructions executed",
		   &sim_total_insn, 0, NULL);
  stat_reg_counter(sdb, "sim_total_refs",
		   "total number of loads and stores executed",
		   &sim_total_refs, 0, NULL);
  stat_reg_counter(sdb, "sim_total_loads",
		   "total number of loads executed",
		   &sim_total_loads, 0, NULL);
  stat_reg_formula(sdb, "sim_total_stores",
		   "total number of stores executed",
		   "sim_total_refs - sim_total_loads", NULL);
  stat_reg_counter(sdb, "sim_total_branches",
		   "total number of branches executed",
		   &sim_total_branches, /* initial value */0, /* format */NULL);

  /* register performance stats */
  stat_reg_counter(sdb, "sim_cycle",
		   "total simulation time in cycles",
		   &sim_cycle, /* initial value */0, /* format */NULL);
  stat_reg_formula(sdb, "sim_IPC",
		   "instructions per cycle",
		   "sim_num_insn / sim_cycle", /* format */NULL);
  stat_reg_formula(sdb, "sim_CPI",
		   "cycles per instruction",
		   "sim_cycle / sim_num_insn", /* format */NULL);
  stat_reg_formula(sdb, "sim_exec_BW",
		   "total instructions (mis-spec + committed) per cycle",
		   "sim_total_insn / sim_cycle", /* format */NULL);
  stat_reg_formula(sdb, "sim_IPB",
		   "instruction per branch",
		   "sim_num_insn / sim_num_branches", /* format */NULL);

  /* register predictor stats */

   for(context = 0; context < context_num; context++) {
    if (pred[context])
      bpred_reg_stats(pred[context], context, sdb);
    if (retstack[context])
      retstack_reg_stats(retstack[context], context, sdb);
   }
  
/*   bpred_reg_stats(pred[0], sdb); */


  /* register cache stats */
  for(context = 0; context < (shared_il1 ? 1 : context_num); context++)
    if (cache_il1[context]
        && (cache_il1[context] != cache_dl1[context] && 
            cache_il1[context] != cache_dl2[context]))
      cache_reg_stats(cache_il1[context], context, sdb);
  for(context = 0; context < (shared_il2 ? 1 : context_num); context++)
    if (cache_il2[context]
        && (cache_il2[context] != cache_dl1[context] &&
            cache_il2[context] != cache_dl2[context]))
      cache_reg_stats(cache_il2[context], context, sdb);
  for(context = 0; context < (shared_dl1 ? 1 : context_num); context++)
    if (cache_dl1[context])
      cache_reg_stats(cache_dl1[context], context, sdb);
  for(context = 0; context < (shared_dl2 ? 1 : context_num); context++)
    if (cache_dl2[context])
      cache_reg_stats(cache_dl2[context], context, sdb);
  for(context = 0; context < (shared_itlb ? 1 : context_num); context++)
    if (itlb[context])
      cache_reg_stats(itlb[context], context, sdb);
  for(context = 0; context < (shared_dtlb ? 1 : context_num); context++)
    if (dtlb[context])
      cache_reg_stats(dtlb[context], context, sdb);

  for (i=0; i<pcstat_nelt; i++)
    {
      char buf[512], buf1[512];
      struct stat_stat_t *stat;

      /* track the named statistical variable by text address */

      /* find it... */
      stat = stat_find_stat(sdb, pcstat_vars[i]);
      if (!stat)
	fatal("cannot locate any statistic named `%s'", pcstat_vars[i]);

      /* stat must be an integral type */
      if (stat->sc != sc_int && stat->sc != sc_uint && stat->sc != sc_counter)
	fatal("`-pcstat' statistical variable `%s' is not an integral type",
	      stat->name);

      /* register this stat */
      pcstat_stats[i] = stat;
      pcstat_lastvals[i] = STATVAL(stat);

      /* declare the sparce text distribution */
      sprintf(buf, "%s_by_pc", stat->name);
      sprintf(buf1, "%s (by text address)", stat->desc);
      pcstat_sdists[i] = stat_reg_sdist(sdb, buf, buf1,
					/* initial value */0,
					/* print format */(PF_COUNT|PF_PDF),
					/* format */"0x%lx %lu %.2f",
					/* print fn */NULL);
    }
}

/* forward declarations */
static void ruu_init(void);
static void lsq_init(void);
static void rslink_init(int nlinks);
static void eventq_init(void);
static void readyq_init(void);
static void cv_init(void);
static void tracer_init(void);
static void fetch_init(void);
static void inter_init(void);

/* default register state accessor, used by DLite */
static char *					/* err str, NULL for no err */
simoo_reg_obj(enum dlite_access_t at,		/* access type */
              SS_CONTEXT_TYPE context,          /* context to access */
	      enum dlite_reg_t rt,		/* reg bank to probe */
	      int reg,				/* register number */
	      union dlite_reg_val_t *val);      /* input, output */

/* default memory state accessor, used by DLite */
static char *					/* err str, NULL for no err */
simoo_mem_obj(enum dlite_access_t at,		/* access type */
              SS_CONTEXT_TYPE context,          /* context to access */
	      SS_ADDR_TYPE addr,		/* address to access */
	      char *p,				/* input/output buffer */
	      int nbytes);			/* size of access */

/* default machine state accessor, used by DLite */
static char *					/* err str, NULL for no err */
simoo_mstate_obj(FILE *stream,			/* output stream */
		 char *cmd);			/* optional command string */

/* total RS links allocated at program start */
#define MAX_RS_LINKS                    32768   /* was 4096 */



/*
 * processor core definitions and declarations
 */

/* inst tag type, used to tag an operation instance in the RUU */
typedef unsigned int INST_TAG_TYPE;

/* inst sequence type, used to order instructions in the ready list, if
   this rolls over the ready list order temporarily will get messed up,
   but execution will continue and complete correctly */
typedef unsigned int INST_SEQ_TYPE;

/* is given context running? */
#define ACTIVE_CXT(CONTEXT) \
   ((regs_ACXT >> (unsigned int)(CONTEXT)) & 0x00000001)

/* set given context running or suspended */
#define SET_CONTEXT_SUSPENDED(CONTEXT) (regs_ACXT &= ~(1 << (CONTEXT)))
#define SET_CONTEXT_ACTIVE(CONTEXT) (regs_ACXT |= (1 << (CONTEXT)))

/* read context priority */
#define GET_CONTEXT_PRIORITY(CONTEXT) \
  ((regs_CXTCTRL[(CONTEXT)] >> 8) & 0x0000001f)
     
/* set context priority */
#define SET_CONTEXT_PRIORITY(CONTEXT,PRIORITY) \
  ((regs_CXTCTRL[(CONTEXT)] | 0xffffe0ff) | ((PRIORITY) << 8))
     
/* context id */
#define GET_CONTEXT_ID(CONTEXT) \
  (regs_CXTCTRL[(CONTEXT)] & 0x0000001f)

/* total input dependencies possible */
#define MAX_IDEPS               3

/* total output dependencies possible */
#define MAX_ODEPS               2

/* a register update unit (RUU) station, this record is contained in the
   processors RUU, which serves as a collection of ordered reservations
   stations.  The reservation stations capture register results and await
   the time when all operands are ready, at which time the instruction is
   issued to the functional units; the RUU is an order circular queue, in which
   instructions are inserted in fetch (program) order, results are stored in
   the RUU buffers, and later when an RUU entry is the oldest entry in the
   machines, it and its instruction's value is retired to the architectural
   register file in program order, NOTE: the RUU and LSQ share the same
   structure, this is useful because loads and stores are split into two
   operations: an effective address add and a load/store, the add is inserted
   into the RUU and the load/store inserted into the LSQ, allowing the add
   to wake up the load/store when effective address computation has finished.
   All RUU unit corresponding to an instructions are *REALLY* executed in
   this stage except instructions following a context halt whose RUU stations
   are reserved but will not be executed (because of the passivation of the
   context) */
struct RUU_station {
  /* inst info */
  SS_INST_TYPE IR;                      /* instruction bits */
  enum ss_opcode op;                    /* decoded instruction opcode */
  SS_ADDR_TYPE PC, next_PC, pred_PC;    /* inst PC, next PC, predicted PC */
  SS_CONTEXT_TYPE context;              /* context number */
  int in_LSQ;                           /* non-zero if op is in LSQ */
  int ea_comp;                          /* non-zero if op is an addr comp */
  int recover_inst;                     /* start of mis-speculation? */
  int stack_recover_idx;		/* non-speculative TOS for RSB pred */
  struct bpred_update dir_update;	/* bpred direction update info */
  int spec_mode;                        /* non-zero if issued in spec_mode */
  SS_ADDR_TYPE addr;                    /* effective address for ld/st's */
  INST_TAG_TYPE tag;                    /* RUU slot tag, increment to
					   squash operation */
  INST_SEQ_TYPE seq;                    /* instruction sequence, used to
					   sort the ready list and tag inst */
  unsigned int ptrace_seq;		/* pipetrace sequence number */

  /* instruction status */
  int idle;                             /* operation has not be executed */
  int queued;                           /* operands ready and queued */
  int issued;                           /* operation is/was executing */
  int completed;                        /* operation has completed execution */
  int retired;                          /* instruction has been retired */
                                        /* commited */
  int valid;				/* RUU_DK: valid bit */
  int my_seq;				/* RUU_DK: sequence no. (0 ~ RUU_size-1) */
  int prev;				/* previous entry */
  int next;				/* next entry */

  /* output operand dependency list, these lists are used to
     limit the number of associative searches into the RUU when
     instructions complete and need to wake up dependent insts */
  int onames[MAX_ODEPS];                /* output logical names (NA=unused) */
  struct RS_link *odep_list[MAX_ODEPS]; /* chains to consuming operations */

  /* input dependent links, the output chains rooted above use these
     fields to mark input operands as ready, when all these fields have
     been set non-zero, the RUU operation has all of its register
     operands, it may commence execution as soon as all of its memory
     operands are known to be read (see lsq_refresh() for details on
     enforcing memory dependencies) */
  int idep_ready[MAX_IDEPS];            /* input operand ready? */

  /* The next fields indicate to the issue stage if a context should
     be resumed or stopped and if resumed which value the PC must
     hold. */
  unsigned int          target_op;
  unsigned int          target_mask;
  unsigned int          context_op_seq;
/*    SS_WORD_TYPE          target_PC; */
/*    SS_WORD_TYPE          target_pred_PC; */
};

/* non-zero if all register operands are ready, update with MAX_IDEPS */
#define OPERANDS_READY(RS)                                              \
  ((RS)->idep_ready[0] && (RS)->idep_ready[1] && (RS)->idep_ready[2])

/* register update unit, combination of reservation stations and reorder
   buffer device, organized as a circular queue */
static struct RUU_station *RUU;
                                        /* register update unit */
/* static int RUU_head, RUU_tail; */
static int RUU_head[SS_NUM_CONTEXT], RUU_tail[SS_NUM_CONTEXT]; /* RUU_DK */ 
                                        /* RUU head and tail pointers  for each context */
static int RUU_sequence[SS_NUM_CONTEXT];
					/* RUU_DK: RUU_sequence - seq. no. pool */

static int RUU_num;                     /* num entries currently in RUU */

static int RUU_num_cxt[SS_NUM_CONTEXT]; /* RUU_DK */
					/* RUU_num for each context */
                                        /* unified list of RUU stations */


/* allocate and initialize register update unit (RUU) */
static void
ruu_init(void)
{
  int i;

  RUU = (struct RUU_station *) calloc(RUU_size, sizeof(struct RUU_station));
  if (!RUU)
    fatal("out of virtual memory");

  RUU_num = 0;

  /* RUU_DK */
  for (i = 0; i < context_num; i++)  
    RUU_head[i] = RUU_tail[i] = RUU_num_cxt[i] = RUU_sequence[i] = 0;

}

/* dump the contents of the RUU */
static void
ruu_dumpent(struct RUU_station *rs,		/* ptr to RUU station */
	    int index,				/* entry index */
	    FILE *stream,			/* output stream */
	    int header)				/* print header? */
{
  int i;

  if (header)
    fprintf(stream, "idx: %2d: opcode: %s, inst: `",
	    index, SS_OP_NAME(rs->op));
  else
    fprintf(stream, "       opcode: %s, inst: `",
	    SS_OP_NAME(rs->op));
  ss_print_insn(rs->IR, rs->PC, stream);
  fprintf(stream, "'\n");
  fprintf(stream, "         context: %d\n", rs->context);
  fprintf(stream, "         PC: 0x%08x, NPC: 0x%08x (pred_PC: 0x%08x)\n",
	  rs->PC, rs->next_PC, rs->pred_PC);
  fprintf(stream, "         in_LSQ: %s, ea_comp: %s, recover_inst: %s\n",
	  rs->in_LSQ ? "t" : "f",
	  rs->ea_comp ? "t" : "f",
	  rs->recover_inst ? "t" : "f");
  fprintf(stream, "         spec_mode: %s, addr: 0x%08x, tag: 0x%08x\n",
	  rs->spec_mode ? "t" : "f", rs->addr, rs->tag);
  fprintf(stream, "         seq: 0x%08x, ptrace_seq: 0x%08x\n",
	  rs->seq, rs->ptrace_seq);
  fprintf(stream, "         queued: %s, issued: %s, completed: %s, retired; %s\n",
	  rs->queued ? "t" : "f",
	  rs->issued ? "t" : "f",
	  rs->completed ? "t" : "f",
          rs->retired ? "t" : "f");
  fprintf(stream, "         target_op:%s, target_mask:0x%08x, idle: %s\n",
          (rs->target_op == CTXT_CHNG_WAKEUP)?"resume":
           ((rs->target_op == CTXT_CHNG_FREEZE)?"stop":"none"),
          rs->target_mask,
          rs->idle ? "t" : "f");
  fprintf(stream, "         operands ready: %s\n",
	  OPERANDS_READY(rs) ? "t" : "f");
  for (i=0; i<3; i++) {
    fprintf(stream, "                 idep[%d] : %s\n",
            i, rs->idep_ready[i] ? "t" : "f");
  } /* for */
  for (i=0; i<2; i++) {
    fprintf(stream, "       onames[%d]:%d ",
            i, rs->onames[i]);
    fprintf(stream, "odep_list[%d]:0x%08x\n",
            i, (unsigned int) rs->odep_list[i]);
  } /* for */  
  fprintf(stream, "       physical address of RUU.PC: 0x%08x\n",
          (unsigned int) &(rs->PC));
}

/* dump the contents of the RUU */
static void
ruu_dump(FILE *stream)				/* output stream */
{
  int i;

  fprintf(stream, "** RUU state **\n");
  fprintf(stream, "RUU_num: %d\n",
          RUU_num);
    
  /* RUU_DK */
  for (i = 0; i < context_num; i++)
    fprintf(stream, "RUU_head[%2d]: %3d   RUU_tail[%2d]: %3d   RUU_num_cxt[%2d]: %3d\n",
	i, RUU_head[i], i, RUU_tail[i], i, RUU_num_cxt[i]);	
}


/*
 * load/store queue (LSQ): holds loads and stores in program order, indicating
 * status of load/store access:
 *
 *   - issued: address computation complete, memory access in progress
 *   - completed: memory access has completed, stored value available
 *   - squashed: memory access was squashed, ignore this entry
 *
 * loads may execute when:
 *   1) register operands are ready, and
 *   2) memory operands are ready (no earlier unresolved store)
 *
 * loads are serviced by:
 *   1) previous store at same address in LSQ (hit latency), or
 *   2) data cache (hit latency + miss latency)
 *
 * stores may execute when:
 *   1) register operands are ready
 *
 * stores are serviced by:
 *   1) depositing store value into the load/store queue
 *   2) writing store value to the store buffer (plus tag check) at commit
 *   3) writing store buffer entry to data cache when cache is free
 *
 * NOTE: the load/store queue can bypass a store value to a load in the same
 *   cycle the store executes (using a bypass network), thus stores complete
 *   in effective zero time after their effective address is known
 */
static int prev_entry;		/* save previous LSQ entry */

static struct RUU_station *LSQ; /* context load/store queue */
static int LSQ_oldest;  	/* points to the oldest entry in the LSQ */
static int LSQ_head[SS_NUM_CONTEXT], LSQ_tail[SS_NUM_CONTEXT]; /* LSQ_DK */
static int LSQ_num;             /* num entries currently in LSQ */

static int LSQ_sequence[SS_NUM_CONTEXT];
					/* LSQ_DK: LSQ_sequence - seq. no. pool */

static int LSQ_num_cxt[SS_NUM_CONTEXT]; /* LSQ_DK: LSQ_num for each context */

/*
 * input dependencies for stores in the LSQ:
 *   idep #0 - operand input (value that is store'd)
 *   idep #1 - effective address input (address of store operation)
 */
#define STORE_OP_INDEX                  0
#define STORE_ADDR_INDEX                1

#define STORE_OP_READY(RS)              ((RS)->idep_ready[STORE_OP_INDEX])
#define STORE_ADDR_READY(RS)            ((RS)->idep_ready[STORE_ADDR_INDEX])

/* allocate and initialize the load/store queue (LSQ) */
static void
lsq_init(void)
{
  int i;

  LSQ = (struct RUU_station *) calloc(LSQ_size, sizeof(struct RUU_station));
  if (!LSQ)
    fatal("out of virtual memory");

  LSQ_num = 0;
  LSQ_oldest = 0; 
  prev_entry = -1; /* -1 means NULL */

  /* LSQ_DK */
  for (i = 0; i < context_num; i++) 
    LSQ_head[i] = LSQ_tail[i] = LSQ_num_cxt[i] = LSQ_sequence[i] = 0;

}

/* dump the contents of the RUU */
static void
lsq_dump(FILE *stream)				/* output stream */
{
  int i;
/*int num, head;
  struct RUU_station *rs; */

  fprintf(stream, "** LSQ state **\n");
  fprintf(stream, "LSQ_head: %d\n", LSQ_oldest); 
  fprintf(stream, "LSQ_num: %d\n", LSQ_num);

  /* LSQ_DK */
  for (i = 0; i < context_num; i++)
    fprintf(stream, "LSQ_head[%2d]: %3d   LSQ_tail[%2d]: %3d   LSQ_num_cxt[%2d]: %3d\n",
	i, LSQ_head[i], i, LSQ_tail[i], i, LSQ_num_cxt[i]);	
}


/*
 * RS_LINK defs and decls
 */

/* a reservation station link: this structure links elements of a RUU
   reservation station list; used for ready instruction queue, event queue, and
   output dependency lists; each RS_LINK node contains a pointer to the RUU
   entry it references along with an instance tag, the RS_LINK is only valid if
   the instruction instance tag matches the instruction RUU entry instance tag;
   this strategy allows entries in the RUU can be squashed and reused without
   updating the lists that point to it, which significantly improves the
   performance of (all to frequent) squash events */
struct RS_link {
  struct RS_link *next;                 /* next entry in list */
  struct RS_link *next_used;            /* next entry in list of used RS_link */
  struct RUU_station *rs;               /* referenced RUU resv station */
  INST_TAG_TYPE tag;                    /* inst instance sequence number */
  union {
    SS_TIME_TYPE when;                  /* time stamp of entry (for eventq) */
    INST_SEQ_TYPE seq;                  /* inst sequence */
    int opnum;                          /* input/output operand number */
  } x;
};

/* RS link free list, grab RS_LINKs from here, when needed */
static struct RS_link *rslink_free_list;

/* RS used list. Used to iterate on used entries. */
static struct RS_link *rslink_used_list = NULL;

/* NULL value for an RS link */
#define RSLINK_NULL_DATA		{ NULL, NULL, NULL, 0 }
static struct RS_link RSLINK_NULL = RSLINK_NULL_DATA;

/* NULL value for an RS link */
#define RSLINK_INIT(RSL, RS)						\
  ((RSL).rs = (RS), (RSL).tag = (RS)->tag)

/* non-zero if RS link is NULL */
#define RSLINK_IS_NULL(LINK)            ((LINK)->rs == NULL)

/* non-zero if RS link is to a valid (non-squashed) entry */
#define RSLINK_VALID(LINK)                                              \
     ((LINK)->rs ? ((LINK)->tag == (LINK)->rs->tag) : FALSE)

/* extra RUU reservation station pointer */
#define RSLINK_RS(LINK)                 ((LINK)->rs)

/* get a new RS link record */
#define RSLINK_NEW(DST, RS)						\
  { struct RS_link *n_link;						\
    if (!rslink_free_list)						\
      panic("out of rs links");						\
    n_link = rslink_free_list;						\
    rslink_free_list = rslink_free_list->next;				\
    n_link->next_used = rslink_used_list;               		\
    n_link->next = NULL;                                       		\
    rslink_used_list = n_link;                                          \
    n_link->rs = (RS); n_link->tag = n_link->rs->tag;			\
    (DST) = n_link;							\
  }

/* free an RS link record */
#define RSLINK_FREE(LINK)						\
  {  struct RS_link *f_link = (LINK);					\
     struct RS_link *f_index = rslink_used_list;                        \
     struct RS_link *f_backup = rslink_used_list;                       \
     if (!rslink_used_list)                     			\
       panic("out of rs used links");					\
     f_link->rs = NULL; f_link->tag = 0;				\
     while (f_index != f_link) {                                        \
       f_backup = f_index;                                              \
       f_index = f_index->next_used;                                    \
       if (!f_index)                                                    \
         panic("link not found");                                       \
     }                                                                  \
     if (f_backup == f_link) {                                          \
       rslink_used_list = f_backup->next_used;                          \
     } else {                                                           \
       f_backup->next_used = f_index->next_used;                        \
     }                                                                  \
     f_link->next = rslink_free_list;					\
     rslink_free_list = f_link;					        \
  }

/* FIXME: could this be faster!!! */
/* free an RS link list */
#define RSLINK_FREE_LIST(LINK)						\
  {  struct RS_link *fl_link, *fl_link_next;				\
     for (fl_link=(LINK); fl_link; fl_link=fl_link_next)		\
       {								\
	 fl_link_next = fl_link->next;					\
	 RSLINK_FREE(fl_link);						\
       }								\
  }

/* the last operation that ruu_dispatch() attempted to dispatch, for
   implementing in-order issue. This variable is declared here
   because it needs to be changed by ruu_strip and has to be initialized 
   with a for loop . */
static struct RS_link last_op[SS_NUM_CONTEXT];

/* initialize the free RS_LINK pool */
static void
rslink_init(int nlinks)			/* total number of RS_LINK available */
{
  int i;
  struct RS_link *link;
  SS_CONTEXT_TYPE context;

  rslink_free_list = NULL;
  for (i=0; i<nlinks; i++)
    {
      link = calloc(1, sizeof(struct RS_link));
      if (!link)
	fatal("out of virtual memory");
      link->next = rslink_free_list;
      rslink_free_list = link;
    }

  for(context = 0; context < context_num; context++)
    last_op[context] = RSLINK_NULL;

}


/* service all functional unit release events, this function is called
   once per cycle, and it used to step the BUSY timers attached to each
   functional unit in the function unit resource pool, as long as a functional
   unit's BUSY count is > 0, it cannot be issued an operation */
static void
ruu_release_fu(void)
{
  int i;

  /* walk all resource units, decrement busy counts by one */
  for (i=0; i<fu_pool->num_resources; i++)
    {
      /* resource is released when BUSY hits zero */
      if (fu_pool->resources[i].busy > 0)
	fu_pool->resources[i].busy--;
    }
}


/*
 * the execution unit event queue implementation follows, the event queue
 * indicates which instruction will complete next, the writeback handler
 * drains this queue
 */

/* pending event queue, sorted from soonest to latest event (in time), NOTE:
   RS_LINK nodes are used for the event queue list so that it need not be
   updated during squash events */
static struct RS_link *event_queue;

/* initialize the event queue structures */
static void
eventq_init(void)
{
  event_queue = NULL;
}

/* dump the contents of the event queue */
static void
eventq_dump(FILE *stream)			/* output stream */
{
  struct RS_link *ev;

  fprintf(stream, "** event queue state **\n");

  for (ev = event_queue; ev != NULL; ev = ev->next)
    {
      /* is event still valid? */
      if (RSLINK_VALID(ev))
	{
	  struct RUU_station *rs = RSLINK_RS(ev);

	  fprintf(stream, "idx: %2d: @ %.0f\n",
		  rs - (rs->in_LSQ ? LSQ : RUU), (double)ev->x.when);
	  ruu_dumpent(rs,
                      rs - (rs->in_LSQ ? LSQ : RUU),
		      stream, /* !header */FALSE);
	}
    }
}


/* insert an event for RS into the event queue, event queue is sorted from
   earliest to latest event, event and associated side-effects will be
   apparent at the start of cycle WHEN */
static void
eventq_queue_event(struct RUU_station *rs, SS_TIME_TYPE when)
{
  struct RS_link *prev, *ev, *new_ev;

  if (rs->completed)
    panic("event completed");

  if (when <= sim_cycle)
    panic("event occurred in the past");

  /* get a free event record */
  RSLINK_NEW(new_ev, rs);
  new_ev->x.when = when;

  /* locate insertion point */
  for (prev=NULL, ev=event_queue;
       ev && ev->x.when < when;
       prev=ev, ev=ev->next);

  if (prev)
    {
      /* insert middle or end */
      new_ev->next = prev->next;
      prev->next = new_ev;
    }
  else
    {
      /* insert at beginning */
      new_ev->next = event_queue;
      event_queue = new_ev;
    }
}

/* return the next event that has already occurred, returns NULL when no
   remaining events or all remaining events are in the future */
static struct RUU_station *
eventq_next_event(void)
{
  struct RS_link *ev;

  if (event_queue && event_queue->x.when <= sim_cycle)
    {
      /* unlink and return first event on priority list */
      ev = event_queue;
      event_queue = event_queue->next;

      /* event still valid? */
      if (RSLINK_VALID(ev))
	{
	  struct RUU_station *rs = RSLINK_RS(ev);

	  /* reclaim event record */
	  RSLINK_FREE(ev);

	  /* event is valid, return resv station */
	  return rs;
	}
      else
	{
	  /* reclaim event record */
	  RSLINK_FREE(ev);

	  /* receiving inst was squashed, return next event */
	  return eventq_next_event();
	}
    }
  else
    {
      /* no event or no event is ready */
      return NULL;
    }
}


/*
 * the ready instruction queue implementation follows, the ready instruction
 * queue indicates which instruction have all of there *register* dependencies
 * satisfied, instruction will issue when 1) all memory dependencies for
 * the instruction have been satisfied (see lsq_refresh() for details on how
 * this is accomplished) and 2) resources are available; ready queue is fully
 * constructed each cycle before any operation is issued from it -- this
 * ensures that instruction issue priorities are properly observed; NOTE:
 * RS_LINK nodes are used for the event queue list so that it need not be
 * updated during squash events
 */

/* the ready instruction queue */
static struct RS_link *ready_queue;

/* initialize the event queue structures */
static void
readyq_init(void)
{
  ready_queue = NULL;
}

/* dump the contents of the ready queue */
static void
readyq_dump(FILE *stream)			/* output stream */
{
  struct RS_link *link;

  fprintf(stream, "** ready queue state **\n");

  for (link = ready_queue; link != NULL; link = link->next)
    {
      /* is entry still valid? */
      if (RSLINK_VALID(link))
	{
	  struct RUU_station *rs = RSLINK_RS(link);

	  ruu_dumpent(rs, rs - (rs->in_LSQ ? LSQ : RUU),
		      stream, /* header */TRUE);
	}
    }
}


/* insert ready node into the ready list using ready instruction scheduling
   policy; currently the following scheduling policy is enforced:

     memory and long latency operands, and branch instructions first

   then

     all other instructions, oldest instructions first

  this policy works well because branches pass through the machine quicker
  which works to reduce branch misprediction latencies, and very long latency
  instructions (such loads and multiplies) get priority since they are very
  likely on the program's critical path */

/* FIXMEDM: readyq should be order by thread context priorities */

static void
readyq_enqueue(struct RUU_station *rs)		/* RS to enqueue */
{
  struct RS_link *prev, *node, *new_node;

  /* node is now queued */
  if (rs->queued)
    panic("node is already queued");
  rs->queued = TRUE;

  /* get a free ready list node */
  RSLINK_NEW(new_node, rs);
  new_node->x.seq = rs->seq;

  /* locate insertion point */
  if (rs->in_LSQ || SS_OP_FLAGS(rs->op) & (F_LONGLAT|F_CTRL))
    {
      /* insert loads/stores and long latency ops at the head of the queue */
      prev = NULL;
      node = ready_queue;
    }
  else
    {
      /* otherwise insert in program order (earliest seq first) */
      for (prev=NULL, node=ready_queue;
	   node && node->x.seq < rs->seq;
	   prev=node, node=node->next);
    }

  if (prev)
    {
      /* insert middle or end */
      new_node->next = prev->next;
      prev->next = new_node;
    }
  else
    {
      /* insert at beginning */
      new_node->next = ready_queue;
      ready_queue = new_node;
    }
}


/*
 * the create vector maps a logical register to a creator in the RUU (and
 * specific output operand) or the architected register file (if RS_link
 * is NULL)
 */

/* an entry in the create vector */
struct CV_link {
  struct RUU_station *rs;               /* creator's reservation station */
  int odep_num;                         /* specific output operand */
};

/* a NULL create vector entry */
static struct CV_link CVLINK_NULL = { NULL, 0 };

/* get a new create vector link */
#define CVLINK_INIT(CV, RS, ONUM)	((CV).rs = (RS), (CV).odep_num = (ONUM))

/* size of the create vector (one entry per architected register) */
#define CV_BMAP_SZ              (BITMAP_SIZE(SS_TOTAL_REGS*SS_NUM_CONTEXT))

/* the create vector, NOTE: speculative copy on write storage provided
   for fast recovery during wrong path execute (see tracer_recover() for
   details on this process */
static BITMAP_TYPE(SS_TOTAL_REGS*SS_NUM_CONTEXT, use_spec_cv);
static struct CV_link create_vector[SS_TOTAL_REGS * SS_NUM_CONTEXT];
static struct CV_link spec_create_vector[SS_TOTAL_REGS * SS_NUM_CONTEXT];

/* these arrays shadow the create vector an indicate when a register was
   last created */
static SS_TIME_TYPE create_vector_rt[SS_TOTAL_REGS * SS_NUM_CONTEXT];
static SS_TIME_TYPE spec_create_vector_rt[SS_TOTAL_REGS * SS_NUM_CONTEXT];

/* read a create vector entry */
#define CREATE_VECTOR(N)        (BITMAP_SET_P(use_spec_cv, CV_BMAP_SZ, (N))\
				 ? spec_create_vector[(N)]              \
				 : create_vector[(N)])

/* read a create vector timestamp entry */
#define CREATE_VECTOR_RT(N)     (BITMAP_SET_P(use_spec_cv, CV_BMAP_SZ, (N))\
				 ? spec_create_vector_rt[(N)]           \
				 : create_vector_rt[(N)])

/* set a create vector entry */
#define SET_CREATE_VECTOR(N, L) (spec_mode[context]                     \
				 ? (BITMAP_SET(use_spec_cv, CV_BMAP_SZ, (N)),\
				    spec_create_vector[(N)] = (L))      \
				 : (create_vector[(N)] = (L)))

/* initialize the create vector */
static void
cv_init(void)
{
  int i;

  /* initially all registers are valid in the architected register file,
     i.e., the create vector entry is CVLINK_NULL */
  for (i=0; i < (SS_TOTAL_REGS * SS_NUM_CONTEXT); i++)
    {
      create_vector[i] = CVLINK_NULL;
      create_vector_rt[i] = 0;
      spec_create_vector[i] = CVLINK_NULL;
      spec_create_vector_rt[i] = 0;
    }

  /* all create vector entries are non-speculative */
  BITMAP_CLEAR_MAP(use_spec_cv, CV_BMAP_SZ);
}

/* dependency index names */
static char *dep_names[SS_TOTAL_REGS] = {
  "n/a", "$r1", "$r2", "$r3", "$r4", "$r5", "$r6", "$r7", "$r8",  "$r9",
  "$r10", "$r11", "$r12", "$r13", "$r14", "$r15", "$r16", "$r17", "$r18",
  "$r19", "$r20", "$r21", "$r22", "$r23", "$r24", "$r25", "$r26", "$r27",
  "$r28",  "$r29", "$r30", "$r31",
  "$f0", "$f1", "$f2", "$f3", "$f4", "$f5", "$f6", "$f7", "$f8",  "$f9",
  "$f10", "$f11", "$f12", "$f13", "$f14", "$f15", "$f16", "$f17", "$f18",
  "$f19", "$f20", "$f21", "$f22", "$f23", "$f24", "$f25", "$f26", "$f27",
  "$f28",  "$f29", "$f30", "$f31",
  "$hi", "$lo", "$fcc", "$tmp", "$acxt", "$cxtctrl",
  "n/a", "n/a"
};

/* dump the contents of the create vector */
static void
cv_dump(FILE *stream)				/* output stream */
{
  int i;
  struct CV_link ent;
  SS_CONTEXT_TYPE context;

  fprintf(stream, "** create vector state **\n");

  for(context = 0; context < context_num; context++) {
    for (i=0; i<SS_TOTAL_REGS; i++) {
      ent = CREATE_VECTOR((SS_TOTAL_REGS * context ) + i);
      if (!ent.rs)
	fprintf(stream, "[%4s]: from architected reg file\n", dep_names[i]);
      else
	fprintf(stream, "[%4s]: from %s, idx: %d\n",
	      dep_names[i], (ent.rs->in_LSQ ? "LSQ" : "RUU"),
		ent.rs - (ent.rs->in_LSQ ? LSQ : RUU));
    } /* for */
  } /* for */
}

/* DK: counters for the first point and the second point */
static unsigned total_point_count[8];
static unsigned prev_point1_time[8];
static unsigned prev_point2_time[8];
static unsigned prev_point3_time[8];
static unsigned total_point2_time[8];
static unsigned total_point3_time[8];
static unsigned total_point4_time[8];

static unsigned prev_point1_insn[8];
static unsigned prev_point2_insn[8];
static unsigned prev_point3_insn[8];
static unsigned total_point2_insn[8];
static unsigned total_point3_insn[8];
static unsigned total_point4_insn[8];

/* program counter */
static SS_ADDR_TYPE pred_PC[SS_NUM_CONTEXT];
static SS_ADDR_TYPE recover_PC[SS_NUM_CONTEXT];

/* context start/stop PC values */
static SS_ADDR_TYPE backup_PC[SS_NUM_CONTEXT];

static unsigned mark1_total[SS_NUM_CONTEXT];
static unsigned mark1_count[SS_NUM_CONTEXT];
static unsigned last_mark1[SS_NUM_CONTEXT];
static unsigned mark2_total[SS_NUM_CONTEXT];
static unsigned mark2_total_insn[SS_NUM_CONTEXT];
static unsigned mark2_count[SS_NUM_CONTEXT];
static unsigned last_mark2[SS_NUM_CONTEXT];
static unsigned last_mark2_insn[SS_NUM_CONTEXT];
static unsigned last_mark[SS_NUM_CONTEXT];

/* DY:  Check for marker debug instruction and print marker
   information if found. */
void check_marker_instruction(int RUU_index, SS_CONTEXT_TYPE context)
{
  unsigned annotate;

  if ((RUU[RUU_index].op == ADDIU) && !(RUU[RUU_index].spec_mode)) {

    annotate = (RUU[RUU_index].IR.a & 0x003f0000) >> 16;
    switch(annotate) {
    case 1:
/*      fprintf(stderr, "%4d MARK1(%x,%2d):  %10u, %10u\n", mark1_count[context],
	     RUU[RUU_index].PC, context, (unsigned)sim_cycle,
	     (unsigned)sim_cycle - last_mark1[context]);*/
      if (last_mark[context] == 1)
	mark1_total[context] += (unsigned) sim_cycle - last_mark1[context];
      last_mark1[context] = (unsigned) sim_cycle;
      mark1_count[context]++;
      last_mark[context] = 1;
      break;
    case 2:
/*      fprintf(stderr, "%4d MARK2(%x,%2d):  %10u, %10u\n", mark2_count[context],
	     RUU[RUU_index].PC, context, (unsigned)sim_cycle,
	     (unsigned)sim_cycle - last_mark2[context]);*/
      if (last_mark[context] == 2) {
	mark2_total[context] += (unsigned) sim_cycle - last_mark2[context];
	mark2_total_insn[context] += (unsigned) sim_num_insn_commit[context] - last_mark2_insn[context];
      }
      last_mark2[context] = (unsigned) sim_cycle;
      last_mark2_insn[context] = (unsigned) sim_num_insn_commit[context];
      mark2_count[context]++;
      last_mark[context] = 2;
      break;
    case 8:
      total_point_count[context]++;
      prev_point1_time[context] = (unsigned) sim_cycle;
      prev_point1_insn[context] = (unsigned) sim_num_insn_commit[context]+1; 
      break;
    case 16:
      total_point2_time[context] += ((unsigned) sim_cycle - prev_point1_time[context]);
      total_point2_insn[context] += ((unsigned) sim_num_insn_commit[context] - prev_point1_insn[context]);
      prev_point2_time[context] = (unsigned) sim_cycle;
      prev_point2_insn[context] = (unsigned) sim_num_insn_commit[context]+1; 
      break;
    case 24:
      total_point3_time[context] += ((unsigned) sim_cycle - prev_point2_time[context]);
      total_point3_insn[context] += ((unsigned) sim_num_insn_commit[context] - prev_point2_insn[context]);
      prev_point3_time[context] = (unsigned) sim_cycle;
      prev_point3_insn[context] = (unsigned) sim_num_insn_commit[context]+1; 
      break;
    case 32:
      total_point4_time[context] += ((unsigned) sim_cycle - prev_point3_time[context]);
      total_point4_insn[context] += ((unsigned) sim_num_insn_commit[context] - prev_point3_insn[context]);
      break;
    }
  }
}


/* RUU_DK: The following two macros are for finding a proper queue entry. */ 
/* find the next free(invalid) entry. 
   check if the queue is full before calling this macro! */
#define FIND_FREE_ENTRY(QUEUE, SIZE, POINTER)			\
  ({ int ptr = POINTER; while (QUEUE[ptr].valid == TRUE) ptr = (ptr+1) % SIZE; ptr; })  

/*
 *  RUU_COMMIT() - instruction retirement pipeline stage
 */

static int prev_cxt = 0; /* RUU_DK: record the last context id, which committed
			       in the previous cycle. In the current cycle, start
			       from (prev_cxt + 1) % context_num */ 
static int no_commit = 0; /* DK: Keep track of how many cycles are gone without committing 
			        any instruction. Terminate the simulation if hanging. */  

/* this function commits the results of the oldest completed entries from the
   RUU and LSQ to the architected reg file, stores in the LSQ will commit
   their store data to the data cache at this point as well */
static void
ruu_commit(void)
{
  int i, lat, events, committed = 0;
  SS_CONTEXT_TYPE context;
  int commit_locked[SS_NUM_CONTEXT];
  int context_locked = 0;
  struct RUU_station *rs;

  for(i = 0; i < SS_NUM_CONTEXT; i++)
    commit_locked[i] = FALSE;

  /* RUU_DK: start from (prev_cxt + 1) % context_num */
  context = (prev_cxt + 1) % context_num;

  /* all values must be retired to the architected reg file in program order */
  /* Nevertheless instructions taken from different contexts can be retired  */
  /* out of order.                                                           */

  /* RUU_DK: Before every 'continue;' in the 'if' statement, I should have 
     decremented RUU_count_cxt[] or LSQ_count_cxt[]. However, since the 
     context id is changed and locked, it is unnecessary. */

  while ((RUU_num > 0) &&
         (committed < ruu_commit_width) &&
         (context_locked < context_num))    /* If no context can be commited */
    {
      /* RUU_DK: no instruction to commit from this context. lock this context
	 and proceed to the next one! */	
      if (RUU_num_cxt[context] == 0) {
	commit_locked[context] = TRUE;
	context_locked++;
	context = (context + 1) % context_num;
	continue;
      }	

      rs = &(RUU[RUU_head[context]]);

      /* RUU_DK: following four 'if' statements */
      if (commit_locked[context]) {
	context = (context + 1) % context_num;
	continue;
      }	

      if (!rs->completed) {
        /* at least RUU entry must be complete */
        commit_locked[context] = TRUE;
        context_locked++;
	context = (context + 1) % context_num;
        continue;
      } /* if */

      /* default commit events */
      events = 0;
  
      /* load/stores must retire load/store queue entry as well */
      if (RUU[RUU_head[context]].ea_comp)
	{
	  /* load/store operation must be complete */
	  if (!LSQ[LSQ_head[context]].completed)
	    {
	      /* load/store operation is not yet complete */
              commit_locked[context] = TRUE;
              context_locked++;
	      /* RUU_DK */
	      context = (context + 1) % context_num;
              continue;
	    }

	  /* changed by dongkeun not to commit speculative STORE */
	  if ((SS_OP_FLAGS(LSQ[LSQ_head[context]].op) & (F_MEM|F_STORE))
	      == (F_MEM|F_STORE) && !LSQ[LSQ_head[context]].spec_mode)  
	    {
	      struct res_template *fu;

	      /* stores must retire their store value to the cache at commit,
		 try to get a store port (functional unit allocation) */
	      fu = res_get(fu_pool, SS_OP_FUCLASS(LSQ[LSQ_head[context]].op));
	      if (fu)
		{
		  /* schedule functional unit release event */
		  fu->master->busy = fu->issuelat;

		  /* go to the data cache */
		  if (cache_dl1[context])
		    {
		      /* commit store value to D-cache */
		      lat =
			cache_access(cache_dl1[context], Write,
                                     (LSQ[LSQ_head[context]].addr&~3),
				     mmu->config[context],
				     NULL, 4, sim_cycle, NULL, NULL);
		      if (lat > cache_dl1_lat)
			events |= PEV_CACHEMISS;
		    }

		  /* all loads and stores must to access D-TLB */
		  if (dtlb[context])
		    {
		      /* access the D-TLB */
		      lat =
			cache_access(dtlb[context], Read,
                                     (LSQ[LSQ_head[context]].addr & ~3),
				     mmu->config[context],
				     NULL, 4, sim_cycle, NULL, NULL);
		      if (lat > 1)
			events |= PEV_TLBMISS;
		    }
		}
	      else
		{
		  /* no store ports left, cannot continue to commit insts */
                  commit_locked[context] = TRUE;
                  context_locked++;
	          /* RUU_DK */
	          context = (context + 1) % context_num;
                  continue;
		}
	    }

	  /* invalidate load/store operation instance */
	  LSQ[LSQ_head[context]].tag++;
	  /* indicate to pipeline trace that this instruction retired */
	  ptrace_newstage(context, LSQ[LSQ_head[context]].ptrace_seq,
                          PST_COMMIT, events);
	  ptrace_endinst(context, LSQ[LSQ_head[context]].ptrace_seq);

	  /* commit head of LSQ as well */
          LSQ[LSQ_head[context]].retired = TRUE;
	  LSQ[LSQ_head[context]].valid = FALSE;
	  LSQ_num--;
	  /* LSQ_DK */
	  LSQ_num_cxt[context]--;

	  /* Manage the order of LSQ entry */
	  if (LSQ[LSQ_head[context]].next != -1) 
	    LSQ[LSQ[LSQ_head[context]].next].prev = LSQ[LSQ_head[context]].prev;
	   
	  if (LSQ[LSQ_head[context]].prev != -1) 
	    LSQ[LSQ[LSQ_head[context]].prev].next = LSQ[LSQ_head[context]].next;

	  if (LSQ_head[context] == LSQ_oldest)
	    LSQ_oldest = LSQ[LSQ_head[context]].next;

	  if (LSQ_head[context] == prev_entry)
	    prev_entry = LSQ[LSQ_head[context]].prev;

	  if (LSQ_num_cxt[context] != 0) {
	    LSQ_head[context] =
	      lsq_table[context][(LSQ[LSQ_head[context]].my_seq+1) % LSQ_size];
	  } else {
	    LSQ_head[context] = (LSQ_head[context] + 1) % LSQ_size;
	  }
	}

      if (pred[context]
	  && bpred_spec_update == spec_CT
	  && (SS_OP_FLAGS(rs->op) & F_CTRL))
	{
	  SS_INST_TYPE inst = rs->IR;

	  bpred_update(pred[context],
		       retstack[context],
		       context,
		       /* branch address */rs->PC,
		       /* actual target address */rs->next_PC,
                       /* taken? */rs->next_PC != (rs->PC +
                                                   sizeof(SS_INST_TYPE)),
                       /* pred taken? */rs->pred_PC != (rs->PC +
                                                        sizeof(SS_INST_TYPE)),
                       /* correct pred? */rs->pred_PC == rs->next_PC,
                       /* opcode */rs->op,
		       /* jump through R31? */(RS) == 31,
                       /* dir predictor update pointer */&rs->dir_update);
	}

      /* invalidate RUU operation instance */

      RUU[RUU_head[context]].tag++;

      /* indicate to pipeline trace that this instruction retired */
      ptrace_newstage(context, RUU[RUU_head[context]].ptrace_seq, PST_COMMIT, events);
      ptrace_endinst(context, RUU[RUU_head[context]].ptrace_seq);
      /* DY:  debug marker instruction support */
      check_marker_instruction(RUU_head[context], context);

      /* commit head entry of RUU */
      RUU[RUU_head[context]].retired = TRUE;
      RUU[RUU_head[context]].valid = FALSE;
      RUU_num--;
      /* RUU_DK */
      RUU_num_cxt[context]--;
      sim_num_insn_commit[context]++;

      if (RUU_num_cxt[context] != 0) {
	RUU_head[context] =
	  ruu_table[context][(RUU[RUU_head[context]].my_seq+1) % RUU_size];
      } else {
	RUU_head[context] = (RUU_head[context] + 1) % RUU_size;
	commit_locked[context] = TRUE;
	context_locked++;
	context = (context + 1) % context_num;
      }
 	
      /* one more instruction committed to architected state */
      committed++;
    }

  if (committed == 0) {
    if (no_commit++ == 10000)
      panic("The simulation is hanging!"); 
  } else {
    no_commit = 0;
  }

  /* We can suppress all entries already retired until we find          */
  /* the first not retired instruction.                                 */

  /* RUU_DK: record prev_cxt. if (context_locked == context_num), 
     we didn't have chance to commit instructions in this context. Go back! */
  if (context_locked == context_num)
    prev_cxt = (context + (context_num - 1)) % context_num;
  else 
    prev_cxt = context;
}



/*
 *  RUU_RECOVER() - squash mispredicted microarchitecture state
 */

#define RUU_STRIP_OP_ALL        1
#define RUU_STRIP_OP_IDLE       2

void recover_create_map(struct RUU_station *rs,
			SS_CONTEXT_TYPE context)
{
  int i;
  struct CV_link *ent;

  /* Correcting create_vector entries with new values */
  for (i = 0; i < SS_TOTAL_REGS; i++) {
    ent = &create_vector[(SS_TOTAL_REGS * context) + i];
    if (ent->rs == rs)
      create_vector[(SS_TOTAL_REGS * context) + i] = CVLINK_NULL;
    ent = &spec_create_vector[(SS_TOTAL_REGS * context) + i];
    if (ent->rs == rs)
      spec_create_vector[(SS_TOTAL_REGS * context) + i] = CVLINK_NULL;
  } /* for */
}


/* recover processor microarchitecture state back to point of the
   mis-predicted branch at RUU[BRANCH_INDEX] */
static void
ruu_recover(int branch_index,           /* index of mis-pred branch */
            SS_CONTEXT_TYPE context,    /* context of the mis-pred branch */
            unsigned int operation)     /* apply on idle RU or all
                                           (see RUU_STRIP_OP_*) */
{
  /* RUU_DK: RUU_tail is now changed to RUU_tail[context] */
  int i, RUU_index = RUU_tail[context], LSQ_index = LSQ_tail[context];

  /* recover from the tail of the RUU towards the head until the branch index
     is reached, this direction ensures that the LSQ can be synchronized with
     the RUU */

  /* go to first element to squash */
  RUU_index = (RUU_index + (RUU_size-1)) % RUU_size;
  LSQ_index = (LSQ_index + (LSQ_size-1)) % LSQ_size;

  /* traverse to older insts until the mispredicted branch is encountered */
  /* DK: If ruu_recover() is called from ruu_issue() with the operation 
	RUU_STRIP_OP_IDLE, the head should be squashed too if it is idle. */

/*  while ((RUU_index != branch_index) && RUU_num_cxt[context])*/
  while (RUU_num_cxt[context] && ((operation == RUU_STRIP_OP_IDLE) ||
	((operation == RUU_STRIP_OP_ALL) && (RUU_index != branch_index))))
    {
      /* the RUU should not drain since the mispredicted branch will remain */
      if (!RUU_num)
	panic("empty RUU");

      /* should meet up with the tail first */
/*      if (RUU_index == RUU_head[context])*/
      if ((operation == RUU_STRIP_OP_ALL) && (RUU_index == RUU_head[context]))
	panic("RUU head and tail broken");

      if ((RUU[RUU_index].context == context) &&
	  ((operation == RUU_STRIP_OP_ALL) ||
	   (((RUU[RUU_index].idle == TRUE) || kill_mode[context]) &&
/* DK: For kill operation, all instructions should be squashed. */
/*	   ((RUU[RUU_index].idle == TRUE) &&*/
	    (operation == RUU_STRIP_OP_IDLE)))) {

        /* is this operation an effective addr calc for a load or store? */
        if (RUU[RUU_index].ea_comp)
          {
            /* should be at least one load or store in the LSQ */
            if (!LSQ_num_cxt[context])
              panic("RUU and LSQ out of sync");
            
	    if ((LSQ[LSQ_index].context == context) &&
		((operation == RUU_STRIP_OP_ALL) ||
		 (((LSQ[LSQ_index].idle == TRUE) || kill_mode[context]) &&
/* DK: For kill operation, all instructions should be squashed. */
/*		 ((LSQ[LSQ_index].idle == TRUE) &&*/
		  (operation == RUU_STRIP_OP_IDLE)))) {

	      /* LSQ_DK: lsq entry with this context id shouldn't be invalid */
	      if (LSQ[LSQ_index].valid == FALSE)
		panic("LSQ entry with this context id shoule be valid!");

	      /* recover any resources consumed by the load or store operation */
	      for (i=0; i<MAX_ODEPS; i++) {
		RSLINK_FREE_LIST(LSQ[LSQ_index].odep_list[i]);
		/* blow away the consuming op list */
		LSQ[LSQ_index].odep_list[i] = NULL;
	      } /* for */
            
	      /* squash this LSQ entry */
	      LSQ[LSQ_index].tag++;

	      /* squash entries in the create map */
	      recover_create_map(&LSQ[LSQ_index], context);

	      /* indicate in pipetrace that this instruction was squashed */
	      ptrace_endinst(context, LSQ[LSQ_index].ptrace_seq);

	      /* go to next earlier LSQ slot */
	      LSQ_num--;

	      /* Manage the order of LSQ entry */
	      if (LSQ[LSQ_index].next != -1) 
	        LSQ[LSQ[LSQ_index].next].prev = LSQ[LSQ_index].prev;

	      if (LSQ[LSQ_index].prev != -1) 
	        LSQ[LSQ[LSQ_index].prev].next = LSQ[LSQ_index].next;

 	      if (LSQ_index == LSQ_oldest)
	        LSQ_oldest = LSQ[LSQ_index].next;

	      if (LSQ_index == prev_entry)
	        prev_entry = LSQ[LSQ_index].prev;

	      /* LSQ_DK */
	      LSQ_num_cxt[context]--;
	      LSQ[LSQ_index].valid = FALSE;
	    }

	    if (LSQ_num_cxt[context])
	      LSQ_index = lsq_table[context][(LSQ[LSQ_index].my_seq+(LSQ_size-1)) % LSQ_size];
          }
        
        /* recover any resources used by this RUU operation */
        for (i=0; i<MAX_ODEPS; i++) {
          RSLINK_FREE_LIST(RUU[RUU_index].odep_list[i]);
          /* blow away the consuming op list */
          RUU[RUU_index].odep_list[i] = NULL;
        } /* for */

        /* squash this RUU entry */
        RUU[RUU_index].tag++;
      
	/* squash entries in the create map */
	recover_create_map(&RUU[RUU_index], context);

        /* indicate in pipetrace that this instruction was squashed */
        ptrace_endinst(context, RUU[RUU_index].ptrace_seq);

        RUU_num--;
	/* RUU_DK */
	RUU_num_cxt[context]--;
	RUU[RUU_index].valid = FALSE;

      } /* if */

      /* Added by DK to allow the idle head instruction to be squashed. */      
      if ((operation == RUU_STRIP_OP_IDLE) && (RUU_index == branch_index))
	break;

      /* go to next earlier slot in the RUU */
      /* RUU_DK: go to the previous entry! */
      if (RUU_num_cxt[context])
	RUU_index = ruu_table[context][(RUU[RUU_index].my_seq+(RUU_size-1)) % RUU_size];
    } /* while */

    /* DK: After squashing instructions that belong to the killed context, 
           reset the kill_mode array element. */
    if ((operation == RUU_STRIP_OP_IDLE) && kill_mode[context]) 
      kill_mode[context] = 0;

    /* reset head/tail pointers to point to the mis-predicted branch */
    /* RUU_DK: tail pointer is adjusted here! */ 
    RUU_tail[context] = RUU_head[context];
    if (RUU_num_cxt[context] != 0) {
      for (i = 0; i < RUU_num_cxt[context] - 1; i++)
	RUU_tail[context] = ruu_table[context][(RUU[RUU_tail[context]].my_seq + 1) % RUU_size];
      RUU_sequence[context] = (RUU[RUU_tail[context]].my_seq + 1) % RUU_size;
      RUU_tail[context] = (RUU_tail[context] + 1) % RUU_size;
    }

    /* LSQ_DK: tail pointer is adjusted here! */
    LSQ_tail[context] = LSQ_head[context];
    if (LSQ_num_cxt[context] != 0) {
      for (i = 0; i < LSQ_num_cxt[context] - 1; i++)
        LSQ_tail[context] = lsq_table[context][(LSQ[LSQ_tail[context]].my_seq + 1) % LSQ_size];
      LSQ_sequence[context] = (LSQ[LSQ_tail[context]].my_seq + 1) % LSQ_size;
      LSQ_tail[context] = (LSQ_tail[context] + 1) % LSQ_size;
    }

    /* revert create vector back to last precise create vector state, NOTE:
       this is accomplished by resetting all the copied-on-write bits in the
       USE_SPEC_CV bit vector */
    
    /* FIXMEDM: This part might be not executed when operation applies
       to idle RU. */
    if (operation == RUU_STRIP_OP_ALL)
      for(i = 0; i < SS_TOTAL_REGS; i++)
        BITMAP_CLEAR(use_spec_cv, CV_BMAP_SZ, i + (context * SS_TOTAL_REGS));

    /* FIXME: could reset functional units at squash time */

} /* ruu_recover */


/*
 *  RUU_WRITEBACK() - instruction result writeback pipeline stage
 */


/* forward declarations */
static void tracer_recover(SS_CONTEXT_TYPE context);
void fetch_dump(FILE *stream);	     

/* writeback completed operation results from the functional units to RUU,
   at this point, the output dependency chains of completing instructions
   are also walked to determine if any dependent instruction now has all
   of its register operands, if so the (nearly) ready instruction is inserted
   into the ready instruction queue */
static void
ruu_writeback(void)
{
  int i;
  struct RUU_station *rs;

  /* service all completed events */
  while ((rs = eventq_next_event()))
    {
      SS_CONTEXT_TYPE context = rs->context;    /* extracting context   */
      current_context = context;                /* global update        */
      /* operation has completed */
      rs->completed = TRUE;

      /* does this operation reveal a mis-predicted branch? */
      if (rs->recover_inst)
	{
	  /* recover processor state and reinit fetch to correct path */
	  ruu_recover(rs - RUU, context, RUU_STRIP_OP_ALL);
	  tracer_recover(context);
	  retstack_recover(retstack[context], rs->PC, rs->stack_recover_idx);
	  /* stall fetch until I-fetch and I-decode recover */
	  ruu_fetch_issue_delay[context] = ruu_branch_penalty;

	  /* continue writeback of the branch/control instruction */
	}

      /* if we speculatively update branch-predictor, do it here */
      if (pred[context]
	  && bpred_spec_update == spec_WB
	  && !rs->in_LSQ
	  && (SS_OP_FLAGS(rs->op) & F_CTRL))
	{
	  SS_INST_TYPE inst = rs->IR;

	  bpred_update(pred[context],
		       retstack[context],
		       context,
		       /* branch address */rs->PC,
		       /* actual target address */rs->next_PC,
		       /* taken? */rs->next_PC != (rs->PC +
						   sizeof(SS_INST_TYPE)),
		       /* pred taken? */rs->pred_PC != (rs->PC +
							sizeof(SS_INST_TYPE)),
		       /* correct pred? */rs->pred_PC == rs->next_PC,
		       /* opcode */rs->op,
		       /* jump through R31? */(RS) == 31,
		       /* dir predictor update pointer */&rs->dir_update);
	}

      /* entered writeback stage, indicate in pipe trace */
      ptrace_newstage(context, rs->ptrace_seq, PST_WRITEBACK,
		      rs->recover_inst ? PEV_MPDETECT : 0);

      /* broadcast results to consuming operations, this is more efficiently
         accomplished by walking the output dependency chains of the
	 completed instruction */
      for (i=0; i<MAX_ODEPS; i++)
	{
	  if (rs->onames[i] != NA)
	    {
	      struct CV_link link;
	      struct RS_link *olink, *olink_next;

	      if (rs->spec_mode)
		{
		  /* update the speculative create vector, future operations
		     get value from later creator or architected reg file */
		  link = spec_create_vector[rs->onames[i]];
		  if (/* !NULL */link.rs
		      && /* refs RS */(link.rs == rs && link.odep_num == i))
		    {
		      /* the result can now be read from a physical register,
			 indicate this as so */
		      spec_create_vector[rs->onames[i]] = CVLINK_NULL;
		      spec_create_vector_rt[rs->onames[i]] = sim_cycle;
		    }
		  /* else, creator invalidated or there is another creator */
		}
	      else
		{
		  /* update the non-speculative create vector, future
		     operations get value from later creator or architected
		     reg file */
		  link = create_vector[rs->onames[i]];
		  if (/* !NULL */link.rs
		      && /* refs RS */(link.rs == rs && link.odep_num == i))
		    {
		      /* the result can now be read from a physical register,
			 indicate this as so */
		      create_vector[rs->onames[i]] = CVLINK_NULL;
		      create_vector_rt[rs->onames[i]] = sim_cycle;
		    }
		  /* else, creator invalidated or there is another creator */
		}

	      /* walk output list, queue up ready operations */
	      for (olink=rs->odep_list[i]; olink; olink=olink_next)
		{
		  if (RSLINK_VALID(olink))
		    {
		      /* input is now ready */
		      olink->rs->idep_ready[olink->x.opnum] = TRUE;

		      /* are all the register operands of target ready? */
		      if (OPERANDS_READY(olink->rs))
			{
			  /* yes! enqueue instruction as ready, NOTE: stores
			     complete at dispatch, so no need to enqueue
			     them */
			  if (!olink->rs->in_LSQ
			      || ((SS_OP_FLAGS(olink->rs->op)&(F_MEM|F_STORE))
				  == (F_MEM|F_STORE)))
			    readyq_enqueue(olink->rs);
			  /* else, ld op, issued when no mem conflict */
			}
		    }

		  /* grab link to next element prior to free */
		  olink_next = olink->next;

		  /* free dependence link element */
		  RSLINK_FREE(olink);
		}
	      /* blow away the consuming op list */
	      rs->odep_list[i] = NULL;

	    } /* if not NA output */
	} /* for all outputs */

   } /* for all writeback events */

}

/*
 *  LSQ_REFRESH() - memory access dependence checker/scheduler
 */

/* this function locates ready instructions whose memory dependencies have
   been satisfied, this is accomplished by walking the LSQ for loads, looking
   for blocking memory dependency condition (e.g., earlier store with an
   unknown address) */
#define MAX_STD_UNKNOWNS		64
static void
lsq_refresh(void)
{
  int i, j, index, n_std_unknowns;
  SS_ADDR_TYPE std_unknowns[MAX_STD_UNKNOWNS];
  SS_CONTEXT_TYPE std_unknowns_context[MAX_STD_UNKNOWNS];

  /* scan entire queue for ready loads: scan from oldest instruction
     (head) until we reach the tail or an unresolved store, after which no
     other instruction will become ready */

  for (i=0, index = LSQ_oldest, n_std_unknowns=0;
       i < LSQ_num;
       i++, index = LSQ[index].next)
    {
      /* terminate search for ready loads after first unresolved store,
	 as no later load could be resolved in its presence */

      if (/* store? */
	  (SS_OP_FLAGS(LSQ[index].op) & (F_MEM|F_STORE)) == (F_MEM|F_STORE))
	{
	  if (!STORE_ADDR_READY(&LSQ[index]))
	    {
	      /* FIXME: a later STD + STD known could hide the STA unknown */
	      /* sta unknown, blocks all later loads, stop search */
	      break;
	    }
	  else if (!OPERANDS_READY(&LSQ[index]))
	    {
	      /* sta known, but std unknown, may block a later store, record
		 this address for later referral, we use an array here because
		 for most simulations the number of entries to search will be
		 very small */
	      if (n_std_unknowns == MAX_STD_UNKNOWNS)
		fatal("STD unknown array overflow, increase MAX_STD_UNKNOWNS");
	      std_unknowns[n_std_unknowns] = LSQ[index].addr;
	      std_unknowns_context[n_std_unknowns++] = LSQ[index].context;
	    }
	  else /* STORE_ADDR_READY() && OPERANDS_READY() */
	    {
	      /* a later STD known hides an earlier STD unknown */
	      for (j=0; j<n_std_unknowns; j++)
		{
		  if ((std_unknowns[j] == /* STA/STD known */LSQ[index].addr) &&
                      (std_unknowns_context[j] == LSQ[index].context))
		    std_unknowns[j] = /* bogus addr */0;
		}
	    }
	}

      if (/* load? */
	  ((SS_OP_FLAGS(LSQ[index].op) & (F_MEM|F_LOAD)) == (F_MEM|F_LOAD))
	  && /* queued? */!LSQ[index].queued
	  && /* waiting? */!LSQ[index].issued
	  && /* completed? */!LSQ[index].completed
	  && /* regs ready? */OPERANDS_READY(&LSQ[index]))
	{
	  /* no STA unknown conflict (because we got to this check), check for
	     a STD unknown conflict */
	  for (j=0; j<n_std_unknowns; j++)
	    {
	      /* found a relevant STD unknown? */
	      if ((std_unknowns[j] == LSQ[index].addr) &&
                  (std_unknowns_context[j] == LSQ[index].context))
		break;
	    }
	  if (j == n_std_unknowns)
	    {
	      /* no STA or STD unknown conflicts, put load on ready queue */
	      readyq_enqueue(&LSQ[index]);
	    }
	}
    }
}

/*
 *  RUU_ISSUE() - issue instructions to functional units
 */


#ifdef CACHE_PROF
#define MAX_CACHE_MISS_PROF 2000
struct cache_miss_prof_ {
  unsigned PC;
  int lat;
  int count;
  int L1_miss;
  int L2_miss;
  int partial_miss;
  int unknown_miss;
  int non_loop_miss;
} cache_miss_prof[MAX_CACHE_MISS_PROF];
static int cache_miss_num = 0;

typedef struct issued_addr_prof_ {
  int valid;
  int count;
  unsigned cycle;
  unsigned addr;
} issued_addr_prof_t;

static issued_addr_prof_t main_addr_prof[MAX_CACHE_MISS_PROF]; 
static issued_addr_prof_t issued_addr_prof[MAX_CACHE_MISS_PROF]; 

static int issued_addr_num = 0;
static int covered_addr = 0;
static int covered_L1_miss = 0;
static int covered_L2_miss = 0;

static int main_addr_num = 0;
static int late_addr = 0;
static int late_L1_miss = 0;
static int late_L2_miss = 0;

static int miss_class; /* 0: Unknown, 1: Partial */

static int non_loop_L1_miss = 0;
static int non_loop_L2_miss = 0;

int find_oldest_entry(issued_addr_prof_t *table_addr)
{
  int i, prev_index;
  unsigned prev_cycle = (unsigned) sim_cycle;

  for (i = 0; i < MAX_CACHE_MISS_PROF; i++)
    if (table_addr[i].cycle < prev_cycle) {
      prev_cycle = table_addr[i].cycle;
      prev_index = i;
    }

  return prev_index;
}

int get_new_index(issued_addr_prof_t *table_addr, int mode)
{
  int index, num_entry;

  if (mode)
    num_entry = issued_addr_num;
  else 
    num_entry = main_addr_num;

  if (num_entry == MAX_CACHE_MISS_PROF) {
    index = find_oldest_entry(table_addr);
    table_addr[index].valid = 0;
    if (mode)
      issued_addr_num--;
    else
      main_addr_num--;
    return index;
  }

  for (index = 0; index < MAX_CACHE_MISS_PROF; index++)
    if (!table_addr[index].valid)
      return index;
}

void register_issued_addr(unsigned addr, issued_addr_prof_t *table_addr, int mode)
{
  int index, valid_entry;

  if (mode)
    valid_entry = issued_addr_num;
  else 
    valid_entry = main_addr_num;

  for (index = 0; index < MAX_CACHE_MISS_PROF && valid_entry; index++) 
    if (table_addr[index].valid) {
      valid_entry--;
      if (table_addr[index].addr == addr) {
        table_addr[index].count++;
        table_addr[index].cycle = (unsigned) sim_cycle;
	return;
      }
    }
  
  /* Register new entry! */
  index = get_new_index(table_addr, mode);
  table_addr[index].valid = 1;
  table_addr[index].count = 1;
  table_addr[index].cycle = (unsigned) sim_cycle;
  table_addr[index].addr = addr;
  if (mode)
    issued_addr_num++;
  else 
    main_addr_num++;
}

int search_issued_addr_table(unsigned addr, int lat, issued_addr_prof_t *table_addr, int mode)
{
  int i, valid_entry;

  if (mode)
    valid_entry = issued_addr_num;
  else
    valid_entry = main_addr_num;

  for (i = 0; i < MAX_CACHE_MISS_PROF && valid_entry; i++)
    if (table_addr[i].valid) {
      valid_entry--;
      if (table_addr[i].addr == addr) {
	if (mode)
	  covered_addr++;
	else 
	  late_addr++;
	table_addr[i].count--;

	if ((lat < mem_lat[0]) && (lat > cache_dl1_lat)) {
	  if (mode) {
	    miss_class = 1;
	    covered_L1_miss++;
	  } else {
	    late_L1_miss++;
	  }
	} else if (lat >= mem_lat[0]) { 
	  if (mode) {
	    miss_class = 1;
	    covered_L2_miss++;
	  } else {
	    late_L2_miss++;
	  }
	}

	if (!table_addr[i].count) {
	  table_addr[i].valid = 0;
	  if (mode)
	    issued_addr_num--;
	  else 
	    main_addr_num--;
	}

	return TRUE;
      }
    }

  return FALSE;
}

void register_cache_miss(unsigned PC, int lat, int context)
{
  int i;

  if (last_mark[context] == 2) {
    if (print_cache_miss_events)
      fprintf(stderr, "     CACHE MISS(%x, %d): %10u\n",
	      PC, context, (unsigned)sim_cycle);

    for (i = 0; i < cache_miss_num; i++) {
      if (cache_miss_prof[i].PC == PC)
	break;
    }
    if (i == cache_miss_num) {
      /* PC hasn't been seen before */
      cache_miss_prof[i].PC = PC;
      cache_miss_prof[i].lat = lat;
      if (lat < mem_lat[0]) { /* L1 miss! */
        cache_miss_prof[i].L1_miss = 1;
	cache_miss_prof[i].L2_miss = 0;
	cache_miss_prof[i].partial_miss = 0;
	cache_miss_prof[i].unknown_miss = 0;
	cache_miss_prof[i].non_loop_miss = 0;
	if (!px_mode) {
	  cache_miss_prof[i].non_loop_miss++;
	  non_loop_L1_miss++;
	} else {
	  if (miss_class == 0)
	    cache_miss_prof[i].unknown_miss++;
	  else if (miss_class == 1)
	    cache_miss_prof[i].partial_miss++;
	}
      } else { /* L2 miss! */
        cache_miss_prof[i].L1_miss = 0;
	cache_miss_prof[i].L2_miss = 1;
	cache_miss_prof[i].partial_miss = 0;
	cache_miss_prof[i].unknown_miss = 0;
	cache_miss_prof[i].non_loop_miss = 0;
	if (!px_mode) {
	  cache_miss_prof[i].non_loop_miss++;
	  non_loop_L2_miss++;
	} else {
	  if (miss_class == 0)
	    cache_miss_prof[i].unknown_miss++;
	  else if (miss_class == 1)
	    cache_miss_prof[i].partial_miss++;
	}
      }
      cache_miss_prof[i].count = 1;
      cache_miss_num++;
      if (cache_miss_num == MAX_CACHE_MISS_PROF)
	panic("ran out of cache miss profile elements!");
    } else {
      /* found PC */
      cache_miss_prof[i].lat += lat;
      if (lat < mem_lat[0]) { /* L1 miss! */
        cache_miss_prof[i].L1_miss++;
	if (!px_mode) {
	  cache_miss_prof[i].non_loop_miss++;
	  non_loop_L1_miss++;
	} else {
	  if (miss_class == 0)
	    cache_miss_prof[i].unknown_miss++;
	  else if (miss_class == 1)
	    cache_miss_prof[i].partial_miss++;
	}
      } else { /* L2 miss! */
	cache_miss_prof[i].L2_miss++;
	if (!px_mode) {
	  cache_miss_prof[i].non_loop_miss++;
	  non_loop_L2_miss++;
	} else {
	  if (miss_class == 0)
	    cache_miss_prof[i].unknown_miss++;
	  else if (miss_class == 1)
	    cache_miss_prof[i].partial_miss++;
	}
      }
      cache_miss_prof[i].count++;
    }
  } /* else don't record */
}
#endif

static unsigned int issue_context_op_seq = 0;
static SS_WORD_TYPE issue_context_op_mask = 0;

/* forward declarations */

static void fetch_queue_strip(SS_CONTEXT_TYPE context);
static SS_ADDR_TYPE fetch_pred_PC[SS_NUM_CONTEXT];

/* attempt to issue all operations in the ready queue; insts in the ready
   instruction queue have all register dependencies satisfied, this function
   must then 1) ensure the instructions memory dependencies have been satisfied
   (see lsq_refresh() for details on this process) and 2) a function unit
   is available in this cycle to commence execution of the operation; if all
   goes well, the function unit is allocated, a writeback event is scheduled,
   and the instruction begins execution */
static void
ruu_issue(void)
{
  int i, load_lat, tlb_lat, n_issued;
  struct RS_link *node, *next_node;
  struct res_template *fu;
  SS_CONTEXT_TYPE context;
  SS_WORD_TYPE start_mask;
  SS_WORD_TYPE stop_mask;
#ifdef CACHE_PROF
  int addr_found;
#endif

  /* FIXME: could be a little more efficient when scanning the ready queue */

  /* copy and then blow away the ready list, NOTE: the ready list is
     always totally reclaimed each cycle, and instructions that are not
     issue are explicitly reinserted into the ready instruction queue,
     this management strategy ensures that the ready instruction queue
     is always properly sorted */

  node = ready_queue;
  ready_queue = NULL;

  /* visit all ready instructions (i.e., insts whose register input
     dependencies have been satisfied, stop issue when no more instructions
     are available or issue bandwidth is exhausted */
  for (n_issued=0;
       node && n_issued < ruu_issue_width;
       node = next_node)
    {
      next_node = node->next;

      /* still valid? */
      if (RSLINK_VALID(node))
	{
	  struct RUU_station *rs = RSLINK_RS(node);

          context = rs->context;
          current_context = context;    /* Mandatory for memory access */
	  /* node is now un-queued */
	  rs->queued = FALSE;

/*            if (rs->idle) { */
/*              readyq_enqueue(rs); */
/*            } else  */
            
	  if (rs->in_LSQ
	      && ((SS_OP_FLAGS(rs->op) & (F_MEM|F_STORE)) == (F_MEM|F_STORE)))
	    {
	      /* stores complete in effectively zero time, result is
		 written into the load/store queue, the actual store into
		 the memory system occurs when the instruction is retired
		 (see ruu_commit()) */
	      rs->issued = TRUE;
              rs->completed = TRUE;
	      /* entered execute stage, indicate in pipe trace */
	      ptrace_newstage(context, rs->ptrace_seq, PST_WRITEBACK, 0);

 	      /* one more inst issued */
              if (!rs->idle) {
                n_issued++;
	      }
	    }
	  else
	    {
	      /* issue the instruction to a functional unit */
	      if (SS_OP_FUCLASS(rs->op) != NA)
		{
		  fu = res_get(fu_pool, SS_OP_FUCLASS(rs->op));
		  if (fu)
		    {
		      /* got one! issue inst to functional unit */
		      rs->issued = TRUE;
		      /* schedule functional unit release event */
		      fu->master->busy = fu->issuelat;

		      /* schedule a result writeback event */
		      if (rs->in_LSQ
			  && ((SS_OP_FLAGS(rs->op) & (F_MEM|F_LOAD))
			      == (F_MEM|F_LOAD)))
			{
			  int events = 0;
			
			  /* for loads, determine cache access latency:
			     first scan LSQ to see if a store forward is
			     possible, if not, access the data cache */
			  load_lat = 0;
			  i = (rs - LSQ);
			  if (i != LSQ_head[context]) /* LSQ_DK */
			    {
			      for (;;)
				{
				  /* LSQ_DK: go to next earlier LSQ entry */
				  i = lsq_table[context][(LSQ[i].my_seq+(LSQ_size-1)) % LSQ_size];
/*				  i = (i + (LSQ_size-1)) % LSQ_size; */

				  /* FIXME: not dealing with partials! */
				  if ((SS_OP_FLAGS(LSQ[i].op) & F_STORE)
				      && (LSQ[i].addr == rs->addr))
				    {
				      /* hit in the LSQ */
				      load_lat = 1;
				      break;
				    }

				  /* LSQ_DK: scan finished? */
				  if (i == LSQ_head[context])
				    break;
				}
			    }

			  /* was the value store forwared from the LSQ? */
			  if (!load_lat)
			    {
			      /* no! go to the data cache */
			      if (cache_dl1[context]
				  /* valid address? */
				  && (rs->addr >= ld_data_base[context]
				      && rs->addr < ld_stack_base[context]))
				{
				  /* access the cache if non-faulting */
				  load_lat =
				    cache_access(cache_dl1[context], Read,
						 (rs->addr & ~3),
						 mmu->config[context],
						 NULL, 4,
						 sim_cycle, NULL, NULL);
/* DK: Used to find covered cache misses. Activate the following if statement and 
       register_issued_addr() below to measure the number of covered cache misses.
       Turn this feature off for normal simulation since this slows down the 
       simulation by more than factor of two. */ 
#ifdef CACHE_PROF
				  miss_class = 0; /* reset miss_class */

				  if (context_master[context]) 
				    addr_found = search_issued_addr_table((rs->addr & ~0x1f), load_lat, issued_addr_prof, 1);
				  else   
				    addr_found = search_issued_addr_table((rs->addr & ~0x1f), load_lat, main_addr_prof, 0);
#endif
				  if (load_lat > cache_dl1_lat) {
				    events |= PEV_CACHEMISS;
#ifdef CACHE_PROF
				    if (context_master[context]) { 
				      register_cache_miss(rs->PC, load_lat, rs->context);
				      if (!addr_found)
					register_issued_addr(rs->addr & ~0x1f, main_addr_prof, 0);
				    } else {
				      if (!addr_found)
			    		register_issued_addr(rs->addr & ~0x1f, issued_addr_prof, 1);
				    }
#endif
				  }
				}
			      else
				{
				  /* no caches defined, just use op latency */
				  load_lat = fu->oplat;
				}
			    }


			  /* all loads and stores must to access D-TLB */
			  if (dtlb[context]
			      /* valid address? */
			      && (rs->addr >= ld_data_base[context]
				  && rs->addr < ld_stack_base[context]))
			    {
			      /* access the D-DLB, NOTE: this code will
				 initiate speculative TLB misses */
			      tlb_lat =
				cache_access(dtlb[context], Read,
                                             (rs->addr & ~3),
					     mmu->config[context],
					     NULL, 4, sim_cycle, NULL, NULL);
			      if (tlb_lat > 1)
				events |= PEV_TLBMISS;

			      /* D-cache/D-TLB accesses occur in parallel */
			      load_lat = MAX(tlb_lat, load_lat);
			    }

			  if (rs->IR.a & 0x00010000) {
			    eventq_queue_event(rs, sim_cycle + 1);
			  } else { 
			    /* use computed cache access latency */
			    eventq_queue_event(rs, sim_cycle + load_lat);
			  }

			  /* entered execute stage, indicate in pipe trace */
			  ptrace_newstage(context,
                                          rs->ptrace_seq, PST_EXECUTE,
					  ((rs->ea_comp ? PEV_AGEN : 0)
					   | events));
			}
		      else if ((SS_OP_FLAGS(rs->op) & (F_CONTEXT))
                               == (F_CONTEXT)) {
                        unsigned int i = 0, context_op_lat = fu->oplat;
                        SS_WORD_TYPE mask;

                        if (rs->target_op) {
                          if (issue_context_op_seq <= rs->context_op_seq) {
                            
                            start_mask = rs->target_mask &
                              (rs->target_mask ^ sim_active_context[0]);
                            stop_mask = sim_active_context[0] &
                              (rs->target_mask ^ sim_active_context[0]);
                            issue_context_op_mask = rs->target_mask;
                            
                            mask = stop_mask;
                            while (mask) {
                              if (mask & 0x00000001) ++context_op_lat;
                              mask = mask >> 1;
                            }
                            /* FIXMEDM: should be at the end of the 
                               bloc but because of ruu_strip, rs is
                               modified... */
                            eventq_queue_event(rs, sim_cycle + context_op_lat);
                            ptrace_newstage(context,
                                            rs->ptrace_seq, PST_EXECUTE, 
                                            rs->ea_comp ? PEV_AGEN : 0);
                        
                            while (stop_mask | start_mask) {
                              if (stop_mask & 0x00000001) {
                                fetch_queue_strip(i);
                                /* FIXMEDM: RUU strip must take place here.
                                 * the ruu_recover() might be to "heavy" for
                                 * this work but I haven't checked it yet, *
                                 * ruu_recover(rs - RUU,  */
				/* RUU_DK: RUU_head -> RUU_head[i] */
                                ruu_recover(RUU_head[i], i, RUU_STRIP_OP_IDLE);
				ruu_fetch_issue_delay[i] = ruu_branch_penalty;
                                /* Check if the stop instruction has been
                                   decoded before a resume instruction */
                                if (BITMAP_SET_P(sim_restart_context,
                                                 SIM_RESTART_SZ, i)) {
                                  BITMAP_CLEAR(sim_restart_context,
                                               SIM_RESTART_SZ, i);
                                  BITMAP_SET(sim_active_context,
                                             SIM_ACTIVE_SZ, i);
                                  BITMAP_CLEAR(sim_shutdown_context,
                                               SIM_SHUTDOWN_SZ, i);
                                  fetch_pred_PC[i] = recover_PC[i] =
                                    regs_PC[i] = rbackup_data[i].regs_PC;
                                  spec_mode[i] = FALSE;
                                } else {
                                  BITMAP_CLEAR(sim_active_context,
                                               SIM_ACTIVE_SZ, i);
                                  BITMAP_CLEAR(sim_shutdown_context,
                                               SIM_SHUTDOWN_SZ, i);
                                } /* if */
                               /*   context_op_lat++; */
                              } else if (start_mask & 0x00000001) {
                                BITMAP_SET(sim_active_context,
                                           SIM_ACTIVE_SZ, i);                
                                BITMAP_CLEAR(sim_shutdown_context,
                                             SIM_SHUTDOWN_SZ, i);          
                                BITMAP_CLEAR(sim_restart_context,
                                             SIM_RESTART_SZ, i);          
                                fetch_pred_PC[i] = recover_PC[i] =
                                  regs_PC[i] = rbackup_data[i].regs_PC; /* reason of the problem */
                                spec_mode[i] = FALSE;
                              }/* if */
                              i++;
                              stop_mask = stop_mask >> 1;
                              start_mask = start_mask >> 1;
                            } /* while */
                            issue_context_op_seq = rs->context_op_seq;
                          } else { 
                            /* Is there contexts not restarted ? */
                            stop_mask = issue_context_op_mask &
                              (rs->target_mask ^ issue_context_op_mask);

                            mask = stop_mask;
                            while (mask) {
                              if (mask & 0x00000001) ++context_op_lat;
                              mask = mask >> 1;
                            }
                            /* FIXMEDM: should be at the end of the 
                               bloc but because of ruu_strip, rs is
                               modified... */
                            eventq_queue_event(rs, sim_cycle + context_op_lat);
                            ptrace_newstage(context,
                                            rs->ptrace_seq, PST_EXECUTE, 
                                            rs->ea_comp ? PEV_AGEN : 0);

                            while (stop_mask) {
                              if (stop_mask & 0x00000001) {
                                if (BITMAP_SET_P(sim_restart_context,
                                                 SIM_RESTART_SZ, i)) {
                                  BITMAP_CLEAR(sim_restart_context,
                                               SIM_RESTART_SZ, i);
                                  BITMAP_SET(sim_active_context,
                                             SIM_ACTIVE_SZ, i);
                                  BITMAP_CLEAR(sim_shutdown_context,
                                               SIM_SHUTDOWN_SZ, i);
                                  fetch_pred_PC[i] = recover_PC[i] =
                                    regs_PC[i] = rbackup_data[i].regs_PC;
                                  spec_mode[i] = FALSE;
                                  fetch_queue_strip(i);
				  /* RUU_DK: RUU_head -> RUU_head[i] */
                                  ruu_recover(RUU_head[i], i, RUU_STRIP_OP_IDLE);
				  ruu_fetch_issue_delay[i] = ruu_branch_penalty;
                                 /*   context_op_lat++; */
                                }
                              }
                              i++;
                              stop_mask = stop_mask >> 1;
                            } /* while */
                          } 
                        }
                      } else /* not a context operation */ {
                        
                        /* use deterministic functional unit latency */
                        eventq_queue_event(rs, sim_cycle + fu->oplat);
                        /* entered execute stage, indicate in pipe trace */
                        ptrace_newstage(context,
                                        rs->ptrace_seq, PST_EXECUTE, 
                                        rs->ea_comp ? PEV_AGEN : 0);
                        
                      }           

		      /* one more inst issued */
                      if (!rs->idle) {
                        n_issued++;
		      }	
		    }
		  else /* no functional unit */
		    {
		      /* insufficient functional unit resources, put operation
			 back onto the ready list, we'll try to issue it
			 again next cycle */
		      readyq_enqueue(rs);
		    }
		}
	      else /* does not require a functional unit! */
		{
		  /* FIXME: need better solution for these */
		  /* the instruction does not need a functional unit */
		  rs->issued = TRUE;

		  /* schedule a result event */
		  eventq_queue_event(rs, sim_cycle + 1);

		  /* entered execute stage, indicate in pipe trace */
		  ptrace_newstage(context,
                                  rs->ptrace_seq, PST_EXECUTE,
				  rs->ea_comp ? PEV_AGEN : 0);

		  /* one more inst issued */
                  if (!rs->idle) {
                    n_issued++;
		  }
		}
	    } /* !store */

	}
      /* else, RUU entry was squashed */

      /* reclaim ready list entry, NOTE: this is done whether or not the
         instruction issued, since the instruction was once again reinserted
         into the ready queue if it did not issue, this ensures that the ready
         queue is always properly sorted */
      RSLINK_FREE(node);
    }

  /* put any instruction not issued back into the ready queue, go through
     normal channels to ensure instruction stay ordered correctly */
  for (; node; node = next_node)
    {
      next_node = node->next;

      /* still valid? */
      if (RSLINK_VALID(node))
        {
	  struct RUU_station *rs = RSLINK_RS(node);

          /* node is now un-queued */
          rs->queued = FALSE;

	  /* not issued, put operation back onto the ready list, we'll try to
	     issue it again next cycle */
          readyq_enqueue(rs);
        } else {
      /* else, RUU entry was squashed */
        }
      /* reclaim ready list entry, NOTE: this is done whether or not the
         instruction issued, since the instruction was once again reinserted
         into the ready queue if it did not issue, this ensures that the ready
         queue is always properly sorted */

      RSLINK_FREE(node);
    }
}


/*
 * routines for generating on-the-fly instruction traces with support
 * for control and data misspeculation modeling
 */

/* integer register file */
#define R_BMAP_SZ (BITMAP_SIZE(SS_NUM_REGS*SS_NUM_CONTEXT))
static BITMAP_TYPE((SS_NUM_REGS*SS_NUM_CONTEXT), use_spec_R);
SS_WORD_TYPE spec_regs_R[SS_NUM_CONTEXT * SS_NUM_REGS];

/* floating point register file */
#define F_BMAP_SZ (BITMAP_SIZE(SS_NUM_REGS*SS_NUM_CONTEXT))
static BITMAP_TYPE((SS_NUM_REGS*SS_NUM_CONTEXT), use_spec_F);
static union regs_FP spec_regs_F;

/* miscellaneous registers */
static int use_spec_HI[SS_NUM_CONTEXT];
static SS_WORD_TYPE spec_regs_HI[SS_NUM_CONTEXT];
static int use_spec_LO[SS_NUM_CONTEXT];
static SS_WORD_TYPE spec_regs_LO[SS_NUM_CONTEXT];
static int use_spec_FCC[SS_NUM_CONTEXT];
static int spec_regs_FCC[SS_NUM_CONTEXT];

/* context management registers */

/* In fact there is just one regs_ACXT. But to make possible */
/* speculative writing in this global register we need to */
/* simulate a behavior similar to integer registers. */
static int use_spec_ACXT[SS_NUM_CONTEXT];
static SS_WORD_TYPE spec_regs_ACXT[SS_NUM_CONTEXT];
static int use_spec_CXTCTRL[SS_NUM_CONTEXT];
static SS_WORD_TYPE spec_regs_CXTCTRL[SS_NUM_CONTEXT];

/* dump speculative register state */
static void
rspec_dump(FILE *stream)			/* output stream */
{
  int i;
  SS_CONTEXT_TYPE context;

  fprintf(stream, "** speculative register contents **\n");

  /* dump speculative integer regs */
  for (context = 0; context < context_num; context++) {
    fprintf(stream, "spec_mode: %s\n", spec_mode[context] ? "t" : "f");
    for (i=0; i<SS_NUM_REGS; i++) {
      if (BITMAP_SET_P(use_spec_R, R_BMAP_SZ, i + (SS_NUM_REGS * context))) {
        /* speculative state */
        fprintf(stream, "[%4s][%02d]: %12d/0x%08x\n", dep_names[i], context,
                spec_regs_R[(context << 5) + i],
                spec_regs_R[(context << 5) + i]);
      } /* if */
    } /* for */

    /* dump speculative FP regs */
    for (i=0; i<SS_NUM_REGS; i++) {
      if (BITMAP_SET_P(use_spec_F, F_BMAP_SZ, i + (SS_NUM_REGS * context))) {
        /* speculative state */
        fprintf(stream,
                "[%4s][%02d]: %12d/0x%08x/%f ([%4s] as double: %f)\n\n",
                dep_names[i+32], 
                context, 
                spec_regs_F.l[context][i],
                spec_regs_F.l[context][i],
                spec_regs_F.f[context][i],
                dep_names[i+32],
                spec_regs_F.d[context][i >> 1]);
      } /* if */
    } /* for */

    /* dump speculative miscellaneous regs */
    if (use_spec_HI)
      fprintf(stream, "[ $hi][%02d]: %12d/0x%08x\n",
              context, 
              spec_regs_HI[context],
              spec_regs_HI[context]);
    if (use_spec_LO)
      fprintf(stream, "[ $lo][%02d]: %12d/0x%08x\n",
              context, 
              spec_regs_LO[context],
              spec_regs_LO[context]);
    if (use_spec_FCC)
      fprintf(stream, "[$fcc][%02d]: 0x%08x\n",
              context, 
              spec_regs_FCC[context]);
    if (use_spec_CXTCTRL)
      fprintf(stream, "[$cxtctrl][%02d]: 0x%08x\n",
              context, spec_regs_CXTCTRL[context]);
    if (use_spec_ACXT)
      fprintf(stream, "[$acxt][%02d]: 0x%08x\n",
              context, spec_regs_ACXT[context]);
  } /* for */
}


/* speculative memory hash table size, NOTE: this must be a power-of-two */
#define STORE_HASH_SIZE		32

/* speculative memory hash table definition, accesses go through this hash
   table when accessing memory in speculative mode, the hash table flush the
   table when recovering from mispredicted branches */
struct spec_mem_ent {
  struct spec_mem_ent *next;		/* ptr to next hash table bucket */
  SS_ADDR_TYPE addr;			/* virtual address of spec state */
  unsigned int data[2];			/* spec buffer, up to 8 bytes */
};

/* speculative memory hash table */
static struct spec_mem_ent *store_htable[SS_NUM_CONTEXT][STORE_HASH_SIZE];

/* speculative memory hash table bucket free list */
static struct spec_mem_ent *bucket_free_list = NULL;


/* fetch unit next fetch address */
static SS_ADDR_TYPE fetch_regs_PC[SS_NUM_CONTEXT];
static SS_ADDR_TYPE fetch_pred_PC[SS_NUM_CONTEXT];

/* IFETCH -> DISPATCH instruction queue definition */
struct fetch_rec {
  SS_INST_TYPE IR;                      /* inst register */
  SS_ADDR_TYPE regs_PC, pred_PC;        /* current PC, predicted next PC */
  SS_CONTEXT_TYPE context;              /* context id */
  struct bpred_update dir_update;	/* bpred direction update info */
  int stack_recover_idx;		/* branch predictor RSB index */
  unsigned int ptrace_seq;		/* print trace sequence id */
};
static struct fetch_rec *fetch_data;	/* IFETCH -> DISPATCH inst queue */
static int fetch_num;			/* num entries in IF -> DIS queue */
static int fetch_tail, fetch_head;	/* head and tail pointers of queue */

/* recover instruction trace generator state to precise state state immediately
   before the first mis-predicted branch; this is accomplished by resetting
   all register value copied-on-write bitmasks are reset, and the speculative
   memory hash table is cleared */
static void
tracer_recover(SS_CONTEXT_TYPE context)
{
  int i, idx, new_num;
  struct spec_mem_ent *ent, *ent_next;
  struct fetch_rec *fetch_restored; 

  fetch_restored = (struct fetch_rec *)
    calloc(ruu_ifq_size, sizeof(struct fetch_rec));
  if (!fetch_restored)
    panic("Unable to allocate memory");

  /* The next lines where suppressed because it is possible to have
     speculative instructions in the RUU while executing not in
     spec_mode instructions in the dispatch unit, when for instance, a
     context was stopped, then resumed */

  /* better be in mis-speculative trace generation mode */
/*    if ((!spec_mode[context]) &&  */
/*        (BITMAP_SET_P(sim_active_context, SIM_ACTIVE_SZ, context))) */
/*      panic("cannot recover unless in speculative mode"); */


  /* reset to non-speculative trace generation mode */
  spec_mode[context] = FALSE;

  /* reset copied-on-write register bitmasks back to non-speculative state */
  for(i = 0; i < SS_NUM_REGS; i++) {
    BITMAP_CLEAR(use_spec_R, R_BMAP_SZ, i + (context * SS_NUM_REGS));
    BITMAP_CLEAR(use_spec_F, F_BMAP_SZ, i + (context * SS_NUM_REGS));
  } /* for */
  use_spec_HI[context] = FALSE;
  use_spec_LO[context] = FALSE;
  use_spec_FCC[context] = FALSE;
  use_spec_ACXT[context] = FALSE;
  use_spec_CXTCTRL[context] = FALSE;

  /* reset memory state back to non-speculative state */
  /* FIXME: could version stamps be used here?!?!? */
  for (i=0; i<STORE_HASH_SIZE; i++)
    {
      /* release all hash table buckets */
      for (ent=store_htable[context][i]; ent; ent=ent_next)
	{
	  ent_next = ent->next;
	  ent->next = bucket_free_list;
	  bucket_free_list = ent;
	}
      store_htable[context][i] = NULL;
    }

  /* Create a new ifetch queue without instr from the current context */
  idx = fetch_head;
  new_num = 0;

  for (i = 0 ; i < fetch_num; i++) {
    if (fetch_data[idx].context != context) {
      memcpy(&fetch_restored[new_num], &fetch_data[idx],
             sizeof(struct fetch_rec));
      new_num++;
    } /* if */
    idx = (idx + 1) & (ruu_ifq_size - 1);    
  } /* for */

  /* if pipetracing, indicate squash of instructions in the inst fetch queue */
  if (ptrace_active) {

    while (fetch_num != 0) {
      /* squash the next instruction from the IFETCH -> DISPATCH queue */

      ptrace_endinst(context, fetch_data[fetch_head].ptrace_seq);
      
      /* consume instruction from IFETCH -> DISPATCH queue */
      fetch_head = (fetch_head+1) & (ruu_ifq_size - 1);
      fetch_num--;

    } /* while */
  } /* if */

  /* Copy the created list not including the emptied context */
  for (i = 0 ; i < new_num; i++) {
    memcpy(&fetch_data[i], &fetch_restored[i], sizeof(struct fetch_rec));
  } /* for */  
  free(fetch_restored);

  /* reset IFETCH state */
  fetch_num = new_num;
  fetch_tail = new_num & (ruu_ifq_size - 1);
  fetch_head = 0;

  fetch_pred_PC[context] = fetch_regs_PC[context] = recover_PC[context];

  /* regs_PC is used to backup PC value when a context going down */
  regs_PC[context] = recover_PC[context];

}

/* initialize the speculative instruction state generator state */
static void
tracer_init(void)
{
  int i;
  SS_CONTEXT_TYPE context;

  /* register state is from non-speculative state buffers */
  BITMAP_CLEAR_MAP(use_spec_R, R_BMAP_SZ);
  BITMAP_CLEAR_MAP(use_spec_F, F_BMAP_SZ);
  for(context = 0; context < SS_NUM_CONTEXT; context++) {
    /* initially in non-speculative mode */
    spec_mode[context] = FALSE;

    use_spec_HI[context] = FALSE;
    use_spec_LO[context] = FALSE;
    use_spec_FCC[context] = FALSE;
    use_spec_ACXT[context] = FALSE;
    use_spec_CXTCTRL[context] = FALSE;

    /* memory state is from non-speculative memory pages */
    for (i=0; i<STORE_HASH_SIZE; i++)
      store_htable[context][i] = NULL;

  } /* for */

}


/* speculative memory hash table address hash function */
#define HASH_ADDR(ADDR)							\
  ((((ADDR) >> 24)^((ADDR) >> 16)^((ADDR) >> 8)^(ADDR)) & (STORE_HASH_SIZE-1))

/* this functional provides a layer of mis-speculated state over the
   non-speculative memory state, when in mis-speculation trace generation mode,
   the simulator will call this function to access memory, instead of the
   non-speculative memory access interfaces defined in memory.h; when storage
   is written, an entry is allocated in the speculative memory hash table,
   future reads and writes while in mis-speculative trace generation mode will
   access this buffer instead of non-speculative memory state; when the trace
   generator transitions back to non-speculative trace generation mode,
   tracer_recover() clears this table */
static void
spec_mem_access(enum mem_cmd cmd,		/* Read or Write access cmd */
                SS_CONTEXT_TYPE context,        /* context number to access */
		SS_ADDR_TYPE addr,		/* virtual address of access */
		void *p,			/* input/output buffer */
		int nbytes)			/* number of bytes to access */
{
  int index;
  struct spec_mem_ent *ent, *prev;

  /* FIXME: partially overlapping writes are not combined... */
  /* FIXME: partially overlapping reads are not handled correctly... */

  /* check alignments, even speculative this test should always pass */
  if ((nbytes & (nbytes-1)) != 0 || (addr & (nbytes-1)) != 0)
    {
      /* FIXMEDM */
      if (cmd == Read)
        switch (nbytes) {
          case 4: *((unsigned long *)p) = 0; break;
          case 2: *((unsigned short *)p) = 0; break;
          case 1: *((unsigned char *)p) = 0; break;
        }
      return;
    }

  /* check permissions */
  if (!((addr >= ld_text_base[context] &&
         addr < (ld_text_base[context]+ld_text_size[context])
	 && cmd == Read)
	|| (addr >= ld_data_base[context] &&
            addr < ld_stack_base[context])))
    {
      /* no can do */
      /* FIXMEDM */
       if (cmd == Read)
         switch (nbytes) {
           case 4: *((unsigned long *)p) = 0; break;
           case 2: *((unsigned short *)p) = 0; break;
           case 1: *((unsigned char *)p) = 0; break;
         }
      return;
    }

  /* has this memory state been copied on mis-speculative write? */
  index = HASH_ADDR(addr);
  for (prev=NULL,ent=store_htable[context][index]; ent; prev=ent,ent=ent->next)
    {
      if (ent->addr == addr)
	{
	  /* reorder chains to speed access into hash table */
	  if (prev != NULL)
	    {
	      /* not at head of list, relink the hash table entry at front */
	      prev->next = ent->next;
              ent->next = store_htable[context][index];
              store_htable[context][index] = ent;
	    }
	  break;
	}
    }

  /* no, if it is a write, allocate a hash table entry to hold the data */
  if (!ent && cmd == Write)
    {
      /* try to get an entry from the free list, if available */
      if (!bucket_free_list)
	{
	  /* otherwise, call calloc() to get the needed storage */
	  bucket_free_list = calloc(1, sizeof(struct spec_mem_ent));
	  if (!bucket_free_list)
	    fatal("out of virtual memory");
	}
      ent = bucket_free_list;
      bucket_free_list = bucket_free_list->next;

      if (!bugcompat_mode)
	{
	  /* insert into hash table */
	  ent->next = store_htable[context][index];
	  store_htable[context][index] = ent;
	  ent->addr = addr;
	  ent->data[0] = 0; ent->data[1] = 0;
	}
    }

  /* handle the read or write to speculative or non-speculative storage */
  switch (nbytes)
    {
    case 1:
      if (cmd == Read)
	{
	  if (ent)
	    {
	      /* read from mis-speculated state buffer */
	      *((unsigned char *)p) = *((unsigned char *)(&ent->data[0]));
	    }
	  else
	    {
	      /* read from non-speculative memory state */
	      *((unsigned char *)p) = MEM_READ_BYTE(addr, context);
	    }
	}
      else
	{
	  /* always write into mis-speculated state buffer */
	  *((unsigned char *)(&ent->data[0])) = *((unsigned char *)p);
	}
      break;
    case 2:
      if (cmd == Read)
	{
	  if (ent)
	    {
	      /* read from mis-speculated state buffer */
	      *((unsigned short *)p) = *((unsigned short *)(&ent->data[0]));
	    }
	  else
	    {
	      /* read from non-speculative memory state */
	      *((unsigned short *)p) = MEM_READ_HALF(addr, context);
	    }
	}
      else
	{
	  /* always write into mis-speculated state buffer */
	  *((unsigned short *)&ent->data[0]) = *((unsigned short *)p);
	}
      break;
    case 4:
      if (cmd == Read)
	{
	  if (ent)
	    {
	      /* read from mis-speculated state buffer */
	      *((unsigned int *)p) = *((unsigned int *)&ent->data[0]);
	    }
	  else
	    {
	      /* read from non-speculative memory state */
	      *((unsigned int *)p) = MEM_READ_WORD(addr, context);
	    }
	}
      else
	{
	  /* always write into mis-speculated state buffer */
	  *((unsigned int *)&ent->data[0]) = *((unsigned int *)p);
	}
      break;
    case 8:
      if (cmd == Read)
	{
	  if (ent)
	    {
	      /* read from mis-speculated state buffer */
	      *((unsigned int *)p) = *((unsigned int *)&ent->data[0]);
	      *(((unsigned int *)p)+1) = *((unsigned int *)&ent->data[1]);
	    }
	  else
	    {
	      /* read from non-speculative memory state */
	      *((unsigned int *)p) = MEM_READ_WORD(addr, context);
	      *(((unsigned int *)p)+1) =
		MEM_READ_WORD(addr + sizeof(SS_WORD_TYPE), context);
	    }
	}
      else
	{
	  /* always write into mis-speculated state buffer */
	  *((unsigned int *)&ent->data[0]) = *((unsigned int *)p);
	}
      break;
    default:
      panic("access size not supported in mis-speculative mode");
    }
}

/* dump speculative memory state */
static void
mspec_dump(FILE *stream)			/* output stream */
{
  int i;
  struct spec_mem_ent *ent;

  fprintf(stream, "** speculative memory contents **\n");

  fprintf(stream, "spec_mode: %s\n", spec_mode[current_context] ? "t" : "f");

  for (i=0; i<STORE_HASH_SIZE; i++)
    {
      /* dump contents of all hash table buckets */
      for (ent=store_htable[current_context][i]; ent; ent=ent->next)
	{
	  fprintf(stream, "[0x%08x]: %12.0f/0x%08x:%08x\n",
		  ent->addr, (double)(*((double *)ent->data)),
		  *((unsigned int *)&ent->data[0]),
		  *(((unsigned int *)&ent->data[0]) + 1));
	}
    }
}

/* default memory state accessor, used by DLite */
static char *					/* err str, NULL for no err */
simoo_mem_obj(enum dlite_access_t at,		/* access type */
              SS_CONTEXT_TYPE context,          /* context to access */
	      SS_ADDR_TYPE addr,		/* address to access */
	      char *p,				/* input/output buffer */
	      int nbytes)			/* size of access */
{
  char *errstr;
  enum mem_cmd cmd;

  if (at == at_read)
    cmd = Read;
  else if (at == at_write)
    cmd = Write;
  else
    panic("bogus access type");

  errstr = mem_valid(cmd, context, addr, nbytes, /* declare */FALSE);
  if (errstr)
    return errstr;

  /* else, no error, access memory */
  if (spec_mode[context])
    spec_mem_access(cmd, context, addr, p, nbytes);
  else
    mem_access(cmd, context, addr, p, nbytes);

  /* no error */
  return NULL;
}


/*
 *  RUU_DISPATCH() - decode instructions and allocate RUU and LSQ resources
 */


/* link RS onto the output chain number of whichever operation will next
   create the architected register value IDEP_NAME */
static INLINE void
ruu_link_idep(struct RUU_station *rs,		/* rs station to link */
	      int idep_num,			/* input dependence number */
	      int idep_name,			/* input register name */
              SS_CONTEXT_TYPE context)          /* context number */
{
  struct CV_link head;
  struct RS_link *link;

  /* any dependence? */
  if (idep_name == NA)
    {
      /* no input dependence for this input slot, mark operand as ready */
      rs->idep_ready[idep_num] = TRUE;
      return;
    }

  /* locate creator of operand */
  head = CREATE_VECTOR(idep_name + (SS_TOTAL_REGS * context));

  /* any creator? */
  if (!head.rs)
    {
      /* no active creator, use value available in architected reg file,
         indicate the operand is ready for use */
      rs->idep_ready[idep_num] = TRUE;
      return;
    }
  /* else, creator operation will make this value sometime in the future */

  /* indicate value will be created sometime in the future, i.e., operand
     is not yet ready for use */
  rs->idep_ready[idep_num] = FALSE;

  if (head.rs->completed) /* DM */ {
    fatal("idep_link on an issued operand");
  }

  /* link onto creator's output list of dependant operand */
  RSLINK_NEW(link, rs); link->x.opnum = idep_num;
  link->next = head.rs->odep_list[head.odep_num];
  head.rs->odep_list[head.odep_num] = link;
}

/* make RS the creator of architected register ODEP_NAME */
static INLINE void
ruu_install_odep(struct RUU_station *rs,	/* creating RUU station */
		 int odep_num,			/* output operand number */
		 int odep_name,			/* output register name */
                 SS_CONTEXT_TYPE context)       /* odep's context number */
{
  struct CV_link cv;

  /* any dependence? */
  if (odep_name == NA)
    {
      /* no value created */
      rs->onames[odep_num] = NA;
      rs->odep_list[odep_num] = NULL; /* FIXMEDM: Was it missing or what ? */
      return;
    }
  /* else, create a RS_NULL terminated output chain in create vector */

  /* record output name, used to update create vector at completion */
  rs->onames[odep_num] = odep_name + (SS_TOTAL_REGS * context);

  /* initialize output chain to empty list */
  rs->odep_list[odep_num] = NULL;

  /* indicate this operation is latest creator of ODEP_NAME */
  CVLINK_INIT(cv, rs, odep_num);
  SET_CREATE_VECTOR(odep_name + (SS_TOTAL_REGS * context), cv);
}

/*
 * configure the instruction decode engine
 */

/* general register dependence decoders */
#define DGPR(N)			(N)
#define DCGPR(N)		(SS_COMP_OP != SS_COMP_NOP ? (N) : 0)
#define DGPR_D(N)		((N)&~1)

/* floating point register dependence decoders */
#define DFPR_L(N)		(((N)+32)&~1)
#define DFPR_F(N)		(((N)+32)&~1)
#define DFPR_D(N)		(((N)+32)&~1)

/* miscellaneous register dependence decoders */
#define DHI			(0+32+32)
#define DLO			(1+32+32)
#define DFCC			(2+32+32)
#define DTMP			(3+32+32)
#define DACXT			(4+32+32)
#define DCXTCTRL		(5+32+32)
#define DNA			(0)


/*
 * configure the execution engine
 */


/* next program counter */
#define SET_NPC(CONTEXT,EXPR)   (next_PC[(CONTEXT)] = (EXPR))

/* target program counter */
#undef  SET_TPC
#define SET_TPC(CONTEXT,EXPR)   (target_PC[(CONTEXT)] = (EXPR))

/* current program counter */
#define CPC(CONTEXT)            (regs_PC[(CONTEXT)])
#define SET_CPC(CONTEXT,EXPR)   (regs_PC[(CONTEXT)] = (EXPR))

#define RPC(CONTEXT)            (recover_PC[(CONTEXT)])

/* general purpose register accessors, NOTE: speculative copy on write storage
   provided for fast recovery during wrong path execute (see tracer_recover()
   for details on this process */
#define GPR(CONTEXT,N)           (BITMAP_SET_P(use_spec_R,                  \
                                             R_BMAP_SZ,                     \
                                             ((CONTEXT)*SS_NUM_REGS + (N))) \
				 ? (spec_regs_R[((CONTEXT) << 5) + (N)])    \
				 : (regs_R[(CONTEXT)][(N)]))
#define SET_GPR(CONTEXT,N,EXPR) (spec_mode[(CONTEXT)]                       \
				 ? (BITMAP_SET(use_spec_R,                  \
                                               R_BMAP_SZ,                   \
                                               (CONTEXT)*SS_NUM_REGS + (N)),\
		              spec_regs_R[((CONTEXT) << 5) + (N)] = (EXPR)) \
				 : (regs_R[CONTEXT][N] = (EXPR)))
                        


/* floating point register accessors, NOTE: speculative copy on write storage
   provided for fast recovery during wrong path execute (see tracer_recover()
   for details on this process */
#define FPR_L(CONTEXT, N)                                                     \
                                (BITMAP_SET_P(use_spec_F,                     \
					   F_BMAP_SZ,                         \
					   ((CONTEXT)*SS_NUM_REGS + ((N)&~1)))\
				 ? spec_regs_F.l[(CONTEXT)][(N)]              \
				 : regs_F.l[(CONTEXT)][(N)])
#define SET_FPR_L(CONTEXT, N, EXPR)                                           \
                                (spec_mode[(CONTEXT)]                         \
				 ? (BITMAP_SET(use_spec_F,                    \
					  F_BMAP_SZ,                          \
					  ((CONTEXT)*SS_NUM_REGS + ((N)&~1))),\
				    spec_regs_F.l[CONTEXT][(N)] = (EXPR))     \
				 : (regs_F.l[CONTEXT][(N)] = (EXPR)))
#define FPR_F(CONTEXT, N)                                                     \
                                (BITMAP_SET_P(use_spec_F,                     \
					   F_BMAP_SZ,                         \
					   ((CONTEXT)*SS_NUM_REGS + ((N)&~1)))\
				 ? spec_regs_F.f[CONTEXT][(N)]                \
				 : regs_F.f[CONTEXT][(N)])
#define SET_FPR_F(CONTEXT, N, EXPR)                                           \
                                (spec_mode[(CONTEXT)]                         \
				 ? (BITMAP_SET(use_spec_F,                    \
					  F_BMAP_SZ,                          \
					  ((CONTEXT)*SS_NUM_REGS + ((N)&~1))),\
				    spec_regs_F.f[CONTEXT][(N)] = (EXPR))     \
				 : (regs_F.f[CONTEXT][(N)] = (EXPR)))
#define FPR_D(CONTEXT, N)       (BITMAP_SET_P(use_spec_F,                     \
					   F_BMAP_SZ,                         \
					   ((CONTEXT)*SS_NUM_REGS + ((N)&~1)))\
				 ? spec_regs_F.d[CONTEXT][(N) >> 1]           \
				 : regs_F.d[CONTEXT][(N) >> 1])
#define SET_FPR_D(CONTEXT, N, EXPR)                                           \
                                (spec_mode[(CONTEXT)]                         \
				 ? (BITMAP_SET(use_spec_F,                    \
					  F_BMAP_SZ,                          \
					  ((CONTEXT)*SS_NUM_REGS + ((N)&~1))),\
				    spec_regs_F.d[CONTEXT][(N) >> 1] = (EXPR))\
				 : (regs_F.d[CONTEXT][(N) >> 1] = (EXPR)))

/* miscellanous register accessors, NOTE: speculative copy on write storage
   provided for fast recovery during wrong path execute (see tracer_recover()
   for details on this process */
#define HI(CONTEXT)             (use_spec_HI[CONTEXT]                   \
                                 ? spec_regs_HI[CONTEXT]                \
                                 : regs_HI[CONTEXT])
#define SET_HI(CONTEXT, EXPR)   (spec_mode[(CONTEXT)]                   \
				 ? (use_spec_HI[CONTEXT]=TRUE,          \
				    spec_regs_HI[CONTEXT] = (EXPR))     \
				 : (regs_HI[CONTEXT] = (EXPR)))
#define LO(CONTEXT)             (use_spec_LO[CONTEXT]                   \
                                 ? spec_regs_LO[CONTEXT]                \
                                 : regs_LO[CONTEXT])
#define SET_LO(CONTEXT, EXPR)   (spec_mode[(CONTEXT)]                   \
				 ? (use_spec_LO[CONTEXT]=TRUE,          \
				    spec_regs_LO[CONTEXT] = (EXPR))     \
				 : (regs_LO[CONTEXT] = (EXPR)))
#define FCC(CONTEXT)            (use_spec_FCC[CONTEXT]                  \
                                 ? spec_regs_FCC[CONTEXT]               \
                                 : regs_FCC[CONTEXT])
#define SET_FCC(CONTEXT, EXPR)  (spec_mode[(CONTEXT)]                   \
				 ? (use_spec_FCC[CONTEXT] = TRUE,       \
				    spec_regs_FCC[CONTEXT] = (EXPR))    \
				 : (regs_FCC[CONTEXT] = (EXPR)))
/* context specific registers */
#define ACXT(CONTEXT)            (use_spec_ACXT[(CONTEXT)]              \
                                 ? spec_regs_ACXT[(CONTEXT)]            \
                                 : regs_ACXT)
#define SET_ACXT(CONTEXT, EXPR)  (spec_mode[(CONTEXT)]                  \
				 ? (use_spec_ACXT[CONTEXT] = TRUE,      \
				    spec_regs_ACXT[CONTEXT] = (EXPR))   \
				 : (regs_ACXT = (EXPR),                 \
                                     ((CONTEXT == context) &&           \
                                      !(ACTIVE_CXT(context))) ?         \
                                  (SET_CPC(context, CPC(context) +      \
                                           sizeof(SS_INST_TYPE)))  : 0))
#define CXTCTRL(CONTEXT)         (use_spec_CXTCTRL[(CONTEXT)]           \
                                 ? ((spec_regs_CXTCTRL[(CONTEXT)]       \
                                    & 0xffffffe0) | (CONTEXT))          \
                                 : ((regs_CXTCTRL[(CONTEXT)]            \
                                     & 0xffffffe0) | (CONTEXT)))
void
test_and_flush(SS_CONTEXT_TYPE context, int value)
{
  if (value & 0x00008000) {
    cache_flush(cache_il1[context], 0);
    cache_flush(cache_il2[context], 0);
    cache_flush(itlb[context], 0);
    fprintf(stderr, "### cache flush for context %d\n", context);
  } /* if */
} /* test_and_flush */
             
/* cxtctrl semantics must be preserved i.e. bits 4:0 represents context id */
#define SET_CXTCTRL(CONTEXT, EXPR) (spec_mode[(CONTEXT)]                \
				 ? (use_spec_CXTCTRL[CONTEXT] = TRUE,   \
				    spec_regs_CXTCTRL[CONTEXT] =        \
                                    ((EXPR) & 0xffffffe0) | (CONTEXT))  \
				 : (test_and_flush(CONTEXT, EXPR),      \
                                    (regs_CXTCTRL[(CONTEXT)] =          \
                                    ((EXPR) & 0xffffffe0) | (CONTEXT))))

#define SUSPEND_CXT(CONTEXT)    (regs_ACXT &= ~(1 << (CONTEXT)))
#define RESUME_CXT(CONTEXT)     (regs_ACXT |= (1 << (CONTEXT)))

#define SET_MMU_CONFIG(CONTEXT,EXPR) \
                                mmu_set_context_config((CONTEXT), (EXPR), mmu)
#define MMU_CONFIG(CONTEXT) \
                                mmu_get_context_config((CONTEXT), mmu)



/* precise architected memory state accessor macros, NOTE: speculative copy on
   write storage provided for fast recovery during wrong path execute (see
   tracer_recover() for details on this process */
#define __READ_SPECMEM(CONTEXT, DST_T, SRC_N, SRC)			\
  (addr = (SRC),							\
   (spec_mode[(CONTEXT)]                				\
    ? spec_mem_access(Read, CONTEXT, addr, SYMCAT(&temp_,SRC_N),	\
		      sizeof(SYMCAT(temp_,SRC_N)))			\
    : mem_access(Read, CONTEXT, addr, SYMCAT(&temp_,SRC_N),	       	\
		 sizeof(SYMCAT(temp_,SRC_N)))),				\
   ((unsigned int)((DST_T)SYMCAT(temp_,SRC_N))))

#define READ_WORD(SRC, CONTEXT)                         		\
  __READ_SPECMEM(CONTEXT, unsigned int, uint, (SRC))
   
#define READ_UNSIGNED_HALF(SRC, CONTEXT)				\
  __READ_SPECMEM(CONTEXT, unsigned int, ushort, (SRC))
#define READ_SIGNED_HALF(SRC, CONTEXT)				       	\
  __READ_SPECMEM(CONTEXT, signed int, short, (SRC))
#define READ_UNSIGNED_BYTE(SRC, CONTEXT)		       		\
  __READ_SPECMEM(CONTEXT, unsigned int, uchar, (SRC))
#define READ_SIGNED_BYTE(SRC, CONTEXT)					\
  __READ_SPECMEM(CONTEXT, signed int, char, (SRC))

#define __WRITE_SPECMEM(CONTEXT, SRC, DST_T, DST_N, DST)      		\
  (addr = (DST),							\
   SYMCAT(temp_,DST_N) = (DST_T)((unsigned int)(SRC)),			\
   (spec_mode[(CONTEXT)]        					\
    ? spec_mem_access(Write, CONTEXT, addr, SYMCAT(&temp_,DST_N),      	\
		      sizeof(SYMCAT(temp_,DST_N)))			\
    : mem_access(Write, CONTEXT, addr, SYMCAT(&temp_,DST_N),	      	\
		 sizeof(SYMCAT(temp_,DST_N)))))

#define WRITE_WORD(SRC, DST, CONTEXT)					\
  __WRITE_SPECMEM(CONTEXT, (SRC), unsigned int, uint, (DST))
#define WRITE_HALF(SRC, DST, CONTEXT)					\
  __WRITE_SPECMEM(CONTEXT, (SRC), unsigned short, ushort, (DST))
#define WRITE_BYTE(SRC, DST, CONTEXT)					\
  __WRITE_SPECMEM(CONTEXT, (SRC), unsigned char, uchar, (DST))

/* system call handler macro */
#define SYSCALL(INST, CONTEXT)						\
  (/* only execute system calls in non-speculative mode */		\
   (spec_mode[(CONTEXT)] ? panic("speculative syscall") : (void) 0),    \
   ss_syscall(mem_access, CONTEXT, INST))

/* non-expression inst implementations */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,CLASS,O1,O2,I1,I2,I3,EXPR)
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)
#define CONNECT(OP)
#define IMPL
#include "ss.def"
#undef DEFINST
#undef DEFLINK
#undef CONNECT
#undef IMPL


/* default register state accessor, used by DLite */
static char *					/* err str, NULL for no err */
simoo_reg_obj(enum dlite_access_t at,		/* access type */
              SS_CONTEXT_TYPE context,          /* context to access */
	      enum dlite_reg_t rt,		/* reg bank to probe */
	      int reg,				/* register number */
	      union dlite_reg_val_t *val)       /* input, output */
{
  if (reg < 0 || reg >= SS_NUM_REGS)
    return "register number out of range";

  if (at == at_read || at == at_write)
    {
      switch (rt)
	{
	case rt_gpr:
	  if (at == at_read)
	    val->as_word = GPR(context, reg);
	  else
	    SET_GPR(context, reg, val->as_word);
	  break;
	case rt_lpr:
	  if (at == at_read)
	    val->as_word = FPR_L(context, reg);
	  else
	    SET_FPR_L(context, reg, val->as_word);
	  break;
	case rt_fpr:
	  if (at == at_read)
	    val->as_float = FPR_F(context, reg);
	  else
	    SET_FPR_F(context, reg, val->as_float);
	  break;
	case rt_dpr:
	  /* 1/2 as many regs in this mode */
	  if (reg < 0 || reg >= SS_NUM_REGS/2)
	    return "register number out of range";

	  if (at == at_read)
	    val->as_double = FPR_D(context, reg * 2);
	  else
	    SET_FPR_D(context, reg * 2, val->as_double);
	  break;
	case rt_hi:
	  if (at == at_read)
	    val->as_word = HI(context);
	  else
	    SET_HI(context, val->as_word);
	  break;
	case rt_lo:
	  if (at == at_read)
	    val->as_word = LO(context);
	  else
	    SET_LO(context, val->as_word);
	  break;
	case rt_FCC:
	  if (at == at_read)
	    val->as_condition = FCC(context);
	  else
	    SET_FCC(context, val->as_condition);
	  break;
	case rt_PC:
	  if (at == at_read)
	    val->as_address = regs_PC[context];
	  else
	    regs_PC[context] = val->as_address;
	  break;
	default:
	  panic("bogus register bank");
	}
    }
  else
    panic("bogus access type");

  /* no error */
  return NULL;
}


/* fetch_queue_strip take out all occurences of instructions of
   a given context. fetch_head, fetch_num and fetch_tail are
   modified accordingly to the new fetch queue data. */
static void
fetch_queue_strip(SS_CONTEXT_TYPE context){
  int head_delete, head_insert = fetch_head;
  int i, num = fetch_num;

  /* Let's start at the first position found where context
     has to be taken out. */

  while ((fetch_data[head_insert].context != context) && num) {
    head_insert = (head_insert + 1) & (ruu_ifq_size - 1);
    num--;
  } /* while */
  head_delete = head_insert;
/*   head_delete = (head_insert + 1) & (ruu_ifq_size - 1); */

  /* Move the entries. */

  while (num-- > 0) {
    if (fetch_data[head_delete].context != context) {
      memcpy(&fetch_data[head_insert], &fetch_data[head_delete],
             sizeof(struct fetch_rec));
      head_insert = (head_insert + 1) & (ruu_ifq_size - 1);
    } else {
      /* Smashed entries */
      ptrace_endinst(fetch_data[head_delete].context, /* FIXMEDM: repl.
                                                         by context */
                     fetch_data[head_delete].ptrace_seq);
      fetch_num--;    

    } /* if */
    head_delete = (head_delete + 1) & (ruu_ifq_size - 1);    
  } /* while */
  fetch_tail = head_insert; 

  /* DK: The following codes were taken from tracer_recover() and were not in 
	 the original simulator code. I think this is necessary. */
  /* reset copied-on-write register bitmasks back to non-speculative state */
  for(i = 0; i < SS_NUM_REGS; i++) {
    BITMAP_CLEAR(use_spec_R, R_BMAP_SZ, i + (context * SS_NUM_REGS));
    BITMAP_CLEAR(use_spec_F, F_BMAP_SZ, i + (context * SS_NUM_REGS));
  } /* for */
  use_spec_HI[context] = FALSE;
  use_spec_LO[context] = FALSE;
  use_spec_FCC[context] = FALSE;
  use_spec_ACXT[context] = FALSE;
  use_spec_CXTCTRL[context] = FALSE;

} /* fetch_queue_strip */


/* DK: macro to print out instructions */
/* DK: DEBUGGING TOOL */
void print_mem_ref(enum ss_opcode op, int context, unsigned inst, unsigned PC) 
{
  char temp_str[8];

  if (spec_mode[context])
    return;

  switch(op) {
    case SB: {
      strcpy(temp_str, "SB ");
      goto PR1;
    }
    case SH: {
      strcpy(temp_str, "SH ");
      goto PR1;
    }
    case SW: {
      strcpy(temp_str, "SW ");
      PR1: 
      fprintf(stderr, "%10d [%d] PC(%x)  %s  R%2d = %d (%x)  addr = %x\n", 
	(unsigned) sim_cycle, context, PC, temp_str, (inst & 0x00ff0000) >> 16, 
	GPR(context, (inst & 0xff0000) >> 16), GPR(context, (inst & 0xff0000) >> 16), 
	(inst & 0x0000ffff) + GPR(context, (inst & 0xff000000) >> 24));
      break;
    }
    case S_S: {
      strcpy(temp_str, "S.S");
      fprintf(stderr, "%10d [%d] PC(%x)  %s  F%2d = %d (%x)  addr = %x\n", 
	(unsigned) sim_cycle, context, PC, temp_str, (inst & 0xff0000) >> 16, 
	FPR_L(context, (inst & 0xff0000) >> 16), 
	FPR_L(context, (inst & 0xff0000) >> 16), 
	(inst & 0x0000ffff) + GPR(context, (inst & 0xff000000) >> 24));
      break;
    }
    case S_D: {
      strcpy(temp_str, "S.D");
      fprintf(stderr, "%10d [%d] PC(%x)  %s  F%2d = %d (%x)  F%2d = %d (%x)  addr = %x\n", 
	(unsigned) sim_cycle, context, PC, temp_str, (inst & 0xff0000) >> 16, 
	FPR_L(context, (inst & 0xff0000) >> 16), FPR_L(context, (inst & 0xff0000) >> 16),
 	((inst & 0xff0000) >> 16) + 1, FPR_L(context, ((inst & 0xff0000) >> 16) + 1),
	FPR_L(context, ((inst & 0xff0000) >> 16) + 1),
	(inst & 0x0000ffff) + GPR(context, (inst & 0xff000000) >> 24));
      break;
    }
    case LB: {
      strcpy(temp_str, "LB ");
      goto PR2;
    }
    case LBU:{
      strcpy(temp_str, "LBU");
      goto PR2;
    }
    case LH: {
      strcpy(temp_str, "LH ");
      goto PR2;
    }
    case LHU:{
      strcpy(temp_str, "LHU");
      goto PR2;
    }
    case LW: {
      strcpy(temp_str, "LW ");
      PR2:
      fprintf(stderr, "%10d [%d] PC(%x)  %s  R%2d = %d (%x)  addr = %x\n", 
	(unsigned) sim_cycle, context, PC, temp_str, (inst & 0xff0000) >> 16, 
	GPR(context, (inst & 0xff0000) >> 16), GPR(context, (inst & 0xff0000) >> 16), 
	(inst & 0x0000ffff) + GPR(context, (inst & 0xff000000) >> 24));
      break;
    }
    case LUI: {
      strcpy(temp_str, "LUI");
      fprintf(stderr, "%10d [%d] PC(%x)  %s  R%2d = %d (%x)  immd = %x\n", 
	(unsigned) sim_cycle, context, PC, temp_str, (inst & 0xff0000) >> 16, 
	GPR(context, (inst & 0xff0000) >> 16), GPR(context, (inst & 0xff0000) >> 16), 
	inst & 0xffff);
      break;
    }
    case L_S: {
      strcpy(temp_str, "L.S");
      fprintf(stderr, "%10d [%d] PC(%x)  %s  F%2d = %d (%x)  addr = %x\n", 
	(unsigned) sim_cycle, context, PC, temp_str, (inst & 0xff0000) >> 16, 
	FPR_L(context, (inst & 0xff0000) >> 16), 
	FPR_L(context, (inst & 0xff0000) >> 16), 
	(inst & 0x0000ffff) + GPR(context, (inst & 0xff000000) >> 24));
      break;
    }
    case L_D: {
      strcpy(temp_str, "L.D");
      fprintf(stderr, "%10d [%d] PC(%x)  %s  F%2d = %d (%x)  F%2d = %d (%x)  addr = %x\n", 
	(unsigned) sim_cycle, context, PC, temp_str, (inst & 0xff0000) >> 16, 
	FPR_L(context, ((inst & 0xff0000) >> 16) ^ sim_swap_words), 
	FPR_L(context, ((inst & 0xff0000) >> 16) ^ sim_swap_words),
 	((inst & 0xff0000) >> 16) + 1, 
	FPR_L(context, (((inst & 0xff0000) >> 16) + 1) ^ sim_swap_words),
	FPR_L(context, (((inst & 0xff0000) >> 16) + 1) ^ sim_swap_words),
	(inst & 0x0000ffff) + GPR(context, (inst & 0xff000000) >> 24));
      break;
    }
    default:
      break;
  }
}

/* DK: dispatch stage stats */
static int ruu_full_cycle = 0;
static int lsq_full_cycle = 0;

/* context operation seq. num. */
static unsigned int context_op_seq = 0;

/* non-zero for a valid address */
#define VALID_ADDR(ADDR) (SS_DATA_BASE <= (ADDR) && (ADDR) <= SS_STACK_BASE)

/* dispatch instructions from the IFETCH -> DISPATCH queue: instructions are
   first decoded, then they allocated RUU (and LSQ for load/stores) resources
   and input and output dependence chains are updated accordingly */
static void
ruu_dispatch(void)
{
  int i;	
  int n_dispatched;                     /* total insts dispatched */
  SS_INST_TYPE inst;                    /* actual instruction bits */
  enum ss_opcode op;                    /* decoded opcode enum */
  int out1, out2, in1, in2, in3;        /* output/input register names */
  SS_ADDR_TYPE next_PC[SS_NUM_CONTEXT]; /* actual target PC address */ 
  SS_ADDR_TYPE target_PC[SS_NUM_CONTEXT]; /* next target PC address */
  SS_ADDR_TYPE addr;                    /* effective address, if load/store */
  struct RUU_station *rs;               /* RUU station being allocated */
  struct RUU_station *lsq;              /* LSQ station for ld/st's */
  struct bpred_update *dir_update_ptr;	/* branch predictor dir update ptr */
  int stack_recover_idx;		/* bpred retstack recovery index */
  unsigned int pseq;			/* pipetrace sequence number */
  int is_write;				/* store? */
  int made_check;			/* used to ensure DLite entry */
  int br_taken, br_pred_taken;		/* if br, taken?  predicted taken? */
  int fetch_redirected[SS_NUM_CONTEXT];
  unsigned int temp_uint;		/* temp variable for spec mem access */
  signed short temp_short;		/* " ditto " */
  unsigned short temp_ushort;		/* " ditto " */
  signed char temp_char;		/* " ditto " */
  unsigned char temp_uchar;		/* " ditto " */
  SS_CONTEXT_TYPE context;              /* context number */
  SS_WORD_TYPE stop_mask;               /* context to suspended */
  SS_WORD_TYPE start_mask;              /* context to suspended */
  SS_WORD_TYPE previous_ACXT;           /* backup of $acxt register */

  made_check = FALSE;
  n_dispatched = 0;
  for (i = 0; i < context_num; i++)
    fetch_redirected[i] = FALSE;

  context = fetch_data[fetch_head].context;

  if (RUU_num == RUU_size) {
    ruu_full_cycle++;
  } else {
    if (LSQ_num == LSQ_size)
      lsq_full_cycle++;
  }

  while (/* instruction decode B/W left? */
	 n_dispatched < (ruu_decode_width * fetch_speed)
	 /* RUU and LSQ not full? */
	 && (RUU_num < RUU_size) && (LSQ_num < LSQ_size)
	 /* insts still available from fetch unit? */
	 && fetch_num != 0
	 /* on an acceptable trace path */
	 && (ruu_include_spec || !spec_mode[context]))
    {
      current_context = context;

      if (interrupt_raised) {
        if (fetch_data[fetch_head].regs_PC == interrupt_vector) {
          interrupt_raised--;
        } else {
          n_dispatched++;
          fetch_head = (fetch_head+1) & (ruu_ifq_size - 1);
          fetch_num--;
          continue;          
        } /* if */
      } /* if */

      /* if issuing in-order, block until last op issues if inorder issue */
      if (ruu_inorder_issue
	  && (last_op[context].rs && RSLINK_VALID(&last_op[context])
	      && !OPERANDS_READY(last_op[context].rs)))
	{
	  /* stall until last operation is ready to issue */
	  break;
	}

      /* get the next instruction from the IFETCH -> DISPATCH queue */
      inst = fetch_data[fetch_head].IR;
      regs_PC[context] = fetch_data[fetch_head].regs_PC;
      pred_PC[context] = fetch_data[fetch_head].pred_PC;
      dir_update_ptr = &(fetch_data[fetch_head].dir_update);
      stack_recover_idx = fetch_data[fetch_head].stack_recover_idx;
      pseq = fetch_data[fetch_head].ptrace_seq;

      /* decode the inst */
      op = SS_OPCODE(inst);

      /* added by dongkeun */
      if ((op == NOP) && (spec_mode[context] == FALSE) &&
	  (inst.a & 0x00070000) == 0x00020000) { /* VSIM_EXIT */
          vsim_mode = FALSE;
      }

#ifdef CACHE_PROF
      if ((op == LPCR) && ((inst.a & 0x00070000) == 0x00010000) &&
	  (spec_mode[context] == FALSE) && context_master[context])
	  px_mode = TRUE; /* Entering a pre-execution region */

      if ((op == SPCS) && ((inst.a & 0x00070000) == 0x00010000) &&
	  (spec_mode[context] == FALSE) && context_master[context])
	  px_mode = FALSE; /* Exiting a pre-execution region */
#endif

      /* compute default next_PC */
      next_PC[context] = regs_PC[context] + sizeof(SS_INST_TYPE);

      /* saving ACXT register */
      previous_ACXT = regs_ACXT;

      /* do not execute CONTEXT instr. type if in speculative mode */
      /* DY:  permit the semaphore instructions to enter the RUU;
	 otherwise, long latency mis-predicted branches will delay
	 issue of other instructions.  */
      /* FIXME 1: Eventually, we should take this line out, allow all the 
	 context instructions in spec_mode to enter the RUU, and not
	 execute speculative context instructions in ss.def. */
      /* FIXME 2: After all, we should allow the context instructions in 
	 spec_mode to enter the RUU but execute them after a branch 
	 instruction is resolved in the ruu_writeback stage. */
      /* FIXME 3: Or, we can execute the context instruction right away
	 and if it is wrong, we should support squashing wrongfully
	 fetched instructions. But, we don't wanna do this. */
      if ((SS_OP_FLAGS(op) & F_CONTEXT) && spec_mode[context] && 
	  !((op == SEMP) || (op == SEMV) || (op == SEMVB) || 
	  (op == SPCS) || (op == SCU) || (op == LPCR) || (op == LCU))) 
        break;

      /* drain RUU for TRAPs and system calls */
      if (SS_OP_FLAGS(op) & F_TRAP)
	{
	  if (RUU_num != 0)
	    break;
	}

      /* maintain $r0 semantics (in spec and non-spec space) */
      regs_R[context][0] = 0; spec_regs_R[(context) << 5] = 0;

      /* default effective address (none) and access */
      addr = 0; is_write = FALSE;

      /* more decoding and execution */
      switch (op)
	{

/* the divide-by-zero instruction check macro is redefined because there
   is no portable and reliable way to hide integer divide-by-zero faults */
#undef  DIV0
#define DIV0(N)		(spec_mode[context]				\
			 ? /* do nothing */(void)0			\
			 : (((N) == 0) ? (IFAIL("divide by 0")) : (void)0))

/* the divide operator semantics are also redefined because there
   are no portable and reliable way to hide integer divide-by-zero faults */
#undef  IDIV
#define IDIV(A, B)	(((B) == 0) ? 0 : ((A) / (B)))

#undef  IMOD
#define IMOD(A, B)	(((B) == 0) ? 0 : ((A) % (B)))

#undef  FDIV
#define FDIV(A, B)	(((B) == 0) ? 0 : ((A) / (B)))

#undef  FINT
#define FINT(A)		(isnan(A) ? 0 : ((int)(A)))

/* decode and execute the instruction */
/* the following macro wraps the instruction check failure declaration
   with a test to see if the trace generator is in non-speculative
   mode, if so the instruction fault is declared, otherwise, the error
   is shunted because instruction faults need to be masked on the
   mis-speculated instruction paths */
#undef  IFAIL
#define IFAIL(S)							\
  (spec_mode[context] ? /* ignore */(void)0 : /* declare */(void)fatal(S))

#define DEFINST(OP,MSK,NAME,OPFORM,RES,CLASS,O1,O2,I1,I2,I3,EXPR)	\
	case OP:							\
	  out1 = O1; out2 = O2;						\
	  in1 = I1; in2 = I2; in3 = I3;					\
	  if (BITMAP_CLEAR_P(sim_shutdown_context, SIM_SHUTDOWN_SZ,     \
                             context)) EXPR;                            \
	  break;
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)					\
	case OP:							\
	  /* could speculatively decode a bogus inst */			\
	  op = NOP;							\
	  out1 = NA; out2 = NA;						\
	  in1 = NA; in2 = NA; in3 = NA;					\
	  /* no EXPR */							\
	  break;
#define CONNECT(OP)
#include "ss.def"
#undef DEFINST
#undef DEFLINK
#undef CONNECT
	default:
	  /* can speculatively decode a bogus inst */
	  op = NOP;
	  out1 = NA; out2 = NA;
	  in1 = NA; in2 = NA; in3 = NA;
	  /* no EXPR */
	} /* switch */

      /* DK: When an exception occurs, suspend the faulting context. */
      if (exception_signal) {
	exception_array[exception_signal]++;
	exception_signal = 0;
	op = SPCS;
	SUSPEND_CXT(context);
	kill_mode[context] = 1;
      }
/*
if (sim_cycle > 300000) {
	fprintf(stderr, "%10d [%d] PC(%x) %d addr = %x ", (unsigned) sim_cycle, 
		context, regs_PC[context], spec_mode[context], addr); 
	ss_print_insn(inst, regs_PC[context], stderr);
	fprintf(stderr, "\n");
}  

/*
print_mem_ref(op, context, inst.b, regs_PC[context]); 

      /* DK: DEBUGGING TOOL  
      if (debug_cycle && (sim_cycle > debug_cycle)) { 
	fprintf(stderr, "%10d [%d] PC(%x) %d ", (unsigned) sim_cycle, 
		context, regs_PC[context], spec_mode[context]); 
	ss_print_insn(inst, regs_PC[context], stderr);
	fprintf(stderr, "\n");
        print_mem_ref(op, context, inst.b, regs_PC[context]); 
      }*/

      /* operation sets next_PC */
      /* update memory access stats */
      if (SS_OP_FLAGS(op) & F_MEM)
	{
	  sim_total_refs++;
	  if (!spec_mode[context])
	    sim_num_refs++;

	  if (SS_OP_FLAGS(op) & F_STORE)
	    is_write = TRUE;
	  else
	    {
	      sim_total_loads++;
	      if (!spec_mode[context])
		sim_num_loads++;
	    }
	}

      if (!spec_mode[context] 
          && BITMAP_CLEAR_P(sim_shutdown_context, SIM_SHUTDOWN_SZ, context)) {
        backup_PC[context] = next_PC[context];
      }

      br_taken = (next_PC[context] != (regs_PC[context] +
                                       sizeof(SS_INST_TYPE)));
      br_pred_taken = (pred_PC[context] != (regs_PC[context] +
                                            sizeof(SS_INST_TYPE)));

      if ((pred_PC[context] != next_PC[context] && pred_perfect)
	  || ((SS_OP_FLAGS(op) & (F_CTRL|F_DIRJMP)) == (F_CTRL|F_DIRJMP)
	      && target_PC[context] != pred_PC[context] && br_pred_taken))
	{
	  /* Either 1) we're simulating perfect prediction and are in a
             mis-predict state and need to patch up, or 2) We're not simulating
             perfect prediction, we've predicted the branch taken, but our
             predicted target doesn't match the computed target (i.e.,
             mis-fetch).  Just update the PC values and do a fetch squash.
             This is just like calling fetch_squash() except we pre-anticipate
             the updates to the fetch values at the end of this function.  If
             case #2, also charge a mispredict penalty for redirecting fetch */

	  fetch_pred_PC[context] = fetch_regs_PC[context] = next_PC[context];

	  /* was: if (pred_perfect) */
#if 1
	  if (pred_perfect)
#endif
	    pred_PC[context] = next_PC[context];

          fetch_queue_strip(context);
                  /* The next lines are needed because fetch_head is */
                  /* modified at the end of this function. */
 	  fetch_head = (fetch_head + ruu_ifq_size - 1 ) & (ruu_ifq_size-1);
	  ++fetch_num;

	  if (!pred_perfect)
	    ruu_fetch_issue_delay[context] = ruu_branch_penalty;
#if 0
	  else
	    {
	      /* only 1 cycle bubble */
	      ruu_fetch_issue_delay[context] = 1;
	    }
#endif

	  fetch_redirected[context] = TRUE;
	}

      /* is this a NOP */
      if (op != NOP)
	{
	  /* for load/stores:
	       idep #0     - store operand (value that is store'ed)
	       idep #1, #2 - eff addr computation inputs (addr of access)

	     resulting RUU/LSQ operation pair:
	       RUU (effective address computation operation):
		 idep #0, #1 - eff addr computation inputs (addr of access)
	       LSQ (memory access operation):
		 idep #0     - operand input (value that is store'd)
		 idep #1     - eff addr computation result (from RUU op)

	     effective address computation is transfered via the reserved
	     name DTMP
	   */

	  /* fill in RUU reservation station */

	  /* RUU_DK: find a free RUU entry! */
	  RUU_tail[context] = FIND_FREE_ENTRY(RUU, RUU_size, RUU_tail[context]);
	  rs = &(RUU[RUU_tail[context]]);
	  rs->my_seq = RUU_sequence[context]; /* assign a sequence number */
	  ruu_table[context][rs->my_seq] = RUU_tail[context];
	  RUU_sequence[context] = (RUU_sequence[context] + 1) % RUU_size;
	
	  if (RUU_num_cxt[context] == 0)  /* new instruction for this context! */
	    RUU_head[context] = RUU_tail[context]; /* start from the same place! */

/*	  rs = &(RUU[RUU_tail]); */

	  rs->IR = inst;
	  rs->op = op;
          rs->context = context;
	  rs->PC = regs_PC[context];
	  rs->next_PC = next_PC[context]; rs->pred_PC = pred_PC[context];
	  rs->in_LSQ = FALSE;
	  rs->ea_comp = FALSE;
	  rs->recover_inst = FALSE;
          rs->dir_update = *dir_update_ptr;
	  rs->stack_recover_idx = stack_recover_idx;
	  rs->spec_mode = spec_mode[context];
	  rs->addr = 0;
	  /* rs->tag is already set */
	  rs->seq = ++inst_seq;
	  rs->queued = rs->issued = rs->completed = rs->retired = FALSE;
          rs->idle =
            BITMAP_SET_P(sim_shutdown_context, SIM_SHUTDOWN_SZ, context);
	  rs->ptrace_seq = pseq;
	  rs->valid = TRUE; /* RUU_DK */

          if (((SS_OP_FLAGS(op) & F_CONTEXT) ||
	       ((op == SYSCALL) && (regs_R[context][2] == SS_SYS_exit)))
	      && (!rs->idle)) {
            unsigned int i = 0;

            rs->target_mask = regs_ACXT;
            rs->context_op_seq = context_op_seq++;
            rs->target_op = 1;

            stop_mask = (previous_ACXT ^ regs_ACXT) & previous_ACXT;
            start_mask = (previous_ACXT ^ regs_ACXT) & regs_ACXT;

	    /* DK: The next resume request should not be processed until
		   all the instructions of the killed context are squashed. */
            while (stop_mask | start_mask) {
              if (stop_mask & 0x00000001) {
                BITMAP_SET(sim_shutdown_context, SIM_SHUTDOWN_SZ, i);
              } else if ((start_mask & 0x00000001) && !kill_mode[i]) {
/*              } else if (start_mask & 0x00000001) {*/
                if ((BITMAP_SET_P(sim_shutdown_context, SIM_SHUTDOWN_SZ, i)) 
                    && (BITMAP_SET_P(sim_active_context, SIM_ACTIVE_SZ, i))) {
                  BITMAP_SET(sim_restart_context, SIM_RESTART_SZ, i);
                } /* if */
                sim_num_start[i]++; 
              } /* if */
              i++;
              stop_mask = stop_mask >> 1;
              start_mask = start_mask >> 1;
            } /* while */
          } else {
            rs->target_op = 0;
          }

	  /* split ld/st's into two operations: eff addr comp + mem access */
	  if (SS_OP_FLAGS(op) & F_MEM)
	    {
	      /* convert RUU operation from ld/st to an add (eff addr comp) */
	      rs->op = ADD;
	      rs->ea_comp = TRUE;

	      /* fill in LSQ reservation station */

	      /* LSQ_DK: find a free LSQ entry! */
	      LSQ_tail[context] = FIND_FREE_ENTRY(LSQ, LSQ_size, LSQ_tail[context]);
	      lsq = &(LSQ[LSQ_tail[context]]);
	      lsq->my_seq = LSQ_sequence[context]; /* assign a sequence number */
	      lsq_table[context][lsq->my_seq] = LSQ_tail[context];
	      LSQ_sequence[context] = (LSQ_sequence[context] + 1) % LSQ_size;
	
	      if (LSQ_num_cxt[context] == 0) /* new instruction for this context! */
		LSQ_head[context] = LSQ_tail[context]; /* start from the same place */	

	      /* Keep track of the order of the LSQ entries */
	      lsq->prev = prev_entry;
	      lsq->next = -1;
	      if (prev_entry != -1) {
		LSQ[prev_entry].next = LSQ_tail[context];
	      } else {
		LSQ_oldest = LSQ_tail[context];
	      }
	      prev_entry = LSQ_tail[context];

	      lsq->IR = inst;
	      lsq->op = op;
              lsq->context = context;
	      lsq->PC = regs_PC[context];
	      lsq->next_PC = next_PC[context]; lsq->pred_PC = pred_PC[context];
	      lsq->in_LSQ = TRUE;
	      lsq->ea_comp = FALSE;
	      lsq->recover_inst = FALSE;
	      lsq->dir_update.pdir1 = lsq->dir_update.pdir2 = NULL;
	      lsq->dir_update.pmeta = NULL;
	      lsq->stack_recover_idx = 0;
	      lsq->spec_mode = spec_mode[context];
	      lsq->addr = addr;
	      /* lsq->tag is already set */
	      lsq->seq = ++inst_seq;
	      lsq->queued = lsq->issued = FALSE;
              lsq->completed = lsq->retired = FALSE;
              lsq->idle = rs->idle;
	      lsq->ptrace_seq = ptrace_seq[context]++;
	      lsq->valid = TRUE; /* LSQ_DK */

	      /* pipetrace this uop */
	      ptrace_newuop(context, lsq->ptrace_seq,
                            "internal ld/st", lsq->PC, 0);
	      ptrace_newstage(context, lsq->ptrace_seq, PST_DISPATCH, 0);

	      /* link eff addr computation onto operand's output chains */
	      ruu_link_idep(rs, /* idep_ready[] index */0, NA, context);
	      ruu_link_idep(rs, /* idep_ready[] index */1, in2, context);
	      ruu_link_idep(rs, /* idep_ready[] index */2, in3, context);

	      /* install output after inputs to prevent self reference */
	      ruu_install_odep(rs, /* odep_list[] index */0, DTMP, context);
	      ruu_install_odep(rs, /* odep_list[] index */1, NA, context);

	      /* link memory access onto output chain of eff addr operation */
	      ruu_link_idep(lsq,
			    /* idep_ready[] index */STORE_OP_INDEX/* 0 */,
			    in1, context);
	      ruu_link_idep(lsq,
			    /* idep_ready[] index */STORE_ADDR_INDEX/* 1 */,
			    DTMP, context);
	      ruu_link_idep(lsq, /* idep_ready[] index */2, NA, context);

	      /* DY:  don't modify create vector if this instruction is idle */
	      if (!(lsq->idle)) {
		/* install output after inputs to prevent self reference */
		ruu_install_odep(lsq, /* odep_list[] index */0, out1, context);
		ruu_install_odep(lsq, /* odep_list[] index */1, out2, context);
	      } else {
		ruu_install_odep(lsq, /* odep_list[] index */0, NA, context);
		ruu_install_odep(lsq, /* odep_list[] index */1, NA, context);
	      }

	      /* install operation in the RUU and LSQ */
	      n_dispatched++;

	      /* RUU_DK & LSQ_DK */
	      RUU_tail[context] = (RUU_tail[context] + 1) % RUU_size;
	      RUU_num++;
	      RUU_num_cxt[context]++; 
	      LSQ_tail[context] = (LSQ_tail[context] + 1) % LSQ_size;
	      LSQ_num++;
	      LSQ_num_cxt[context]++;

	      if (OPERANDS_READY(rs))
		{
		  /* eff addr computation ready, queue it on ready list */
		  readyq_enqueue(rs);
		}
	      /* issue may continue when the load/store is issued */
	      RSLINK_INIT(last_op[context], lsq);

	      /* issue stores only, loads are issued by lsq_refresh() */
	      if (((SS_OP_FLAGS(op) & (F_MEM|F_STORE)) == (F_MEM|F_STORE))
		  && OPERANDS_READY(lsq))
		{
		  /* panic("store immediately ready"); */
		  /* put operation on ready list, ruu_issue() issue it later */
		  readyq_enqueue(lsq);
		}
	    }
	  else /* !(SS_OP_FLAGS(op) & F_MEM) */
	    {
	      /* link onto producing operation */
	      ruu_link_idep(rs, /* idep_ready[] index */0, in1, context);
	      ruu_link_idep(rs, /* idep_ready[] index */1, in2, context);
	      ruu_link_idep(rs, /* idep_ready[] index */2, in3, context);

	      /* DY:  don't modify create vector if this instruction is idle */
	      if (!(rs->idle)) {
		/* install output after inputs to prevent self reference */
		ruu_install_odep(rs, /* odep_list[] index */0, out1, context);
		ruu_install_odep(rs, /* odep_list[] index */1, out2, context);
	      } else {
		ruu_install_odep(rs, /* odep_list[] index */0, NA, context);
		ruu_install_odep(rs, /* odep_list[] index */1, NA, context);
	      }

	      /* install operation in the RUU */
	      n_dispatched++;

	      /* RUU_DK */
	      RUU_tail[context] = (RUU_tail[context] + 1) % RUU_size;
	      RUU_num++;
	      RUU_num_cxt[context]++; 

	      /* issue op if all its reg operands are ready (no mem input) */
	      if (OPERANDS_READY(rs))
		{
		  /* put operation on ready list, ruu_issue() issue it later */
		  readyq_enqueue(rs);
		  /* issue may continue */
		  last_op[context] = RSLINK_NULL;
		}
	      else
		{
		  /* could not issue this inst, stall issue until we can */
		  RSLINK_INIT(last_op[context], rs);
		}
	    }
	}
      else
	{
	  /* this is a NOP, no need to update RUU/LSQ state */
	  rs = NULL;
	}
      /* one more instruction executed, speculative or otherwise */
      sim_total_insn++;

      if (SS_OP_FLAGS(op) & F_CTRL)
	sim_total_branches++;
      if (!spec_mode[context])
	{
	  /* one more non-speculative instruction executed */
          if (BITMAP_CLEAR_P(sim_shutdown_context, SIM_SHUTDOWN_SZ, context)) {
            sim_num_insn++;
            /* FIXMEDM: Just there to hide differences when stopping
               simulator */
            sim_num_insn_ctxt[context]++;
          } /* if */

	  /* if this is a branching instruction update BTB, i.e., only
	     non-speculative state is committed into the BTB */
	  if (SS_OP_FLAGS(op) & F_CTRL)
	    {
	      sim_num_branches++;
	      if (pred[context] && bpred_spec_update == spec_ID)
		{
		  bpred_update(
                    pred[context],
		    retstack[context],
		    context,
                    /* branch address */regs_PC[context],
                    /* actual target address */next_PC[context],
                    /* taken? */next_PC[context] != (regs_PC[context] +
                                                     sizeof(SS_INST_TYPE)),
                    /* pred taken? */pred_PC[context] != (regs_PC[context] +
                                                          sizeof(SS_INST_TYPE)),
                    /* correct pred? */pred_PC[context] == next_PC[context],
                    /* opcode */op,
                    /* jump through R31? */(RS) == 31,
                    /* predictor update ptr */&rs->dir_update);
		}
	    }

	  /* is the trace generator trasitioning into mis-speculation mode? */
	  if ((pred_PC[context] != next_PC[context])
              && !fetch_redirected[context])
	    {
	      /* entering mis-speculation mode, indicate this and save PC */
              if (!spec_mode[context] && 
                BITMAP_CLEAR_P(sim_shutdown_context, SIM_SHUTDOWN_SZ, context))
                backup_PC[context] = next_PC[context];
	      spec_mode[context] = TRUE;
	      rs->recover_inst = TRUE;
	      recover_PC[context] = next_PC[context];
              for(i = 0; i < SS_NUM_REGS; i++) {
                spec_regs_R[(context << 5) + i] = regs_R[context][i];
              } /* for */
              spec_regs_HI[context] = regs_HI[context];
              spec_regs_LO[context] = regs_LO[context];
              spec_regs_FCC[context] = regs_FCC[context];
              spec_regs_ACXT[context] = regs_ACXT;
              spec_regs_CXTCTRL[context] = regs_CXTCTRL[context];
	    }
	}

      /* entered decode/allocate stage, indicate in pipe trace */
      ptrace_newstage(context, pseq, PST_DISPATCH,
		  (pred_PC[context] != next_PC[context]) ? PEV_MPOCCURED : 0);
      if (op == NOP)
	{
	  /* end of the line */
	  ptrace_endinst(context, pseq);

	}

      /* update any stats tracked by PC */
      for (i=0; i<pcstat_nelt; i++)
	{
	  SS_COUNTER_TYPE newval;
	  int delta;

	  /* check if any tracked stats changed */
	  newval = STATVAL(pcstat_stats[i]);
	  delta = newval - pcstat_lastvals[i];
	  if (delta != 0)
	    {
	      stat_add_samples(pcstat_sdists[i], regs_PC[context], delta);
	      pcstat_lastvals[i] = newval;
	    }
	}

      /* consume instruction from IFETCH -> DISPATCH queue */
      fetch_head = (fetch_head+1) & (ruu_ifq_size - 1);
      fetch_num--;

      /* check for DLite debugger entry condition */
      made_check = TRUE;
      if (dlite_check_break(pred_PC[context],
			    is_write ? ACCESS_WRITE : ACCESS_READ,
			    addr, sim_num_insn, sim_cycle))
	dlite_main(regs_PC[context], pred_PC[context], sim_cycle);

      /* Next context number. Needed by next 'while' condition text */
      context = fetch_data[fetch_head].context;

    } /* end while */

  /* need to enter DLite at least once per cycle */
  if (!made_check)
    {
      if (dlite_check_break(/* no next PC */0,
			    is_write ? ACCESS_WRITE : ACCESS_READ,
			    addr, sim_num_insn, sim_cycle))
	dlite_main(regs_PC[context], /* no next PC */0, sim_cycle);
    }
}
  

/*
 *  RUU_FETCH() - instruction fetch pipeline stage(s)
 */

/* initialize the instruction fetch pipeline stage */
static void
fetch_init(void)
{
  /* allocate the IFETCH -> DISPATCH instruction queue */
  fetch_data =
    (struct fetch_rec *)calloc(ruu_ifq_size, sizeof(struct fetch_rec));
  if (!fetch_data)
    fatal("out of virtual memory");

  fetch_num = 0;
  fetch_tail = fetch_head = 0;
}

/* dump contents of fetch stage registers and fetch queue */
void
fetch_dump(FILE *stream)			/* output stream */
{
  int num, head;
  SS_CONTEXT_TYPE context;

  fprintf(stream, "** fetch stage state **\n");
  
  for(context = 0; context < context_num; context++) {
    fprintf(stream, "spec_mode: %s\n", spec_mode[context] ? "t" : "f");
    fprintf(stream, "pred_PC[%d]: 0x%08x, recover_PC: 0x%08x\n",
            context, pred_PC[context], recover_PC[context]);
    fprintf(stream, "fetch_regs_PC[%d]: 0x%08x, fetch_pred_PC: 0x%08x\n",
            context, fetch_regs_PC[context], fetch_pred_PC[context]);
    fprintf(stream, "\n");
  } /* for */

  fprintf(stream, "** fetch queue contents **\n");
  fprintf(stream, "fetch_num: %d\n", fetch_num);
  fprintf(stream, "fetch_head: %d, fetch_tail: %d\n",
          fetch_head, fetch_tail);

  num = fetch_num;
  head = fetch_head;
  while (num)
    {
      fprintf(stream, "idx: %2d: inst: `", head);
      ss_print_insn(fetch_data[head].IR, fetch_data[head].regs_PC, stream);
      fprintf(stream, "'\n");
      fprintf(stream, "         context: %d\n", fetch_data[head].context);
      fprintf(stream, "         regs_PC: 0x%08x, pred_PC: 0x%08x\n",
	      fetch_data[head].regs_PC, fetch_data[head].pred_PC);
      head = (head + 1) & (ruu_ifq_size - 1);
      num--;
    }
}


static int last_inst_missed = FALSE;
static int last_inst_tmissed = FALSE;

static int main_assigned_cycle = 0;
static int main_fetch_cycle = 0;
static int main_locked_cycle = 0;

/* fetch up as many instruction as one branch prediction and one cache line
   acess will support without overflowing the IFETCH -> DISPATCH QUEUE */
/* the fetch logic is changed by dongkeun */
static void
ruu_fetch(void)
{
  int i, j, k, lat[SS_NUM_CONTEXT], tlb_lat[SS_NUM_CONTEXT];
  SS_INST_TYPE inst;
  int stack_recover_idx[SS_NUM_CONTEXT];
  int branch_cnt = 0;
  static SS_CONTEXT_TYPE context = 0;
  static int idle_cycles_cnt = 0;
  int temp_element, used_context = 0; 
  int fetch_locked[SS_NUM_CONTEXT], context_locked = 0;
  int temp_icount[SS_NUM_CONTEXT], next_CXT[SS_NUM_CONTEXT];
  int active_array[SS_NUM_CONTEXT], active_num = 0;
  int main_assigned = 0, main_fetched = 0;

  /* DK: Copy the ICOUNT value (RUU_num_cxt) to 'temp_icount'. 
     In addition, add 'main offset' value to sub-threads' ICOUNT 
     to give higher priority to the main thread */
  for (i = 0; i < context_num; i++) { 

    temp_icount[i] = RUU_num_cxt[i];
    if (!context_master[i])
      temp_icount[i] += main_offset;

    /* If a context is inactive or blocked, lock it. */
    if (BITMAP_CLEAR_P(sim_active_context, SIM_ACTIVE_SZ, i) ||	ruu_fetch_issue_delay[i]) {
      fetch_locked[i] = TRUE;
    } else {
      fetch_locked[i] = FALSE;
      active_array[active_num++] = i;
    }
  }

  /* FIXME: This trick should be changed to more general form. */
  /* DK: Limit the sum of sub thread's ICOUNT < 0.75 * RUU_size */ 
  if ((RUU_num - temp_icount[0] >= (3*RUU_size)/4) || 
      (LSQ_num - LSQ_num_cxt[0] >= (3*LSQ_size)/4)) {
    for (i = 1; i < context_num; i++) 
      fetch_locked[i] = TRUE;

    if (fetch_locked[0]) {
      active_num = 0;
    } else {
      active_num = 1;
      active_array[0] = 0;
    }
  }

  context_locked = context_num - active_num;

  /* Check if all contexts are idle. Maybe a deadlock! */
  if (BITMAP_EMPTY_P(sim_active_context, SIM_ACTIVE_SZ)) 
    idle_cycles_cnt++;
  else
    idle_cycles_cnt = 0;
  if (idle_cycles_cnt == fetch_waiting) {
    fprintf(stderr, "All contexts are inactive for %d cycles!!!\n", fetch_waiting);
    longjmp(sim_exit_buf, 1);
  }

  if (fetch_locked[0])
    main_locked_cycle++;

  if (active_num == 0)
    return;

  for (k = 0; k < num_thread; k++) {
    /* DK: Sort the icount array in ascending order! */
    for (i = 0; i < active_num; i++) {
      for (j = 0; j < (active_num-i-1); j++) {
        if (temp_icount[active_array[j]] > temp_icount[active_array[j+1]]) {
	  temp_element = active_array[j];
          active_array[j] = active_array[j+1];
	  active_array[j+1] = temp_element;
        }
      }
    }

    if (context_master[active_array[0]])
      main_assigned = 1;

    next_CXT[k] = active_array[0];
    temp_icount[active_array[0]] += ruu_decode_width * fetch_speed / num_thread;
  } /* end for */

  if (main_assigned)
    main_assigned_cycle++;

  /* Regular behavior */
  for (i=0;
       /* fetch up to as many instruction as the DISPATCH stage can decode */
       i < (ruu_decode_width * fetch_speed)
       /* fetch until IFETCH -> DISPATCH queue fills */
       && fetch_num < ruu_ifq_size && context_locked != context_num /* DK */
       /* There are still active context */
       && !BITMAP_EMPTY_P(sim_active_context, SIM_ACTIVE_SZ);
       i++)
    {
      /* DK: Pick up the next active context. If we already fetched 
	 'ruu_decode_width * fetch_speed / num_thread' instructions
	 from one thread, go to the next active context. 
	 If we are running out of contexts, stop fetching. */  
      if ((i % (ruu_decode_width * fetch_speed / num_thread)) == 0) {
	context = next_CXT[used_context];

	if (++used_context > num_thread)
	  break;

	branch_cnt = 0; /* go to the next context */

	/* if this context is locked, go to the next context! */
	if (fetch_locked[context] == TRUE) {
	  i = (ruu_decode_width * fetch_speed / num_thread * used_context) - 1;
	  continue;
	}

	/* if this context is inactive or the fetch_issue_delay is non-zero, 
	   go to the next context! */
	if (BITMAP_CLEAR_P(sim_active_context, SIM_ACTIVE_SZ, context) ||
	    ruu_fetch_issue_delay[context]) {
	  fetch_locked[context] = TRUE;
	  context_locked++;
	  i = (ruu_decode_width * fetch_speed / num_thread * used_context) - 1;
	  continue;
	}
      } /* end if */

      if (ruu_fetch_issue_delay[context]) {
	fetch_locked[context] = TRUE;
	context_locked++;
	i = (ruu_decode_width * fetch_speed / num_thread * used_context) - 1;
	branch_cnt = 0; /* go to the next context */
	continue;
      }	

      stack_recover_idx[context] = 0;
      current_context = context;

      /* fetch an instruction at the next predicted fetch address */
      fetch_regs_PC[context] = fetch_pred_PC[context];

      if (interrupt_raised) {
        fprintf(stderr, "fetch_regs_PC[%d] = 0x%08x\n",
                context, fetch_regs_PC[context]);
        interrupt_raised--;
      }

      /* is this a bogus text address? (can happen on mis-spec path) */
      if (ld_text_base[context] <= fetch_regs_PC[context]
	  && fetch_regs_PC[context] <
          (ld_text_base[context]+ld_text_size[context])
	  && !(fetch_regs_PC[context] & (sizeof(SS_INST_TYPE)-1)))
	{
	  /* read instruction from memory */
	  mem_access(Read, context, fetch_regs_PC[context],
                     &inst, sizeof(SS_INST_TYPE));

	  /* address is within program text, read instruction from memory */
	  lat[context] = cache_il1_lat;
	  if (cache_il1[context])
	    {
	      /* access the I-cache */
	      lat[context] =
		cache_access(cache_il1[context], Read, fetch_regs_PC[context],
			     mmu->config[context],
			     NULL, sizeof(SS_INST_TYPE),
                             sim_cycle, NULL, NULL);
	      if (lat[context] > cache_il1_lat)
		last_inst_missed = TRUE;
	    }

	  if (itlb[context])
	    {
	      /* access the I-TLB, NOTE: this code will initiate
		 speculative TLB misses */
	      tlb_lat[context] =
		cache_access(itlb[context], Read, fetch_regs_PC[context],
			     mmu->config[context],
			     NULL, sizeof(SS_INST_TYPE),
                             sim_cycle, NULL, NULL);
	      if (tlb_lat[context] > 1)
		last_inst_tmissed = TRUE;

	      /* I-cache/I-TLB accesses occur in parallel */
	      lat[context] = MAX(tlb_lat[context], lat[context]);
	    }

	  /* I-cache/I-TLB miss? assumes I-cache hit >= I-TLB hit */
	  if ((lat[context] != cache_il1_lat) &&
	      (!icache_thrash_detect[context]))
	    {
	      /* I-cache miss, block fetch until it is resolved */
	      ruu_fetch_issue_delay[context] = lat[context] - 1;
              ruu_fetch_issue_ready =
                MAX(ruu_fetch_issue_ready, ruu_fetch_issue_delay[context]);
	      fetch_locked[context] = TRUE;
	      context_locked++;
	      i = (ruu_decode_width * fetch_speed / num_thread * used_context) - 1;
	      branch_cnt = 0; /* go to the next context */
	      continue;
	    }
	  icache_thrash_detect[context] = FALSE;
	  /* else, I-cache/I-TLB hit */
	}
      else
	{
	  /* fetch PC is bogus, send a NOP down the pipeline */
	  inst = SS_NOP_INST;
	}

      /* have a valid inst, here */

      /* possibly use the BTB target */
      if (pred[context])
	{
	  /* get the next predicted fetch address; only use branch predictor
	     result for branches (assumes pre-decode bits); NOTE: returned
	     value may be 1 if bpred can only predict a direction */
	  if (SS_OP_FLAGS(SS_OPCODE(inst)) & F_CTRL) {
	    fetch_pred_PC[context] =
	      bpred_lookup(pred[context],
			   retstack[context],
			   context,
			   /* branch address */fetch_regs_PC[context],
			   /* target address *//* FIXME: not computed */0,
			   /* opcode */SS_OPCODE(inst),
			   /* jump through r31? */(RS) == 31,
			   /* updt */&(fetch_data[fetch_tail].dir_update),
			   /* RSB index */&(stack_recover_idx[context]));
          } else {
	    fetch_pred_PC[context] = 0;
          } /* if */

	  /* valid address returned from branch predictor? */
	  if (!fetch_pred_PC[context])
	    {
	      /* no predicted taken target, attempt not taken target */
	      fetch_pred_PC[context] =
                fetch_regs_PC[context] + sizeof(SS_INST_TYPE);
	    }
	  else
	    {
	      /* go with target, NOTE: discontinuous fetch, so terminate */
	      branch_cnt++;
	      if (branch_cnt >= fetch_speed) {
	        fetch_locked[context] = TRUE;
	        context_locked++;
		i = (ruu_decode_width * fetch_speed / num_thread * used_context) - 1;
		branch_cnt = 0; /* go to the next context */
	      } /* end if */
	    }
	}
      else
	{
	  /* no predictor, just default to predict not taken, and
	     continue fetching instructions linearly */
	  fetch_pred_PC[context] = fetch_regs_PC[context] +
            sizeof(SS_INST_TYPE);
	}

      /* commit this instruction to the IFETCH -> DISPATCH queue */
      fetch_data[fetch_tail].context = context;
      fetch_data[fetch_tail].IR = inst;
      fetch_data[fetch_tail].regs_PC = fetch_regs_PC[context];
      fetch_data[fetch_tail].pred_PC = fetch_pred_PC[context];
      fetch_data[fetch_tail].stack_recover_idx = stack_recover_idx[context];
      fetch_data[fetch_tail].ptrace_seq = ptrace_seq[context]++;

      /* for pipe trace */
      ptrace_newinst(context, fetch_data[fetch_tail].ptrace_seq,
		     inst, fetch_data[fetch_tail].regs_PC, 0);
      ptrace_newstage(context, fetch_data[fetch_tail].ptrace_seq, PST_IFETCH,
		      ((last_inst_missed ? PEV_CACHEMISS : 0)
		       | (last_inst_tmissed ? PEV_TLBMISS : 0)));
      last_inst_missed = FALSE;
      last_inst_tmissed = FALSE;

      /* adjust instruction fetch queue */
      fetch_tail = (fetch_tail + 1) & (ruu_ifq_size - 1);
      fetch_num++;

      if (context_master[context] && !main_fetched) {
	main_fetched = 1;
	main_fetch_cycle++;
      } 
    }
} /* ruu_fetch() */

/* default machine state accessor, used by DLite */
static char *					/* err str, NULL for no err */
simoo_mstate_obj(FILE *stream,			/* output stream */
		 char *cmd)			/* optional command string */
{
  if (!cmd || !strcmp(cmd, "help"))
    fprintf(stream,
"mstate commands:\n"
"\n"
"    mstate help   - show all machine-specific commands (this list)\n"
"    mstate stats  - dump all statistical variables\n"
"    mstate res    - dump current functional unit resource states\n"
"    mstate ruu    - dump contents of the register update unit\n"
"    mstate lsq    - dump contents of the load/store queue\n"
"    mstate eventq - dump contents of event queue\n"
"    mstate readyq - dump contents of ready instruction queue\n"
"    mstate cv     - dump contents of the register create vector\n"
"    mstate rspec  - dump contents of speculative regs\n"
"    mstate mspec  - dump contents of speculative memory\n"
"    mstate fetch  - dump contents of fetch stage registers and fetch queue\n"
"\n"
	    );
  else if (!strcmp(cmd, "stats"))
    {
      /* just dump intermediate stats */
      sim_print_stats(stream);
    }
  else if (!strcmp(cmd, "res"))
    {
      /* dump resource state */
      res_dump(fu_pool, stream);
    }
  else if (!strcmp(cmd, "ruu"))
    {
      /* dump RUU contents */
      ruu_dump(stream);
    }
  else if (!strcmp(cmd, "lsq"))
    {
      /* dump LSQ contents */
      lsq_dump(stream);
    }
  else if (!strcmp(cmd, "eventq"))
    {
      /* dump event queue contents */
      eventq_dump(stream);
    }
  else if (!strcmp(cmd, "readyq"))
    {
      /* dump event queue contents */
      readyq_dump(stream);
    }
  else if (!strcmp(cmd, "cv"))
    {
      /* dump event queue contents */
      cv_dump(stream);
    }
  else if (!strcmp(cmd, "rspec"))
    {
      /* dump event queue contents */
      rspec_dump(stream);
    }
  else if (!strcmp(cmd, "mspec"))
    {
      /* dump event queue contents */
      mspec_dump(stream);
    }
  else if (!strcmp(cmd, "fetch"))
    {
      /* dump event queue contents */
      fetch_dump(stream);
    }
  else
    return "unknown mstate command";

  /* no error */
  return NULL;
}


/*
 * return true if given context has no instruction in the upper
 * stages of the pipeline
 */

unsigned int ready_to_resume(SS_CONTEXT_TYPE context)
{
  return 1;
} /* upstages_used_by_cxt */


/* DY:  redefine the execute engine for sim-fast mode execution */
#undef GPR
#undef SET_GPR
#undef FPR_L
#undef SET_FPR_L
#undef FPR_F
#undef SET_FPR_F
#undef FPR_D
#undef SET_FPR_D
#undef HI
#undef SET_HI
#undef LO
#undef SET_LO
#undef FCC
#undef SET_FCC
#undef READ_WORD
#undef READ_UNSIGNED_HALF
#undef READ_SIGNED_HALF
#undef READ_UNSIGNED_BYTE
#undef READ_SIGNED_BYTE
#undef WRITE_WORD
#undef WRITE_HALF
#undef WRITE_BYTE
#undef SYSCALL

/* general purpose registers */
#define GPR(CONTEXT,N)                  (regs_R[CONTEXT][N])
#define SET_GPR(CONTEXT,N,EXPR)         (regs_R[CONTEXT][N] = (EXPR))

/* floating point registers, L->word, F->single-prec, D->double-prec */
#define FPR_L(CONTEXT,N)                (regs_F.l[CONTEXT][(N)])
#define SET_FPR_L(CONTEXT,N,EXPR)	(regs_F.l[CONTEXT][(N)] = (EXPR))
#define FPR_F(CONTEXT,N)		(regs_F.f[CONTEXT][(N)])
#define SET_FPR_F(CONTEXT,N,EXPR)	(regs_F.f[CONTEXT][(N)] = (EXPR))
#define FPR_D(CONTEXT,N)		(regs_F.d[CONTEXT][(N) >> 1])
#define SET_FPR_D(CONTEXT,N,EXPR)	(regs_F.d[CONTEXT][(N) >> 1] = (EXPR))

/* miscellaneous register accessors */
#define HI(CONTEXT)			(regs_HI[CONTEXT])
#define SET_HI(CONTEXT,EXPR)   		(regs_HI[CONTEXT] = (EXPR))
#define LO(CONTEXT)			(regs_LO[CONTEXT])
#define SET_LO(CONTEXT,EXPR)		(regs_LO[CONTEXT] = (EXPR))
#define FCC(CONTEXT)			(regs_FCC[CONTEXT])
#define SET_FCC(CONTEXT,EXPR)		(regs_FCC[CONTEXT] = (EXPR))

/* precise architected memory state accessor macros, all unsafe for speed  */
#define READ_WORD(SRC, CONTEXT)						\
  (__MEM_READ_WORD(SRC, CONTEXT))
#define READ_UNSIGNED_HALF(SRC, CONTEXT)				\
  ((unsigned int)((unsigned short)__MEM_READ_HALF(SRC, CONTEXT)))
#define READ_SIGNED_HALF(SRC, CONTEXT)					\
  ((signed int)((signed short)__MEM_READ_HALF(SRC, CONTEXT)))
#define READ_UNSIGNED_BYTE(SRC, CONTEXT)				\
  ((unsigned int)((unsigned char)__MEM_READ_BYTE(SRC, CONTEXT)))
#define READ_SIGNED_BYTE(SRC, CONTEXT)					\
  ((unsigned int)((signed int)((signed char)__MEM_READ_BYTE(SRC, CONTEXT))))

#define WRITE_WORD(SRC, DST, CONTEXT)					\
  (__MEM_WRITE_WORD((DST), (CONTEXT), (unsigned int)(SRC)))
#define WRITE_HALF(SRC, DST, CONTEXT)					\
  (__MEM_WRITE_HALF((DST), (CONTEXT), (unsigned short)((unsigned int)(SRC))))
#define WRITE_BYTE(SRC, DST, CONTEXT)					\
  (__MEM_WRITE_BYTE((DST), (CONTEXT), (unsigned char)((unsigned int)(SRC))))

/* system call handler macro */
#define SYSCALL(INST, CONTEXT)	(ss_syscall(mem_access, CONTEXT, INST))


/* start simulation, program loaded, processor precise state initialized */
void
sim_main(void)
{
  SS_INST_TYPE inst;			/* actual instruction bits */
  SS_ADDR_TYPE next_PC[SS_NUM_CONTEXT]; /* actual target PC address */ 
  SS_ADDR_TYPE target_PC[SS_NUM_CONTEXT]; /* next target PC address */
  SS_ADDR_TYPE addr;			/* effective address, if load/store */
  int is_write;				/* store? */
  SS_CONTEXT_TYPE context;		/* context number */
  double icount = 0.0;				/* instructions in fast mode */

  fprintf(stderr, "sim: ** starting *fast* functional simulation **\n");
  
  /* ignore any floating point exceptions, they may occur on mis-speculated
     execution paths */
  signal(SIGFPE, SIG_IGN);

  for (context = 0; context < context_num; context++) {

    if (!context_master[context])
      continue;

    regs_PC[context] = ld_prog_entry[context] - SS_INST_SIZE;
    next_PC[context] = regs_PC[context] + SS_INST_SIZE;

    current_context = context;
    vsim_mode = FALSE;
    icount = 0;

    /* main simulator loop, NOTE: the pipe stages are traverse in reverse order
       to eliminate this/next state synchronization and relaxation problems */
    for (;;)
      {
	icount++;
	/* maintain $r0 semantics (in spec and non-spec space) */
	regs_R[context][0] = 0;

	/* default effective address (none) and access */
	addr = 0; is_write = FALSE;

	/* get the next instruction to execute */
	inst = __UNCHK_MEM_ACCESS(SS_INST_TYPE, regs_PC[context], context);

	if ((SS_OPCODE(inst) == NOP) &&
	    (inst.a & 0x00070000) == 0x00010000) /* VSIM_ENTER */
	  vsim_mode = TRUE;	  

      /* execute the instruction */
	switch (SS_OPCODE(inst)) {
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR)	\
	case OP:							\
          EXPR;								\
          break;
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)					\
	case OP:							\
          panic("attempted to execute a linking opcode");
#define CONNECT(OP)
#include "ss.def"
#undef DEFINST
#undef DEFLINK
#undef CONNECT
	default:
	  panic("bogus opcode");
	}

	/* execute next instruction */
	regs_PC[context] = next_PC[context];
	next_PC[context] += SS_INST_SIZE;	

	/* VSIM_ENTER */
	if ((vsim_mode == TRUE) || (icount == fastfwd_count && fastfwd_count)) {
	  vsim_mode = TRUE;
	  fprintf(stderr, "sim: ** %.0f instructions in fast mode for context %d **\n",
		  icount, context);
	  break;
	}
      } /* end for */
  }

  sim_vsim();
}


/* start simulation, program loaded, processor precise state initialized */
void
sim_vsim(void)
{
  SS_CONTEXT_TYPE context;

  fprintf(stderr, "sim: ** starting performance simulation **\n");

  /* set up program entry state */
  for(context = 0; context < context_num; context++){
    if (context_master[context]) {
      fetch_regs_PC[context] = regs_PC[context] - sizeof(SS_INST_TYPE);
      fetch_pred_PC[context] = regs_PC[context];
      backup_PC[context] = fetch_pred_PC[context];
      regs_PC[context] = regs_PC[context] - sizeof(SS_INST_TYPE);
    } else {
      SET_CPC(context, ld_prog_entry[context]-sizeof(SS_INST_TYPE));
      fetch_regs_PC[context] = regs_PC[context];
      fetch_pred_PC[context] = ld_prog_entry[context];
      backup_PC[context] = fetch_pred_PC[context];
    }
  }

  /* main simulator loop, NOTE: the pipe stages are traverse in reverse order
     to eliminate this/next state synchronization and relaxation problems */
  for (;;)
    {
      /* LSQ_DK: RUU/LSQ sanity checks - 
	 I should have checked RUU_num_cxt[context] < LSQ_num_cxt[context] 
	 for all contexts but I didn't, not to reduce the simulation speed. */
      /* check if pipetracing is still active */
      ptrace_check_active(regs_PC[0] /*regs_PC[context=0]*/,
                          sim_num_insn, sim_cycle);

      /* indicate new cycle in pipetrace */
      ptrace_newcycle(sim_cycle);

      /* commit entries from RUU/LSQ to architected register file */
      ruu_commit();

      /* service function unit release events */
      ruu_release_fu();

      /* ==> may have ready queue entries carried over from previous cycles */

      /* service result completions, also readies dependent operations */
      /* ==> inserts operations into ready queue --> register deps resolved */
      ruu_writeback();

      if (!bugcompat_mode)
	{
	  /* try to locate memory operations that are ready to execute */
	  /* ==> inserts operations into ready queue --> mem deps resolved */
	  lsq_refresh();

	  /* issue operations ready to execute from a previous cycle */
	  /* <== drains ready queue <-- ready operations commence execution */
	  ruu_issue();
	}

      /* decode and dispatch new operations */
      /* ==> insert ops w/ no deps or all regs ready --> reg deps resolved */
      ruu_dispatch();

      if (bugcompat_mode)
	{
	  /* try to locate memory operations that are ready to execute */
	  /* ==> inserts operations into ready queue --> mem deps resolved */
	  lsq_refresh();

	  /* issue operations ready to execute from a previous cycle */
	  /* <== drains ready queue <-- ready operations commence execution */
	  ruu_issue();
	}

      /* call instruction fetch unit if it is not blocked */
      ruu_fetch();

      for(context = 0; context < context_num; context++)
        if (ruu_fetch_issue_delay[context]) {
          ruu_fetch_issue_delay[context]--;
	  if (!ruu_fetch_issue_delay[context])
	    icache_thrash_detect[context] = TRUE;
	}

      /* go to next cycle */
      sim_cycle++;
      if (ruu_fetch_issue_ready)
        ruu_fetch_issue_ready--;
      if (next_interruption == sim_cycle) {
        if (next_interruption) {
          if (msg_port_test(interrupt_config[1]) == 1) {
            context = context_using_port(interrupt_config[1]);
                    
            fetch_pred_PC[context] =
              fetch_regs_PC[context] =
              recover_PC[context] =
              regs_PC[context] =
              rbackup_data[context].regs_PC;
            msg_port_write(/* interrupt source */ SS_INTER_CONTEXT, 
                       /* port */ interrupt_config[1],
                       /* FIXMEDM: value passed */ 0xffffffff,
                       /* receiver */ &context);
            /* FIXMEDM: wake up PC have to be set correctly 
                        but should be done in ruu_dispatch */
            if ((BITMAP_SET_P(sim_shutdown_context, 
                              SIM_SHUTDOWN_SZ, context)) 
                && (BITMAP_SET_P(sim_active_context,
                                 SIM_ACTIVE_SZ, context))) {
              BITMAP_SET(sim_restart_context,
                         SIM_RESTART_SZ, context);
            } else {              
              BITMAP_SET(sim_active_context, SIM_ACTIVE_SZ, context);       
              BITMAP_CLEAR(sim_shutdown_context, SIM_SHUTDOWN_SZ, context);
              BITMAP_CLEAR(sim_restart_context, SIM_RESTART_SZ, context);
            } /* if */
            spec_mode[context] = FALSE;
            regs_ACXT |= (1 << context);

          }
          if ((!interrupt_config[1]) && (interrupt_vector)) {
            context = 0;
            fprintf(stderr, "#### hard interrupt ####\n");
            interrupted_PC = CPC(context);
            fetch_pred_PC[context] =
              fetch_regs_PC[context] =
              recover_PC[context] =
              regs_PC[context] = interrupt_vector;
            interrupt_raised = TRUE;
            spec_mode[context] = FALSE;
          }
          /* FIXMEDM: doesn't support sim_shutdown_context !!! */
          next_interruption = interrupt_next_cycle(sim_cycle, interrupt_db);
        }
      }

      /* VSIM_EXIT */ 
      if (vsim_mode == FALSE) {
	fprintf(stderr, "sim: ** VSIM_EXIT\n");
	fflush(stderr);
	return; 
      }

      /* finish early? */
      if (max_insts && sim_num_insn >= max_insts) 
	return;
	
    } /* end for */

}



/* initialize the interrupt "handler" */
void
inter_init(void)
{
  interrupt_db = interrupt_init();

  interrupt_setconfig(interrupt_config[0] /* first cycle */,
                      interrupt_config[2] /* repetition */,
                      interrupt_config[1] /* port */,
                      interrupt_db);

  next_interruption = interrupt_next_cycle(0, interrupt_db);

} /* inter_init */




/* initialize the simulator */
void
sim_init(void)
{
  SS_CONTEXT_TYPE context;
  unsigned int i, j;

  for(context = 0; context < context_num; context++) {
    sim_num_insn_ctxt[context] = 0;
    spec_mode[context] = FALSE;
    ptrace_seq[context] = 0;
    fetch_dynamic_priority[context] = (context | FETCH_CXT_CACHE_MISS);
    ruu_fetch_issue_delay[context] = 0;
    icache_thrash_detect[context] = 0;
    for(i = 0; i < SS_NUM_REGS; i++) {
      spec_regs_R[(context << 5) + i] = 1;
    }
    for(i = 0; i < SS_NUM_REGS; i++) {
      spec_regs_R[(context << 5) + i] -= 1;
    } /* for */
  } /* for */

  /* disabling all contexts */
  BITMAP_CLEAR_MAP(sim_active_context, SIM_ACTIVE_SZ);
  BITMAP_CLEAR_MAP(sim_shutdown_context, SIM_SHUTDOWN_SZ);
  BITMAP_CLEAR_MAP(sim_restart_context, SIM_RESTART_SZ);

  /* clear message ports */
  msg_port_init();
  lock_port_init();
  semaphore_init();

  BITMAP_CLEAR_MAP(use_spec_R, R_BMAP_SZ);
  BITMAP_CLEAR_MAP(use_spec_F, F_BMAP_SZ);

  /* enabling selected contexts */
  /* use context_master mask to determine which contexts to enable */
  regs_ACXT = 0;
  for(context = 0; context < context_num; context++) {
    sim_num_insn_commit[context] = 0; 
    if (context_master[context]) {
      sim_num_start[context]++;
      BITMAP_SET(sim_active_context, SIM_ACTIVE_SZ, context);
      regs_ACXT |= (1 << context);
    }
  } /* for */

  sim_num_insn = 0;
  sim_num_refs = 0;

  /* initialize here, so symbols can be loaded */
  if (ptrace_nelt == 2)
    {
      /* generate a pipeline trace */
      ptrace_open(/* fname */ptrace_opts[0], /* range */ptrace_opts[1]);
    }
  else if (ptrace_nelt == 0)
    {
      /* no pipetracing */;
    }
  else
    fatal("bad pipetrace args, use: <fname|stdout|stderr> <range>");

  for(context = 0; context < context_num; context++)
  /* decode all instructions */
  {
    SS_ADDR_TYPE addr;
    SS_INST_TYPE inst;

    if (OP_MAX > 255)
      fatal("cannot do fast decoding, too many opcodes");

    debug("sim: decoding text segment...");
    for (addr=ld_text_base[context];
	 addr < (ld_text_base[context]+ld_text_size[context]);
	 addr += SS_INST_SIZE)
      {
	inst = __UNCHK_MEM_ACCESS(SS_INST_TYPE, addr, context);
	inst.a = (inst.a & ~0xff) | (unsigned int)SS_OP_ENUM(SS_OPCODE(inst));
	__UNCHK_MEM_ACCESS(SS_INST_TYPE, addr, context) = inst;
      }
  }

  /* initialize the simulation engine */
  fu_pool = res_create_pool("fu-pool", fu_config, N_ELT(fu_config));
  rslink_init(MAX_RS_LINKS);
  tracer_init();
  fetch_init();
  cv_init();
  eventq_init();
  readyq_init();
  ruu_init();
  lsq_init();
  inter_init();

  /* initialize the DLite debugger */
  dlite_init(simoo_reg_obj, simoo_mem_obj, simoo_mstate_obj);

  /* DY:  initialize the marker statistics */
  for (i = 0; i < SS_NUM_CONTEXT; i++) {
    mark1_total[i] = mark1_count[i] = last_mark1[i] = 0;
    mark2_total[i] = mark2_total_insn[i] = mark2_count[i] = 0;
    last_mark2[i] = last_mark2_insn[i] = last_mark[i] = 0;

    for (j = 0; j < 8; j++) {
      sem_p_ops[i][j] = sem_v_ops[i][j] = sem_suspends[i][j] = 0;
      prev_sem_time[i][j] = total_sleep_time[i][j] = 0;
    }
  }

  for (i = 0; i < 8; i++) {
    total_point_count[i] = 0;
    total_point2_time[i] = total_point3_time[i] = total_point4_time[i] = 0;
    total_point2_insn[i] = total_point3_insn[i] = total_point4_insn[i] = 0;
  }

#ifdef CACHE_PROF
  /* DK: initialize issued_addr_prof[] table */
  for (i = 0; i < MAX_CACHE_MISS_PROF; i++) {
    main_addr_prof[i].valid = 0;
    issued_addr_prof[i].valid = 0;
  }
#endif

  /* DK: initialize the kill_info table */
  for (i = 0; i < 100; i++) 
    kill_table[i].PC = kill_table[i].access = kill_table[i].kill = 0;

  /* DK: initialize the exception array */
  for (i = 0; i < 10; i++)
    exception_array[i] = 0;

  /* DK: initialize the kill_mode array */
  for (i = 0; i < SS_NUM_CONTEXT; i++)
    kill_mode[i] = 0; 
}

#ifdef CACHE_PROF
/* dump simulator-specific auxiliary simulator statistics */
int
prof_cmp(struct cache_miss_prof_ *i, struct cache_miss_prof_ *j) {
  if ((i)->count < (j)->count)
    return (1);
  else if ((i)->count > (j)->count)
    return (-1);
  else
    return(0);
}
#endif


void
sim_aux_stats(FILE *stream)             /* output stream */
{
  int i, j, total_lat = 0, total_cache_misses = 0;
  int total_L1_misses = 0, total_L2_misses = 0;

#ifdef CACHE_PROF
  if (cache_miss_num) {
    qsort(cache_miss_prof, cache_miss_num,
	  sizeof(struct cache_miss_prof_), (void *)prof_cmp);

    for (i = 0; i < cache_miss_num; i++) {
      total_cache_misses += cache_miss_prof[i].count;
      total_L1_misses += cache_miss_prof[i].L1_miss;
      total_L2_misses += cache_miss_prof[i].L2_miss;
    }

    fprintf(stderr, "\n\nCache Miss Profile Stats\n\n");
    fprintf(stderr, "There were %d unique cache missing loads\n",
	    cache_miss_num);
    fprintf(stderr, "Total cache misses: %d\n", total_cache_misses);
    fprintf(stderr, "Total L1 misses: %d\n", total_L1_misses);
    fprintf(stderr, "Total L2 misses: %d\n", total_L2_misses);
    fprintf(stderr, "Covered addresses: %d\n", covered_addr);
    fprintf(stderr, "Covered L1 misses: %d\n", covered_L1_miss);
    fprintf(stderr, "Covered L2 misses: %d\n", covered_L2_miss);
    fprintf(stderr, "Late addresses: %d\n", late_addr);
    fprintf(stderr, "Late L1 misses: %d\n", late_L1_miss);
    fprintf(stderr, "Late L2 misses: %d\n", late_L2_miss);
    fprintf(stderr, "Non loop misses: %d\n", non_loop_L1_miss + non_loop_L2_miss);
    fprintf(stderr, "Non loop L1 misses: %d\n", non_loop_L1_miss);
    fprintf(stderr, "Non loop L2 misses: %d\n", non_loop_L2_miss);
    fprintf(stderr, "\nPC       miss  latency  unknown  partial  non-loop\n\n");
    for (i = 0; i < cache_miss_num; i++) {
      fprintf(stderr, "%x:  %4d(%.2f), %.2f  %d  %d  %d\n", cache_miss_prof[i].PC,
	      cache_miss_prof[i].count,
	      (float)cache_miss_prof[i].count / (float)total_cache_misses,
	      (float)cache_miss_prof[i].lat / (float)cache_miss_prof[i].count,
	      cache_miss_prof[i].unknown_miss, cache_miss_prof[i].partial_miss,
	      cache_miss_prof[i].non_loop_miss);
      total_lat += cache_miss_prof[i].lat;
    }

    fprintf(stderr, "\nAverage Latency: %.2f\n", (float)total_lat/total_cache_misses);    
    fprintf(stderr, "\n\n");
  }
#endif

  fprintf(stderr, "\n\nMarker Stats\n\n");
  fprintf(stderr, "marker 1\n");
  fprintf(stderr, "total:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%d\t", mark1_total[i]);
  fprintf(stderr, "\ncount:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%d\t", mark1_count[i]);
  fprintf(stderr, "\navg:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%.2f\t", (float)mark1_total[i] / (float)mark1_count[i]);
  fprintf(stderr, "\nmarker 2\n");
  fprintf(stderr, "total:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%d\t", mark2_total[i]);
  fprintf(stderr, "\ncount:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%d\t", mark2_count[i]);
  fprintf(stderr, "\navg:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%.2f\t", (float)mark2_total[i] / (float)mark2_count[i]);
  fprintf(stderr, "\ninsn:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%d\t", mark2_total_insn[i]);
  fprintf(stderr, "\navg:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%.2f\t", (float)mark2_total_insn[i] / (float)mark2_count[i]);
  fprintf(stderr, "\n\n\n");

  fprintf(stderr, "Point Stats\n");
  fprintf(stderr, "\ncount:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%d\t", total_point_count[i]);
  fprintf(stderr, "\ntotal2:\n");
  fprintf(stderr, "\ncycle:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%d\t", total_point2_time[i]);
  fprintf(stderr, "\navg:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%.2f\t", (float)total_point2_time[i] / (float)total_point_count[i]);
  fprintf(stderr, "\ninsn:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%d\t", total_point2_insn[i]);
  fprintf(stderr, "\navg:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%.2f\t", (float)total_point2_insn[i] / (float)total_point_count[i]);
  fprintf(stderr, "\ntotal3:\n");
  fprintf(stderr, "\ncycle:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%d\t", total_point3_time[i]);
  fprintf(stderr, "\navg:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%.2f\t", (float)total_point3_time[i] / (float)total_point_count[i]);
  fprintf(stderr, "\ninsn:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%d\t", total_point3_insn[i]);
  fprintf(stderr, "\navg:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%.2f\t", (float)total_point3_insn[i] / (float)total_point_count[i]);
  fprintf(stderr, "\ntotal4:\n");
  fprintf(stderr, "\ncycle:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%d\t", total_point4_time[i]);
  fprintf(stderr, "\navg:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%.2f\t", (float)total_point4_time[i] / (float)total_point_count[i]);
  fprintf(stderr, "\ninsn:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%d\t", total_point4_insn[i]);
  fprintf(stderr, "\navg:\t");
  for (i = 0; i < context_num; i++)
    fprintf(stderr, "%.2f\t", (float)total_point4_insn[i] / (float)total_point_count[i]);
  fprintf(stderr, "\n\n\n");

  fprintf(stderr, "Semaphore Stats\n");
  fprintf(stderr, "\nsem_v count:\n");
  for (i = 0; i < 8; i++) {
    fprintf(stderr, "sem %d\t", i);
    for (j = 0; j < context_num; j++)
      fprintf(stderr, "%d\t", sem_v_ops[i][j]);
    fprintf(stderr, "\n");
  }
  fprintf(stderr, "\nsem_p count:\n");
  for (i = 0; i < 8; i++) {
    fprintf(stderr, "sem %d\t", i);
    for (j = 0; j < context_num; j++)
      fprintf(stderr, "%d\t", sem_p_ops[i][j]);
    fprintf(stderr, "\n");
  }
  fprintf(stderr, "\nsem_p block:\n");
  for (i = 0; i < 8; i++) {
    fprintf(stderr, "sem %d\t", i);
    for (j = 0; j < context_num; j++)
      fprintf(stderr, "%d\t", sem_suspends[i][j]);
    fprintf(stderr, "\n");
  }
  fprintf(stderr, "\nsem blocked cycle:\n");
  for (i = 0; i < 8; i++) {
    fprintf(stderr, "sem %d\t", i);
    for (j = 0; j < context_num; j++)
      fprintf(stderr, "%d\t", total_sleep_time[i][j]);
    fprintf(stderr, "\n");
  }

  /* DK: kill table stats */
  fprintf(stderr, "\n\nKill Table Stats\n\n");
  for (i = 0; i < kill_table_num; i++)
    fprintf(stderr, "PC: %x  Access: %8d  Kill: %8d  Rate: %.2f\n", 
	kill_table[i].PC, kill_table[i].access, kill_table[i].kill,
	(float) kill_table[i].kill/ (float) kill_table[i].access);
  fprintf(stderr, "\n");

  /* DK: exception stats */
  fprintf(stderr, "\nException Stats\n\n");
  fprintf(stderr, "bad size or alignment (mem_valid)     %4d\n", exception_array[1]);
  fprintf(stderr, "segmentation violation (mem_valid)    %4d\n", exception_array[2]); 
  fprintf(stderr, "bad size or alignment (mem_access)    %4d\n", exception_array[3]); 	
  fprintf(stderr, "segmentation violation (mem_access)   %4d\n", exception_array[4]); 
  fprintf(stderr, "bad jump alignment (ss.def)           %4d\n", exception_array[5]); 
  fprintf(stderr, "\n\n");

  fprintf(stderr, "Fetch Stage Stats (total cycle = %d)\n\n", (unsigned) sim_cycle);
  fprintf(stderr, "Main context was fetch-locked - %d cycles (%.2f\%)\n", 
	main_locked_cycle, (float) (100.0 * main_locked_cycle / (unsigned) sim_cycle));
  fprintf(stderr, "Main context had lower priority, so couldn't get slots - %d cycles (%.2f\%)\n", 
	(unsigned) sim_cycle - main_locked_cycle - main_assigned_cycle, (float) (100.0 * 
	((unsigned) sim_cycle - main_locked_cycle - main_assigned_cycle) / (unsigned) sim_cycle));
  fprintf(stderr, "Main context got fetch slots and fetched - %d cycles (%.2f\%)\n", 
	main_fetch_cycle, (float) (100.0 * main_fetch_cycle / (unsigned) sim_cycle));
  fprintf(stderr, "Main context got fetch slots but didn't fetch - %d cycles (%.2f\%)\n", 
	main_assigned_cycle - main_fetch_cycle, 
	(float) (100.0 * (main_assigned_cycle - main_fetch_cycle) / (unsigned) sim_cycle));
  fprintf(stderr, "\n\n");

  fprintf(stderr, "Dispatch Stage Stats\n\n");
  fprintf(stderr, "RUU was full for %d cycles (%.2f\%)\n", ruu_full_cycle,
	(float) (100.0 * ruu_full_cycle / (unsigned) sim_cycle));
  fprintf(stderr, "LSQ was full for %d cycles whereas RUU was not full (%.2f\%)\n", 
	lsq_full_cycle, (float) (100.0 * lsq_full_cycle / sim_cycle));
  fprintf(stderr, "\n"); 
}

/* un-initialize the simulator */
void
sim_uninit(void)
{
  if (ptrace_nelt > 0)
    ptrace_close();
}
